
#include "fl_conf_id.h"


#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <sys/errno.h>


#if FL_OS_NAME_V_IRIX
#       if FL_OS_VERSION_V_6_5
#               if FL_OS_RELEASE_V_30


#include <errno.h>
#include <limits.h>


#               endif
#       endif
#endif


%GLOB:ufl_c:START%


#include "fl_test.h"


extern unsigned long long fl_test_env;
extern char *fl_test_env_function_name;
extern char *fl_test_env_test_name;


/******************************************************************************
 * Data type definitions (for portability)
 ******************************************************************************/
typedef unsigned char fl_uint8_t;
typedef signed char fl_int8_t;
typedef unsigned short fl_uint16_t;
typedef signed short fl_int16_t;
typedef unsigned int fl_uint32_t;
typedef signed int fl_int32_t;
typedef unsigned long long fl_uint64_t;
typedef signed long long fl_int64_t;


#ifdef FL_DEBUG
 
 
#define FL_DEBUG_PRINTF(ARG) printf ## ARG;


#else
 
 
#define FL_DEBUG_PRINTF(ARG)
 
 
#endif


#define FL_LOOP_FOREVER 1


#define FL_VERSION "1.2.1"


#if FL_OS_NAME_V_OSF1


#elif FL_OS_NAME_V_IRIX


#define FL_OS "IRIX"


#endif


// NOTE: File looper commands


#define FL_CMD_NULL 0
#define FL_CMD_ATTACH 1
#define FL_CMD_DETACH 2
#define FL_CMD_LIST 3
#define FL_CMD_HELP 4
#define FL_CMD_VERSION 5


// NOTE: Path length vs path size
//
//  The path length is the maximum number of characters allowed
//  in a path given to file looper (both user space programs and
//  kernel modules). This length accounts for the leading 'r' letter
//  that is prepended to the CHR device file name and does not
//  include the terminating NULL byte.
//
//  The path size is the total number of bytes required to store a path.
//
//  It is always one byte more than the path length, to account for the
//  terminating NULL byte.


#define PATH_LENGTH_MAX (PATH_MAX - 1 - 1)
#define PATH_SIZE_MAX (PATH_LENGTH_MAX + 1)


// Request size
#define FL_REQ_MAX_SIZE (4 + 2 + PATH_MAX + 2 + PATH_MAX)


// Pathname buffers for device and file
char dev_real[PATH_MAX + 1];
char file_real[PATH_MAX + 1];


// NOTE: Command line arguments


struct args_s
{
	char *dev;
	char *file;
	fl_uint8_t attach;
	fl_uint8_t detach;
	fl_uint8_t list;
	fl_uint8_t help;
	fl_uint8_t version;
	fl_uint32_t req;
	fl_uint32_t cmd;
};


// NOTE: Packet
//
//  A packet holds a logical representation of the
//  data sent to or received from the kernel.


struct packet_s
{
	char *data;
	char *data_req;
	fl_uint32_t size;


	// Return code from kernel module
	fl_int32_t ret;
};


// NOTE: Global variables


static struct args_s args;
static struct packet_s packet;


// NOTE: Private functions


fl_uint64_t strnlen(char *string_source,	// -- IN
		    fl_uint64_t string_max_num)	// -- IN
{
	char *string_pointer;
	fl_uint64_t string_length;


	// NOTE: Arguments


	// IN: Check for source string
	if (string_source == NULL)
	{
		return 0;	// ERROR: No source string
	}


	// IN: Check for zero limit
	if (string_max_num == 0)
	{
		return 2;	// SUCCESS: Zero string limit
	}


	// NOTE: Processing


	// Traverse string
	string_length = 0;
	string_pointer = string_source;
	while (*string_pointer != '\0')
	{
		// Increment length
		string_length++;


		// Check for length limit
		if (string_length == string_max_num)
		{
			break;
		}


		// Increment string pointer
		string_pointer++;
	}


	return string_length;	// SUCCESS
}


int fl_packet_new_req()
{
	char *dp;
	char *fp;
	fl_uint16_t dev_size;
	fl_uint16_t file_size;
	fl_uint16_t field_index;


	//FL_TEST_RET;


	// NOTE: Processing


        // NOTE: Request packet format
        //
        //  A request packet is composed of the following fields, sequentially
        //  laid out in a byte stream:
        //
        //  Command:                    C (4 bytes - unsigned int)
        //  Command data:               ? (? bytes - unsigned int)
        //----------------------------------------------------------------------------------
        //  Command: FL_CMD_ATTACH
        //  Command data:               D  (? bytes - unsigned int)
        //    Device file name size:    DS (2 bytes - unsigned short)
        //    Device file name:         DN (? bytes - char[])
        //    File name size:           FS (2 bytes - unsigned short)
        //    File name:                FN (? bytes - char[])
        //
        //  Command: FL_CMD_DETACH
        //  Command data:               D  (? bytes - unsigned int)
        //    Device file name size:    DS (2 bytes - unsigned short)
        //    Device file name:         DN (? bytes - char[])
        //
        //  Command: FL_CMD_LIST
        //  Command data:               D (0 bytes - unsigned int)
        //
        //  All multibyte data values are copied in host endianness.


	// Set minimum packet size
	packet.size = 4;


	// Switch commands for packet allocation
	switch (args.cmd)
	{
		case FL_CMD_ATTACH:

			// Calculate packet size
                        dev_size = strlen(args.dev) + 1;
                        file_size = strlen(args.file) + 1;
                        packet.size += 2 +		// Device file name size
                                       dev_size +	// Device file name
                                       2 +		// File name size
                                       file_size;	// File name


                        FL_DEBUG_PRINTF(("[ufl][packet_new] INFO: dev_size = %u\n",
					 dev_size))


                        FL_DEBUG_PRINTF(("[ufl][packet_new] INFO: file_size = %u\n",
					 file_size))


                        FL_DEBUG_PRINTF(("[ufl][packet_new] INFO: packet.size = %u\n",
					 packet.size))


			break;

		case FL_CMD_DETACH:

                        // Calculate packet size
                        dev_size = strlen(args.dev) + 1;
                        packet.size += 2 +		// Device file name size
                                       dev_size;	// Device file name


			FL_DEBUG_PRINTF(("[ufl][packet_new] INFO: dev_size = %u\n",
					 dev_size))


			FL_DEBUG_PRINTF(("[ufl][packet_new] INFO: packet.size = %u\n",
					 packet.size))


			break;

		case FL_CMD_LIST:

			FL_DEBUG_PRINTF(("[ufl][packet_new] INFO: packet.size = %u\n",
					 packet.size))


			break;

		default:

			return 0;	// ERROR: Unknown command


			break;
	}


	packet.data = packet.data_req;


	// Switch commands
	switch (args.cmd)
	{
		case FL_CMD_ATTACH:

			// NOTE: Copy fields
			//
        		// Copy successive fields (fp) to data (dp).


        		// Set data pointer
        		dp = packet.data;


        		// Copy command code
        		fp = (char *) &args.cmd;
        		for (field_index = 0;
			     field_index < 4;
			     field_index++)
			{
                		*dp++ = *fp++;
			}


			// Copy device file name size
			fp = (char *) &dev_size;
			for (field_index = 0;
			     field_index < 2;
			     field_index++)
			{
				*dp++ = *fp++;
			}


			// Copy device file name
			fp = args.dev;
			for (field_index = 0;
			     field_index < dev_size;
			     field_index++)
			{
				*dp++ = *fp++;
			}


			// Copy file name size
        		fp = (char *) &file_size;
        		for (field_index = 0;
			     field_index < 2;
			     field_index++)
			{
                		*dp++ = *fp++;
			}


        		// Copy file name
        		fp = args.file;
        		for (field_index = 0;
			     field_index < file_size;
			     field_index++)
			{
                		*dp++ = *fp++;
			}


			break;


		case FL_CMD_DETACH:

			// NOTE: Copy fields
			//
        		// Copy successive fields (fp) to data (dp).


                        // Set data pointer
                        dp = packet.data;


                        // Copy command code
                        fp = (char *) &args.cmd;
                        for (field_index = 0;
			     field_index < 4;
			     field_index++)
			{
                                *dp++ = *fp++;
			}


                        // Copy device file name size
                        fp = (char *) &dev_size;
                        for (field_index = 0;
			     field_index < 2;
			     field_index++)
			{
                                *dp++ = *fp++;
			}


                        // Copy device file name
                        fp = args.dev;
                        for (field_index = 0;
			     field_index < dev_size;
			     field_index++)
			{
                                *dp++ = *fp++;
			}


			break;

		case FL_CMD_LIST:

			// NOTE: Copy fields
			//
        		// Copy successive fields (fp) to data (dp).


                        // Set data pointer
                        dp = packet.data;


                        // Copy command
                        fp = (char *) &args.cmd;
                        for (field_index = 0;
			     field_index < 4;
			     field_index++)
			{
                                *dp++ = *fp++;
			}


			break;

		default:


			break;
	}


	return 1;	// SUCCESS
}


int fl_packet_parse()
{
	char *dp;
	char *ep;
	char *fp;
	char *dev;
	char *file;
	fl_uint16_t field_index;
	fl_uint16_t dev_size;
	fl_uint16_t file_size;
	fl_uint16_t dev_length;
	fl_uint16_t file_length;
	fl_uint32_t cmd;
	fl_uint32_t list_index;
	fl_uint32_t list_length;
	fl_uint64_t size;


	// NOTE: Processing


        // NOTE: Response packet format
        //
        //  A response packet is composed of the following fields, sequentially
        //  laid out in a byte stream:
        //
        //  Command:                            C (4 bytes - unsigned int)
        //  Command data:                       D (? bytes - unsigned int)
        //--------------------------------------------------------------------------------------
        //  Command: FL_CMD_LIST
        //  Command data:                       D  (? bytes - unsigned int)
        //    Length of list:                   L  (4 bytes - unsigned int)
        //    0. Device file name size:         DS (2 bytes - unsigned short)
        //    0. Device file name:              DN (? bytes - char[])
        //    0. File name size:                FS (2 bytes - unsigned short)
        //    0. File name:                     FN (? bytes - char[])
        //    0. File size:                     S  (8 bytes - unsigned long long)
        //    ...
        //    L - 1. Device file name size:     DS (2 bytes - unsigned short)
        //    L - 1. Device file name:          DN (? bytes - char[])
        //    L - 1. File name size:            FS (2 bytes - unsigned short)
        //    L - 1. File name:                 FN (? bytes - char[])
        //    L - 1. File size:                 S  (8 bytes - unsigned long long)
        //
        //  All multibyte data values are copied in host endianness.


	// Set data pointer
	dp = packet.data;


	// Set data end pointer
	ep = packet.data + _PAGESZ - 1;


	FL_DEBUG_PRINTF(("[ufl][packet_parse] INFO: Get command dp = %p\n",
			 dp))


	// Get command
	fp = (char *) &cmd;
	for (field_index = 0;
	     field_index < 4;
	     field_index++)
	{
		*fp++ = *dp++;
	}


	FL_DEBUG_PRINTF(("[ufl][packet_parse] INFO: switch (cmd)\n"))


	// Switch commands
	switch (cmd)
	{
		case FL_CMD_ATTACH:
		case FL_CMD_DETACH:

			FL_DEBUG_PRINTF(("[ufl][packet_parse] INFO: Command: %s\n",
					 cmd == FL_CMD_ATTACH ? "FL_CMD_ATTACH" : "FL_CMD_DETACH"))


			// Get return code
        		fp = (char *) &packet.ret;
			for (field_index = 0;
			     field_index < 4;
			     field_index++)
			{
				*fp++ = *dp++;
			}


			break;


		case FL_CMD_LIST:

			FL_DEBUG_PRINTF(("[ufl][packet_parse] INFO: Command: FL_CMD_LIST\n"))


			// Get length of list
			fp = (char *) &list_length;
			for (field_index = 0;
			     field_index < 4;
			     field_index++)
			{
				*fp++ = *dp++;
			}


			FL_DEBUG_PRINTF(("[ufl][packet_parse] INFO: Length of list: %u\n",
					 list_length))


			// Check length of list
			if (list_length > 0)
			{
				// Traverse listing
				for (list_index = 0;
				     list_index < list_length;
				     list_index++)
				{
					// Get device file name size
					fp = (char *) &dev_size;
					for (field_index = 0;
					     field_index < 2;
					     field_index++)
					{
						*fp++ = *dp++;
					}


					// Get device file name
					dev = dp;
					dp += dev_size;


					// Check device file name size
					if (dev_size == 0)
					{
						return 0;	// ERROR: Device file name size is zero
					}

					if (dp >= ep)
					{
						return -1;	// ERROR: Device file name size too large
					}


					// Check device file name string size
					dev_length = strnlen(dev, PATH_LENGTH_MAX);
					if (dev_length != (dev_size - 1))
					{
						return -2;	// ERROR: Device file name has less bytes than specified
					}

					if (dev_length == PATH_LENGTH_MAX)
					{
						if (dev[PATH_LENGTH_MAX] != '\0')
						{
							return -3;	// ERROR: Device pathname too long, only %u bytes are allowed\n", PATH_LENGTH_MAX
						}
					}


					// Get file name size
					fp = (char *) &file_size;
                                        for (field_index = 0;
					     field_index < 2;
					     field_index++)
					{
                                                *fp++ = *dp++;
					}


                                        // Get file name
                                        file = dp;
					dp += file_size;


					// Check file name size
					if (file_size == 0)
					{
						return -4;	// ERROR: File name size is zero
					}

					if (dp >= ep)
					{
						return -5;	// ERROR: File name size too large
					}


					// Check file name string size
                                        file_length = strnlen(file,
							      PATH_LENGTH_MAX);
                                        if (file_length != (file_size - 1))
                                        {
                                                return -6;      // ERROR: File name has less bytes than specified
                                        }


					if (file_length == PATH_LENGTH_MAX)
					{
						if (file[PATH_LENGTH_MAX] != '\0')
						{
							return -7;	// ERROR: File pathname too long, only %u bytes are allowed\n", PATH_LENGTH_MAX
						}
					}


					// Get file size
					fp = (char *) &size;
					for (field_index = 0;
					     field_index < 8;
					     field_index++)
					{
                                                *fp++ = *dp++;
					}


					// NOTE: Display /dev/loop/ prefixed paths


#if FL_OS_NAME_V_OSF1
#       if FL_OS_VERSION_V_5_1
#               if FL_OS_RELEASE_V_B


#               endif
#       endif
#elif FL_OS_NAME_V_IRIX
#       if FL_OS_VERSION_V_6_5
#               if FL_OS_RELEASE_V_30


					// Print entry
					printf("%s: %s (%llu bytes)\n",
					       dev,
					       file,
					       size);


#               endif
#       endif
#endif


				}
			}


			break;

		default:

			// NOTE: Never reached


			break;
	}


	return 1;	// SUCCESS
}


#if FL_OS_NAME_V_OSF1
#       if FL_OS_VERSION_V_5_1
#               if FL_OS_RELEASE_V_B


#               endif
#       endif
#endif


int fl_args_process(int argc,		// -- IN
		    char *argv[])	// -- IN
{
	// NOTE: Arguments


	// IN: Check number of arguments
	if (argc == 1)
	{
		FL_TEST("fl_args_process",\
			"no_args")
		return 0;	// ERROR: No arguments
	}


	// NOTE: Processing


	// Check for known command
	if (strcmp("-a",
		   argv[1]) == 0)
	{
		args.attach = 1;
		args.req = 4;
	}
	else if (strcmp("-d",
			argv[1]) == 0)
	{
		args.detach = 1;
		args.req = 3;
	}
	else
	{
		if ((strcmp("-l",
			    argv[1]) != 0) &&
		    (strcmp("-h",
			    argv[1]) != 0) &&
		    (strcmp("-v",
			    argv[1]) != 0))
		{
			fprintf(stderr,
				"ERROR: Unknown command\n");


			return -1;	// ERROR: Unknown command
		}


		args.req = 2;
	}


	// Check for required arguments
	if (argc > args.req)
	{
		fprintf(stderr,
			"ERROR: Too many arguments\n");


		FL_TEST("fl_args_process",\
			"too_many_args")
		return -2;	// ERROR: Too many arguments
	}
	else if (argc < args.req)
	{
                fprintf(stderr,
			"ERROR: Too few arguments\n");


                FL_TEST("fl_args_process",\
			"too_few_args")
                return -3;      // ERROR: Too few arguments
	}


	// Switch through argc
	switch (argc)
	{
		case 2:

			// NOTE: Switches -l, -v and -h
			//
			//  Syntax:
			//
			//    $ fl -l
			//    $ fl -v
			//    $ fl -h


			// Switch flags
			if (strcmp("-l",
				   argv[1]) == 0)
			{
				// NOTE: $ fl -l


				// Set flags
				args.list = 1;
				args.cmd = FL_CMD_LIST;
			}
			else if (strcmp("-v",
					argv[1]) == 0)
			{
				// NOTE: $ fl -v


				// Set flags
				args.version = 1;
				args.cmd = FL_CMD_VERSION;
			}
			else if (strcmp("-h",
					argv[1]) == 0)
			{
				// NOTE: $ fl -h


				// Set flags
				args.help = 1;
				args.cmd = FL_CMD_HELP;
			}


			break;

		case 3:

			// NOTE: Switch -d
			//
			//  Syntax:
			//
			//    $ fl -d <dev>


			// Check -d flag
                        if (args.detach == 1)
                        {
				args.cmd = FL_CMD_DETACH;
			}


                        // Set flag arguments
                        args.dev = argv[2];


			break;

		case 4:

			// NOTE: Switch -d
			//
			//  Syntax:
			//
			//    $ fl -a <dev> <file>


			// Check -a flag
			if (args.attach == 1)
			{
				args.cmd = FL_CMD_ATTACH;
			}


			// Set flag arguments
			args.dev = argv[2];
			args.file = argv[3];


			break;

		default:

			break;
	}


	return 1;	// SUCCESS
}


void fl_version_print()
{
	// NOTE: Processing


	printf("File looper version %s on %s\n",
	       FL_VERSION,
	       FL_OS);
}


void fl_syntax_print()
{
	// NOTE: Processing


        printf("File looper syntax:\n\
\n\
  $ fl -a <dev> <file>\n\
  Attach file to device.\n\
\n\
  $ fl -d <dev>\n\
  Detach file from device.\n\
\n\
  $ fl -l\n\
  List attached files.\n\
\n\
  $ fl -h\n\
  Print help.\n\
\n\
  $ fl -v\n\
  Print version.\n\
\n\
For more info type man fl\n");
}


int fl_req_init_data(void **req,	// -- OUT
		     fl_uint64_t size)	// -- IN
{
	// NOTE: Arguments


	// OUT: Check for request
	if (req == NULL)
	{
		return 0;	// ERROR: No request
	}


	// IN: Check for zero size
	if (size == 0ULL)
	{
		return -1;	// ERROR: Size is zero
	}


	// NOTE: Processing


	// NOTE: Page alignment
	//
	//  The request data buffer is used for transferring data
	//  from user space to kernel space.
	//
	//  Alignment of the data buffer has direct impact on the
	//  performance of this transfer. If aligned, the data is
	//  transferred faster. Non-aligned buffers require the kernel
	//  code to compensate for, using extra allocations and code
	//  logic.


	// Allocate page aligned data buffer
	*req = (void *) valloc(size);
	if (*req == NULL)
	{
		return -2;	// ERROR: Couldn't allocate page aligned data buffer
	}


	return 1;	// SUCCESS
}


int fl_ioctl(fl_int32_t fd,		// -- IN
	     fl_uint32_t cmd,		// -- IN
	     char *data,		// -- IN
	     fl_int32_t *ret_code)	// -- OUT
{
	fl_int32_t ret;


	// NOTE: Arguments


	// IN: Check for valid file descriptor
	if (fd < 0)
	{
		return 0;	// ERROR: Invalid file descriptor
	}


	// IN: Check for valid command
	if ((cmd != FL_CMD_ATTACH) &&
	    (cmd != FL_CMD_DETACH) &&
	    (cmd != FL_CMD_LIST))
	{
		return -1;	// ERROR: Invalid command
	}


	// IN: Check for data
	if (data == NULL)
	{
		return -2;	// ERROR: No data
	}


	// OUT: Check for return code
	if (ret_code == NULL)
	{
		return -3;	// ERROR: No return code
	}


	// NOTE: Processing


	// NOTE: I/O control return codes
	//
	//  For an efficient code path with just a single I/O control
	//  call, the return code must be encoded upon return.
	//
	//  This implies that all I/O control invocations of commands
	//  return an error value, with the actual encoded return code
	//  stored in errno.


	// Send I/O control command
	ret = ioctl(fd,
		    cmd,
		    data);
	if (ret != -1)
	{
		// NOTE: I/O control return value
		//
		//  On some OSs (IRIX for example), the return value for
		//  I/O control commands are settable by the driver writer
		//  and not necessarily imposed by the kernel.

		
		// NOTE: Unexpected return value
		//
		//  I/O control returned a value different from -1.


		return -4;	// ERROR: Unexpected return value
	}


	// NOTE: I/O control success


	FL_DEBUG_PRINTF(("[ufl][main] INFO: ioctl() errno = %i strerror = \"%s\"\n",
			 errno,
			 strerror(errno)))


	// NOTE: Decode return code


	if (errno <= 256)
	{
		// NOTE: Positive return code


		// Set return code
		*ret_code = errno;


		return 1;	// SUCCESS: Positive return code
	}
	else if (errno <= 512)
	{
		// NOTE: Negative return code
		//
		//  Never reached for list command in this version.


		// Set return code
		*ret_code = -1 * (errno - 257);


		return 2;	// SUCCESS: Negative return code
	}
	else
	{
		// NOTE: Unexpected return code
		//
		//  Complement cases.


		return -5;	// ERROR: Unexpected return code
	}
}


%FUNC:main:START%


int main(int argc,	// -- IN
	 char *argv[])	// -- IN
{
	fl_int32_t fd;
	fl_int32_t ret;
	char *cp;
	char dev_name[NAME_MAX + 1];
	fl_uint32_t comp_length;
	fl_uint64_t file_length;


	//FL_TEST_RET;


#if FL_OS_NAME_V_OSF1
#       if FL_OS_VERSION_V_5_1
#               if FL_OS_RELEASE_V_B


#               endif
#       endif
#endif


	// NOTE: Initialization


	// Request data buffer
	if (fl_req_init_data((void **) &packet.data_req,
			     _PAGESZ) < 1)
	{
                printf("INFO: Not enough memory available\n");


		FL_TEST("main",\
			"NOT_DATA_REQ_ALLOC")
                return %NOT_DATA_REQ_ALLOC[in,0,]%;      // INFO: Not enough memory available
	}


	// NOTE: Processing


        FL_TEST_INIT()


	// Check for root user
	FL_TEST_F_main_N_NOT_SUSER()
	if (FL_TEST_COND((fl_test_ret == 0), || )
	    (getuid() != 0))
	{
		printf("INFO: You must be superuser to use file looper\n");


		FL_TEST("main",\
			"NOT_SUSER")
                return %NOT_SUSER[in,,]%;      // INFO: You must be superuser to use file looper
	}


	// Process arguments
	ret = fl_args_process(argc,
			      argv);
	if (ret < 1)
	{
		// Print help
		fl_syntax_print();


		switch (ret)
		{
			case 0:

				FL_TEST("main",\
					"ARGS_NO_ARGUMENTS")
                                return %ARGS_NO_ARGUMENTS[in,,]%;

			case -1:

				FL_TEST("main",\
					"ARGS_UNKNOWN_COMMAND")
                                return %ARGS_UNKNOWN_COMMAND[in,,]%;

			case -2:

				FL_TEST("main",\
					"ARGS_TOO_MANY_ARGUMENTS")
                                return %TOO_MANY_ARGUMENTS[in,,]%;

			case -3:

				FL_TEST("main",\
					"ARGS_TOO_FEW_ARGUMENTS")
                                return %TOO_FEW_ARGUMENTS[in,,]%;

			default:

				// NOTE: Never reached


				break;
		}
	}


	// Check for help command
	if (args.cmd == FL_CMD_HELP)
	{
		fl_syntax_print();


		FL_TEST("main",\
			"HELP_PRINT")
		return %HELP_PRINT[ip,6,]%;
	}


        // Check for version command
        if (args.cmd == FL_CMD_VERSION)
        {
                fl_version_print();


		FL_TEST("main",\
			"VERSION_PRINT")
		return %VERSION_PRINT[ip,7,]%;
	}


	// Check for attach command
	if (args.cmd == FL_CMD_ATTACH)
	{
                // NOTE: PATH_MAX and NULL byte
                //
                //  POSIX specification is not clear on PATH_MAX including
                //  terminating NULL byte or not.
                //
                //  Behaviour of stat(), open(), creat() and realpath() on Tru64 V5.1B-3 is
		//  consistent with this code: PATH_MAX is the number of bytes in a pathname
		//  including the terminating NULL byte.
		//
		//  Note, however, the discrepancy with some of its documentation. For example,
		//  man realpath stating that:
		//
		//    [ENAMETOOLONG]
                //              The file_name argument is longer than PATH_MAX or a pathname component
                //              is longer than NAME_MAX.
		//
		//  Actual behaviour of realpath() allows a maximum of one byte less: PATH_MAX - 1.
		//
		//  Any attempt to resolve paths of PATH_MAX or more number of bytes, result in
		//  ENAMETOOLONG errors.
		//
                //
		//  Besides that, all pathname handling functions mentioned above behave the same
		//  with regards to PATH_MAX. And so does this user command.
		//
		//  For convenience, two macros are defined:
		//
		//  PATH_LENGTH_MAX : Maximum number of bytes in a pathname string (not including NULL byte)
		//  PATH_SIZE_MAX : Maximum number of bytes in a pathname string/buffer (including NULL byte)


		// Get length of device name
		file_length = strlen(args.dev);


		// Check minimum length
		if (file_length < 11)
		{
			fprintf(stderr,
				"ERROR: Device pathname too short\n");


			FL_TEST("main",\
				"DEV_PATHNAME_TOO_SHORT")
                        return %DEV_PATHNAME_TOO_SHORT[in,,]%;      // ERROR: Device pathname too short
		}


		// Check maximum length
		if (file_length > PATH_LENGTH_MAX)
		{
			fprintf(stderr,
				"ERROR: Device pathname too long, only %u bytes are allowed\n",
				PATH_LENGTH_MAX);


			FL_TEST("main",\
				"DEV_PATHNAME_TOO_LONG_NO_CANON")
                        return %DEV_PATHNAME_TOO_LONG_NO_CANON[in,,]%;    // ERROR: Device pathname too long, only 1022 bytes are allowed
		}


		// NOTE: Device pathname
		//
		//  For the user command, only device pathnames starting with /dev/loop/
		//  are allowed to be specified on the command line. Note though, /dev may
		//  be a symbolic link, which is handled below.


		// Check for /dev/loop/
		if ((args.dev[0] != '/') ||
		    (args.dev[1] != 'd') ||
		    (args.dev[2] != 'e') ||
		    (args.dev[3] != 'v') ||
		    (args.dev[4] != '/') ||
		    (args.dev[5] != 'l') ||
		    (args.dev[6] != 'o') ||
		    (args.dev[7] != 'o') ||
		    (args.dev[8] != 'p') ||
		    (args.dev[9] != '/'))
		{
			fprintf(stderr,
				"ERROR: Device pathname must begin with /dev/loop/\n");


			FL_TEST("main",\
				"DEV_PATHNAME_DEVLOOP")
                        return %DEV_PATHNAME_DEVLOOP[in,,]%;      // ERROR: Device pathname must begin with /dev/loop/
		}


		// Check for trailing directory separator
		if (args.dev[file_length - 1] == '/')
		{
			fprintf(stderr,
				"ERROR: Device name not specified\n");


			FL_TEST("main",\
				"NO_DEV_NAME")
                        return %NO_DEV_NAME[in,,]%;      // ERROR: Device name not specified
		}


		// Get start of last component
		for (cp = &args.dev[file_length - 2];
		     cp > &args.dev[9];
		     cp--)
		{
			// Check for / (slash)
			if (*cp == '/')
			{
				break;
			}
		}

		cp++;


		// Get length of last component
		comp_length = (fl_uint32_t) (&args.dev[file_length] - cp);


		// Check length of component
		if (comp_length > NAME_MAX)
		{
			fprintf(stderr,
				"ERROR: Device name too long, only %u bytes are allowed\n",
				NAME_MAX);


			FL_TEST("main",\
				"DEV_NAME_TOO_LONG")
                        return %DEV_NAME_TOO_LONG[in,,]%;      // ERROR: Device name too long, only 255 bytes are allowed
		}


		// Copy device name (including preceeding / - slash)
		cp--;
		dev_name[0] = '\0';
		strcat(dev_name,
		       cp);


		// Truncate device name at preceeding / (slash)
		*cp = '\0';


		FL_DEBUG_PRINTF(("[ufl][main] INFO: dev_name = \"%s\"\n",
				 dev_name))


                // Canonicalize device path
                errno = 0;
		dev_real[0] = '\0';
                if (realpath(args.dev,
			     dev_real) == NULL)
                {
			FL_TEST_F_main_N_ATTACH_EACCES_DEV()
			FL_TEST_F_main_N_ATTACH_EIO_DEV()
			FL_TEST_F_main_N_ATTACH_ELOOP_DEV()
			FL_TEST_F_main_N_ATTACH_ENOMEM_DEV()
			FL_TEST_F_main_N_ATTACH_ENOTDIR_DEV()


                        // Switch through errors
                        switch (errno)
                        {
                                case EACCES:

                                        fprintf(stderr,
						"ERROR: Read or search permission was denied for a component of device pathname\n");


					FL_TEST("main",\
						"ATTACH_EACCES_DEV")
                                        return %ATTACH_EACCES_DEV[in,,]%;

                                case EIO:

                                        fprintf(stderr,
						"ERROR: An error occurred while reading from the file system\n");


					FL_TEST("main",\
						"ATTACH_EIO_DEV")
                                        return %ATTACH_EIO_DEV[in,,]%;

                                case ELOOP:

                                        fprintf(stderr,
						"ERROR: Too many symbolic links were encountered in resolving path\n");


					FL_TEST("main",\
						"ATTACH_ELOOP_DEV")
                                        return %ATTACH_ELOOP_DEV[in,,]%;

                                case ENAMETOOLONG:

                                        fprintf(stderr,
						"ERROR: A pathname component is longer than %u bytes\n",
						NAME_MAX);


					FL_TEST("main",\
						"ATTACH_ENAMETOOLONG_DEV")
                                        return %ATTACH_ENAMETOOLONG_DEV[in,,]%;

                                case ENOENT:

                                        fprintf(stderr,
						"ERROR: A component of device pathname does not name an existing file or file name points to an empty string\n");


					FL_TEST("main",\
						"ATTACH_ENOENT_DEV")
                                        return %ATTACH_ENOENT_DEV[in,,]%;

                                case ENOMEM:

                                        fprintf(stderr,
						"ERROR: Insufficient storage space available\n");


					FL_TEST("main",\
						"ATTACH_ENOMEM_DEV")
                                        return %ATTACH_ENOMEM_DEV[in,,]%;

                                case ENOTDIR:

                                        fprintf(stderr,
						"ERROR: A component of either path prefix is not a directory\n");


					FL_TEST("main",\
						"ATTACH_ENOTDIR_DEV")
                                        return %ATTACH_ENOTDIR_DEV[in,,]%;

                                default:

                                        break;
                        }
                }


                // NOTE: Recheck device pathname length
                //
                //  Since realpath() may have expanded the pathname because of symbolic links, for example,
		//  and the fact the device pathname does not include the device name, a second check must
                //  be done at this point to ensure the full device pathname is up to PATH_LENGTH_MAX in length.


#ifdef FL_COMP_BITS_V_32


		FL_DEBUG_PRINTF(("[ufl][main] INFO: realpath(args.dev=\"%s\" (%lu), dev_real=\"%s\" (%lu))\n",
				 args.dev,
				 strlen(args.dev),
				 dev_real,
				 strlen(dev_real)))


#elif FL_COMP_BITS_V_64


		FL_DEBUG_PRINTF(("[ufl][main] INFO: realpath(args.dev=\"%s\" (%llu), dev_real=\"%s\" (%llu))\n",
				 args.dev,
				 strlen(args.dev),
				 dev_real,
				 strlen(dev_real)))


#endif


                // Get length of device pathname, without device name
                file_length = strlen(dev_real);


                // Check final length of device pathname (device pathname length + / + device name length)
                if ((file_length + 1 + comp_length) > PATH_LENGTH_MAX)
                {
			fprintf(stderr,
				"ERROR: Device pathname too long, only %u bytes are allowed\n",
				PATH_LENGTH_MAX);


			FL_TEST("main",\
				"DEV_PATHNAME_TOO_LONG_CANON")
                        return %DEV_PATHNAME_TOO_LONG_CANON[in,,]%;      // ERROR: Device pathname too long, only 1022 bytes are allowed
                }


		FL_DEBUG_PRINTF(("[ufl][main] INFO: dev_real = \"%s\"\n",
				 dev_real))


		FL_DEBUG_PRINTF(("[ufl][main] INFO: dev_name = \"%s\"\n",
				 dev_name))


		// Add device name to device pathname
		strcat(dev_real,
		       dev_name);


		FL_DEBUG_PRINTF(("[ufl][main] INFO: dev_real = \"%s\"\n",
				 dev_real))


		// NOTE: Device pathname is valid


                // Update device
                args.dev = dev_real;


		FL_DEBUG_PRINTF(("[ufl][main] INFO: args.dev = \"%s\" (%llu bytes)\n",
				 args.dev,
				 file_length + 1 + comp_length))


                // Canonicalize file pathname
                errno = 0;
		file_real[0] = '\0';
                if (realpath(args.file,
			     file_real) == NULL)
                {
			FL_TEST_F_main_N_ATTACH_EACCES_FILE()
                        FL_TEST_F_main_N_ATTACH_EIO_FILE()
                        FL_TEST_F_main_N_ATTACH_ELOOP_FILE()
                        FL_TEST_F_main_N_ATTACH_ENOMEM_FILE()
                        FL_TEST_F_main_N_ATTACH_ENOTDIR_FILE()


                        // Switch through errors
                        switch (errno)
                        {
                                case EACCES:

                                        fprintf(stderr,
						"ERROR: Read or search permission was denied for a component of file pathname\n");


					FL_TEST("main",\
						"ATTACH_EACCES_FILE")
                                        return %ATTACH_EACCES_FILE[in,,]%;

                                case EIO:

                                        fprintf(stderr,
						"ERROR: An error occurred while reading from the file system\n");


					FL_TEST("main",\
						"ATTACH_EIO_FILE")
                                        return %ATTACH_EIO_FILE[in,,]%;

                                case ELOOP:

                                        fprintf(stderr,
						"ERROR: Too many symbolic links were encountered in resolving path\n");


					FL_TEST("main",\
						"ATTACH_ELOOP_FILE")
                                        return %ATTACH_ELOOP_FILE[in,,]%;

                                case ENAMETOOLONG:

                                        // Check PATH_MAX
                                        if (strlen(args.file) > PATH_LENGTH_MAX)
                                        {
						fprintf(stderr,
							"ERROR: File pathname too long, only %u bytes are allowed\n",
							PATH_LENGTH_MAX);
                                        }
                                        else
                                        {
                                                fprintf(stderr,
							"ERROR: A pathname component is longer than %u bytes\n",
							NAME_MAX);
                                        }


					FL_TEST("main",\
						"ATTACH_ENAMETOOLONG_FILE")
                                        return %ATTACH_ENAMETOOLONG_FILE[in,,]%;

                                case ENOENT:

                                        fprintf(stderr,
						"ERROR: A component of file pathname does not name an existing file or file pathname points to an empty string\n");


					FL_TEST("main",\
						"ATTACH_ENOENT_FILE")
                                        return %ATTACH_ENOENT_FILE[in,,]%;

                                case ENOMEM:

                                        fprintf(stderr,
						"ERROR: Insufficient storage space available\n");


					FL_TEST("main",\
						"ATTACH_ENOMEM_FILE")
                                        return %ATTACH_ENOMEM_FILE[in,,]%;

                                case ENOTDIR:

                                        fprintf(stderr,
						"ERROR: A component of either path prefix is not a directory\n");


					FL_TEST("main",\
						"ATTACH_ENOTDIR_FILE")
                                        return %ATTACH_ENOTDIR_FILE[in,,]%;

                                default:

                                        break;
			}
		}


		// NOTE: File pathname is valid


		// Update file
		args.file = file_real;


		FL_DEBUG_PRINTF(("[ufl][main] INFO: args.file = \"%s\" (%llu bytes)\n",
				 args.file,
				 file_length))
	}


	// Check for detach command
        if (args.cmd == FL_CMD_DETACH)
        {
		// Canonicalize device path
                errno = 0;
                if (realpath(args.dev,
			     dev_real) == NULL)
                {
			FL_TEST_F_main_N_DETACH_EACCES()
                        FL_TEST_F_main_N_DETACH_EIO()
                        FL_TEST_F_main_N_DETACH_ELOOP()
                        FL_TEST_F_main_N_DETACH_ENOMEM()
                        FL_TEST_F_main_N_DETACH_ENOTDIR()


                        // Switch through errors
                        switch (errno)
                        {
                                case EACCES:

                                        fprintf(stderr,
						"ERROR: Read or search permission was denied for a component of device pathname\n");


					FL_TEST("main",\
						"DETACH_EACCES")
                                        return %DETACH_EACCES[in,,]%;

                                case EIO:

                                        fprintf(stderr,
						"ERROR: An error occurred while reading from the file system\n");


					FL_TEST("main",\
						"DETACH_EIO")
                                        return %DETACH_EIO[in,,]%;

                                case ELOOP:

                                        fprintf(stderr,
						"ERROR: Too many symbolic links were encountered in resolving path\n");


					FL_TEST("main",\
						"DETACH_ELOOP")
                                        return %DETACH_ELOOP[in,,]%;

                                case ENAMETOOLONG:

                                        // Check PATH_MAX
                                        if (strlen(args.dev) > PATH_LENGTH_MAX)
                                        {
                                                fprintf(stderr,
							"ERROR: The device pathname is longer than %u\n",
							PATH_LENGTH_MAX);
                                        }
                                        else
                                        {
                                                fprintf(stderr,
							"ERROR: A pathname component is longer than %u bytes\n",
							NAME_MAX);
                                        }


					FL_TEST("main",\
						"DETACH_ENAMETOOLONG")
                                        return %DETACH_ENAMETOOLONG[in,,]%;

                                case ENOENT:

                                        fprintf(stderr,
						"ERROR: A component of device pathname does not name an existing file or file name points to an empty string\n");


					FL_TEST("main",\
						"DETACH_ENOENT")
                                        return %DETACH_ENOENT[in,,]%;

                                case ENOMEM:

                                        fprintf(stderr,
						"ERROR: Insufficient storage space available\n");


					FL_TEST("main",\
						"DETACH_ENOMEM")
                                        return %DETACH_ENOMEM[in,,]%;

                                case ENOTDIR:

                                        fprintf(stderr,
						"ERROR: A component of either path prefix is not a directory\n");


					FL_TEST("main",\
						"DETACH_ENOTDIR")
                                        return %DETACH_ENOTDIR[in,,]%;

                                default:

                                        break;
                        }
                }


                FL_DEBUG_PRINTF(("[ufl][main] INFO: realpath(args.dev=\"%s\", dev_real=\"%s\")\n",
				 args.dev,
				 dev_real))


		// NOTE: Device pathname is valid


                // Update device
                args.dev = dev_real;


                FL_DEBUG_PRINTF(("[ufl][main] INFO: args.dev = \"%s\" (%llu bytes)\n",
				 args.dev,
				 file_length + 1 + comp_length))
	}


#ifdef FL_COMP_BITS_V_32


	// Print pid and ppid
	FL_DEBUG_PRINTF(("[ufl][main] pid = %i ppid = %i\n",
			 getpid(),
			 getppid()))


#elif FL_COMP_BITS_V_64


	// Print pid and ppid
	FL_DEBUG_PRINTF(("[ufl][main] pid = %llu ppid = %llu\n",
			 getpid(),
			 getppid()))


#endif


#if FL_OS_NAME_V_OSF1
#       if FL_OS_VERSION_V_5_1
#               if FL_OS_RELEASE_V_B


#               endif
#       endif
#endif


main_open:


	FL_DEBUG_PRINTF(("[ufl][main] INFO: open(/dev/fl) (pid = %i ppid = %i)\n",
			 getpid(),
			 getppid()))


	// Open /dev/fl
	errno = 0;
	fd = open("/dev/fl",
		  O_RDWR,
		  0);
	if (fd == -1)
	{
		FL_DEBUG_PRINTF(("[ufl][main] INFO: open(/dev/fl) == -1 (pid = %i ppid = %i)\n",
				 getpid(),
				 getppid()))


		// Check for EUSERS
		if (errno == EUSERS)
		{
			fprintf(stderr,
				"ERROR: User must be root\n");


			FL_TEST("main",\
				"NOT_SUSER_OPEN_DEVFL")
                        return %NOT_SUSER_OPEN_DEVFL[in,,]%;     // ERROR: User must be root
		}


		// Check for EACCES
                if (errno == EACCES)
                {
                        fprintf(stderr,
				"ERROR: No permission to use file looper, license needed\n");


			FL_TEST("main",\
				"NO_PERMISSION_CPU_CHECK")
                        return %NO_PERMISSION_CPU_CHECK[in,,]%;      // ERROR: No permission to use file looper, license needed
                }


		// Check for EOVERFLOW
		if (errno == EOVERFLOW)
		{
                        fprintf(stderr,
				"ERROR: Please reload file looper's kernel module and try again\n");


			FL_TEST("main",\
				"SEM_OVER_UNDER")
                        return %SEM_OVER_UNDER[in,,]%;      // ERROR: Please reload file looper's kernel module and try again
		}


#if FL_OS_NAME_V_OSF1
#       if FL_OS_VERSION_V_5_1
#               if FL_OS_RELEASE_V_B


#		endif
#	endif
#elif FL_OS_NAME_V_IRIX
#       if FL_OS_VERSION_V_6_5
#               if FL_OS_RELEASE_V_30


		// NOTE: Registered modules
		//
		//  In IRIX, once a module is registered, any access
		//  to an associated hardware graph device file will
		//  automatically load the module into memory.
		//
		//  No further user space program intervention is
		//  necessary.


#               endif
#       endif
#endif


		fprintf(stderr,
			"ERROR: Couldn't open /dev/fl\n");


		FL_TEST("main",\
			"NOT_OPEN_DEVFL")
                return %NOT_OPEN_DEVFL[in,,]%;     // ERROR: Couldn't open /dev/fl
	}


	FL_DEBUG_PRINTF(("[ufl][main] INFO: open(/dev/fl) != -1 (pid = %i ppid = %i)\n",
			 getpid(),
			 getppid()))


	// Create new request packet
	fl_packet_new_req();


	FL_TEST_F_main_N_FILE_NOT_FOUND()
        FL_TEST_F_main_N_WRONG_DIR_PREFIX()
        FL_TEST_F_main_N_EMPTY_BLK_DEV_NAME()
        FL_TEST_F_main_N_BLK_DEV_PATHNAME_HAS_INVALID_CHAR()
        FL_TEST_F_main_N_DEV_NAME_NOT_FOUND()


	// NOTE: I/O control command


	switch (args.cmd)
	{
		case FL_CMD_ATTACH:
		case FL_CMD_DETACH:

			// Set packet data
			packet.data = packet.data_req;


			FL_DEBUG_PRINTF(("[ufl][main] INFO: START ioctl(%s)\n",
					 args.cmd == FL_CMD_ATTACH ? "FL_CMD_ATTACH" : "FL_CMD_DETACH"))


			switch (fl_ioctl(fd,
					 args.cmd,
					 packet.data,
					 &packet.ret))
			{
				case 1:

					// NOTE: Positive return code
					//
					//  Handled in command postprocessing.


					break;

				case 2:

					// NOTE: Negative return code
					//
					//  Handled in command postprocessing.


					break;

				case -4:

					// NOTE: Unexpected return value


					FL_DEBUG_PRINTF(("[ufl][main] ERROR: ioctl() errno = %i strerror = \"%s\"\n",
							 errno,
							 strerror(errno)))


					if (args.cmd == FL_CMD_ATTACH)
					{
						FL_DEBUG_PRINTF(("[ufl][main] INFO: FL_CMD_ATTACH I/O control error\n"))


						fprintf(stderr,
							"ERROR: Couldn't execute attach command\n");


						return %UNEXPECTED_IOCTL_ATTACH_RETURN_VALUE[in,,]%;
					}
					else
					{
						FL_DEBUG_PRINTF(("[ufl][main] INFO: FL_CMD_DETACH I/O control error\n"))


						fprintf(stderr,
							"ERROR: Couldn't execute detach command\n");


						return %UNEXPECTED_IOCTL_DETACH_RETURN_VALUE[in,,]%;
					}

				case -5:

					// NOTE: Unexpected return code


					FL_DEBUG_PRINTF(("[ufl][main] ERROR: ioctl() errno = %i strerror = \"%s\"\n",
							 errno,
							 strerror(errno)))


					if (args.cmd == FL_CMD_ATTACH)
					{
						FL_DEBUG_PRINTF(("[ufl][main] INFO: FL_CMD_ATTACH I/O control error\n"))


						fprintf(stderr,
							"ERROR: Couldn't execute attach command\n");


						return %UNEXPECTED_IOCTL_ATTACH_RETURN_CODE[in,,]%;
					}
					else
					{
						FL_DEBUG_PRINTF(("[ufl][main] INFO: FL_CMD_DETACH I/O control error\n"))


						fprintf(stderr,
							"ERROR: Couldn't execute detach command\n");


						return %UNEXPECTED_IOCTL_DETACH_RETURN_CODE[in,,]%;
					}
			}


			break;

		case FL_CMD_LIST:

			// Set packet data
			packet.data = packet.data_req;


			FL_DEBUG_PRINTF(("[ufl][main] INFO: START ioctl(%s)\n",
					 "FL_CMD_LIST"))


			switch (fl_ioctl(fd,
					 args.cmd,
					 packet.data,
					 &packet.ret))
			{
				case 1:

					// NOTE: Positive return code
					//
					//  Handled in command postprocessing.


					break;

				case 2:

					// NOTE: Negative return code
					//
					//  Handled in command postprocessing.
					//
					//  The list command currently has no negative
					//  return codes.


					break;

				case -4:

					// NOTE: Unexpected return value


					FL_DEBUG_PRINTF(("[ufl][main] ERROR: ioctl() errno = %i strerror = \"%s\"\n",
							 errno,
							 strerror(errno)))


					FL_DEBUG_PRINTF(("[ufl][main] INFO: FL_CMD_LIST I/O control error\n"))


					fprintf(stderr,
						"ERROR: Couldn't execute list command\n");


					return %UNEXPECTED_IOCTL_LIST_RETURN_VALUE[in,,]%;

				case -5:

					// NOTE: Unexpected return code


					FL_DEBUG_PRINTF(("[ufl][main] INFO: ioctl() return value = %i\n",
							 errno))


					FL_DEBUG_PRINTF(("[ufl][main] INFO: FL_CMD_LIST I/O control error\n"))


					fprintf(stderr,
						"ERROR: Couldn't execute list command\n");


					return %UNEXPECTED_IOCTL_LIST_RETURN_CODE[in,,]%;
			}
	}


	FL_DEBUG_PRINTF(("[ufl][main] INFO: I/O control packet.ret = %i\n",
			 packet.ret))


	// Check for encoded EACCES
	if (packet.ret == %fl_c:*:CHECK_IOCTL_CPU_EACCESS[on,,]%)
	{
		fprintf(stderr,
			"ERROR: No permission to use file looper, license needed\n");


		FL_TEST("main",\
			"NO_PERMISSION_WRITE_REQUEST_PACKET")
		return %NO_PERMISSION_WRITE_REQUEST_PACKET[in,,]%;     // ERROR: No permission to use file looper, license needed
	}


	FL_DEBUG_PRINTF(("[ufl][main] Return code = %i\n",
			 packet.ret))


	// NOTE: Command postprocessing
	//
	//  Packet return codes come directly from the file looper kernel
	//  module and need to be mapped to user friendly return codes as
	//  they indicate kernel module internal errors that are not
	//  visible to the user.


	// Check for return code (FL_CMD_ATTACH)
	if (args.cmd == FL_CMD_ATTACH)
	{
		FL_DEBUG_PRINTF(("[main] INFO: FL_CMD_ATTACH packet.ret = %i\n",
				 packet.ret))


		switch (packet.ret)
		{
			case %fl_c:*:ATTACH_SUCCESS[op,,]%:

				//  %fl_c:*:ATTACH_SUCCESS[po,,]% - fl_loop_add:SUCCESS
				FL_TEST("main",\
					"ATTACH_SUCCESS")
				ret = %ATTACH_SUCCESS[ip,1,]%;


				break;


#if FL_OS_NAME_V_OSF1
#       if FL_OS_VERSION_V_5_1
#               if FL_OS_RELEASE_V_B


#               endif
#       endif
#endif


			case %fl_c:*:BLK_DEV_ALREADY_ATTACHED[on,,]%:

				//  %fl_c:*:BLK_DEV_ALREADY_ATTACHED[on,,]% - fl_loop_new:ERROR: BLK device is already attached
				fprintf(stderr,
					"ERROR: Block device is already attached\n");


				FL_TEST("main",\
					"BLK_DEV_ALREADY_ATTACHED")
                                ret = %BLK_DEV_ALREADY_ATTACHED[in,,]%;


                                break;

			case %fl_c:*:NOT_LOOKUP_FILENAME[on,,]%:

				//  %fl_c:*:NOT_LOOKUP_FILENAME[on,,]% - fl_loop_new:ERROR: Couldn't lookup file name
				fprintf(stderr,
					"ERROR: Couldn't lookup file name\n");


				FL_TEST("main",\
					"NOT_LOOKUP_FILENAME")
                                ret = %NOT_LOOKUP_FILENAME[in,,]%;


                                break;

			case %fl_c:*:FILE_NOT_FOUND[on,,]%:

				//  %fl_c:*:FILE_NOT_FOUND[on,,]% - fl_loop_new:ERROR: Couldn't find file
				fprintf(stderr,
					"ERROR: Couldn't find requested file\n");


				FL_TEST("main",\
					"FILE_NOT_FOUND")
                                ret = %FILE_NOT_FOUND[in,,]%;


                                break;

			case %fl_c:*:NOT_FILE_STAT[on,,]%:

				// %fl_c:*:NOT_FILE_STAT[on,,]% - fl_loop_new:ERROR: Couldn't stat file
				fprintf(stderr,
					"ERROR: Couldn't stat file\n");


				FL_TEST("main",\
					"NOT_FILE_STAT")
                                ret = %NOT_FILE_STAT[in,,]%;


                                break;

			case %fl_c:*:FILE_NEGATIVE_SIZE[on,,]%:

				// %fl_c:*:FILE_NEGATIVE_SIZE[on,,]% - fl_loop_new:ERROR: File has negative size
				fprintf(stderr,
					"ERROR: File has negative size\n");


				FL_TEST("main",\
					"FILE_NEGATIVE_SIZE")
                                ret = %FILE_NEGATIVE_SIZE[in,,]%;


                                break;

			case %fl_c:*:WRONG_DIR_PREFIX[on,,]%:

				// %fl_c:*:WRONG_DIR_PREFIX[on,,]% - fl_dev_check_path:ERROR: Wrong directory prefix
				fprintf(stderr,
					"ERROR: Wrong directory prefix, must be /dev/loop\n");


				FL_TEST("main",\
					"WRONG_DIR_PREFIX")
                                ret = %WRONG_DIR_PREFIX[in,,]%;


                                break;

			case %fl_c:*:EMPTY_BLK_DEV_NAME[on,,]%:

                                // %fl_c:*:EMPTY_BLK_DEV_NAME[on,,]% - fl_dev_check_path:ERROR: Empty BLK device name
                                fprintf(stderr,
					"ERROR: Empty block device name\n");


				FL_TEST("main",\
					"EMPTY_BLK_DEV_NAME")
                                ret = %EMPTY_BLK_DEV_NAME[in,,]%;


                                break;


#if FL_OS_NAME_V_IRIX
#       if FL_OS_VERSION_V_6_5
#               if FL_OS_RELEASE_V_30


			case %fl_c:*:NOT_START_R_LETTER_BLK_DEV_NAME[on,,]%:

                                fprintf(stderr,
					"ERROR: Block device names may not start with the letter 'r'\n");


				FL_TEST("main",\
					"NOT_START_R_LETTER_DEV_NAME")
                                ret = %NOT_START_R_LETTER_BLK_DEV_NAME[in,,]%;


                                break;

			case %fl_c:*:DEVICE_NAME_OVERFLOW[on,,]%:

                                fprintf(stderr,
					"ERROR: Block device names may have a maximum length of %u\n",
					PATH_LENGTH_MAX);


				FL_TEST("main",
					"DEVICE_NAME_OVERFLOW")
                                ret = %DEVICE_NAME_OVERFLOW[in,,]%;


                                break;


#               endif
#       endif
#endif


			case %fl_c:*:BLK_DEV_PATHNAME_HAS_INVALID_CHAR[on,,]%:

                                // %fl_c:*:BLK_DEV_PATHNAME_HAS_INVALID_CHAR[on,,]% - fl_dev_check_path:ERROR: Invalid characters in BLK device pathname
                                fprintf(stderr,
					"ERROR: Invalid characters in block device pathname. Allowed are: a-z, A-Z, 0-9, . (dot), - (minus) and _ (underscore)\n");


				FL_TEST("main",\
					"BLK_DEV_PATHNAME_HAS_INVALID_CHAR")
                                ret = %BLK_DEV_PATHNAME_HAS_INVALID_CHAR[in,,]%;


                                break;

			case %fl_c:*:BLK_DEV_PATHNAME_HAS_INVALID_FIRST_CHAR[on,,]%:

                                // %fl_c:*:BLK_DEV_PATHNAME_HAS_INVALID_FIRST_CHAR[on,,]% - fl_dev_check_path:ERROR: Invalid characters in BLK device pathname
                                fprintf(stderr,
					"ERROR: Invalid first character in block device pathname. Allowed are: a-z, A-Z, . (dot), - (minus) and _ (underscore)\n");


				FL_TEST("main",\
					"BLK_DEV_PATHNAME_HAS_INVALID_FIRST_CHAR")
                                ret = %BLK_DEV_PATHNAME_HAS_INVALID_FIRST_CHAR[in,,]%;


                                break;

			case %fl_c:*:NOT_BLK_DEV_LOOKUP[on,,]%:

				// %fl_c:*:NOT_BLK_DEV_LOOKUP[on,,]% - fl_loop_new:ERROR: Couldn't lookup BLK device file
				fprintf(stderr,
					"ERROR: Couldn't lookup block device file\n");


				FL_TEST("main",\
					"NOT_BLK_DEV_LOOKUP")
                                ret = %NOT_BLK_DEV_LOOKUP[in,,]%;


                                break;


#if FL_OS_NAME_V_OSF1
#       if FL_OS_VERSION_V_5_1
#               if FL_OS_RELEASE_V_B


#               endif
#       endif
#endif


			case %fl_c:*:BLK_DEV_NAME_NOT_SYMLINK_DEV_FILE[on,,]%:

				// %fl_c:*:BLK_DEV_NAME_NOT_SYMLINK_DEV_FILE[on,,]% - fl_loop_new:ERROR: Device file name is not a symbolic link to a BLK device file
				fprintf(stderr,
					"ERROR: Device file name is not a symbolic link to a block device file\n");


				FL_TEST("main",\
					"BLK_DEV_NAME_NOT_SYMLINK_DEV_FILE")
                                ret = %BLK_DEV_NAME_NOT_SYMLINK_DEV_FILE[in,,]%;


                                break;


#if FL_OS_NAME_V_OSF1
#       if FL_OS_VERSION_V_5_1
#               if FL_OS_RELEASE_V_B


#               endif
#       endif
#endif


			case %fl_c:*:NOT_BLK_DEV_FILE_DELETE[on,,]%:

				// %fl_c:*:NOT_BLK_DEV_FILE_DELETE[on,,]% - fl_loop_new:ERROR: Couldn't delete device file
				fprintf(stderr,
					"ERROR: Couldn't delete existing block device file with the same name. Delete manually and retry\n");


				FL_TEST("main",\
					"NOT_BLK_DEV_FILE_DELETE")
                                ret = %NOT_BLK_DEV_FILE_DELETE[in,,]%;


                                break;


#if FL_OS_NAME_V_IRIX
#       if FL_OS_VERSION_V_6_5
#               if FL_OS_RELEASE_V_30


			case %fl_c:*:NOT_CHR_DEV_LOOKUP[on,,]%:

				// %fl_c:*:NOT_CHR_DEV_LOOKUP[on,,]% - fl_loop_new:ERROR: Couldn't lookup CHR device file
				fprintf(stderr,
					"ERROR: Couldn't lookup char device file\n");


				FL_TEST("main",\
					"NOT_CHR_DEV_LOOKUP")
                                ret = %NOT_CHR_DEV_LOOKUP[in,,]%;


                                break;

			case %fl_c:*:CHR_DEV_NAME_NOT_SYMLINK_DEV_FILE[on,,]%:

				// %fl_c:*:CHR_DEV_NAME_NOT_SYMLINK_DEV_FILE[on,,]% - fl_loop_new:ERROR: Device file name is not a symbolic link to a CHR device file
				fprintf(stderr,
					"ERROR: Device file name is not a symbolic link to a char device file\n");


				FL_TEST("main",\
					"CHR_DEV_NAME_NOT_SYMLINK_DEV_FILE")
                                ret = %CHR_DEV_NAME_NOT_SYMLINK_DEV_FILE[in,,]%;


                                break;

			case %fl_c:*:NOT_CHR_DEV_FILE_DELETE[on,,]%:

				// %fl_c:*:NOT_CHR_DEV_FILE_DELETE[on,,]% - fl_loop_new:ERROR: Couldn't delete CHR device file
				fprintf(stderr,
					"ERROR: Couldn't delete existing char device file with the same name. Delete manually and retry\n");


				FL_TEST("main",\
					"NOT_CHR_DEV_FILE_DELETE")
                                ret = %NOT_CHR_DEV_FILE_DELETE[in,,]%;


                                break;

			case %fl_c:*:NOT_OPEN_FILE_FOR_VH_CHECK[on,,]%:

				// %fl_c:*:NOT_OPEN_FILE_FOR_VH_CHECK[on,,]% - fl_loop_new:ERROR: Couldn't open attached file for volume header check
				fprintf(stderr,
					"ERROR: Couldn't open attached file for volume header check\n");


				FL_TEST("main",\
					"NOT_OPEN_FILE_FOR_VH_CHECK")
                                ret = %NOT_OPEN_FILE_FOR_VH_CHECK[in,,]%;


                                break;

			case %fl_c:*:NOT_READ_FILE_FOR_VH_CHECK[on,,]%:

				// %fl_c:*:NOT_READ_FILE_FOR_VH_CHECK[on,,]% - fl_loop_new:ERROR: Couldn't read from attached file for volume header check
				fprintf(stderr,
					"ERROR: Couldn't read from attached file for volume header check\n");


				FL_TEST("main",\
					"NOT_READ_FILE_FOR_VH_CHECK")
                                ret = %NOT_READ_FILE_FOR_VH_CHECK[in,,]%;


                                break;

			case %fl_c:*:NOT_CLOSE_FILE_FOR_VH_CHECK[on,,]%:

				// %fl_c:*:NOT_CLOSE_FILE_FOR_VH_CHECK[on,,]% - fl_loop_new:ERROR: Couldn't close attached file in volume header check
				fprintf(stderr,
					"ERROR: Couldn't close attached file in volume header check\n");


				FL_TEST("main",\
					"NOT_CLOSE_FILE_FOR_VH_CHECK")
                                ret = %NOT_CLOSE_FILE_FOR_VH_CHECK[in,,]%;


                                break;

			case %fl_c:*:BITFIELD_IS_FULL_BLK[on,,]%:

				// %fl_c:*:BITFIELD_IS_FULL_BLK[on,,]% - fl_dev_new:ERROR: Device bitfield is full
				fprintf(stderr,
					"ERROR: Couldn't create new device special file, system limit reached\n");


				FL_TEST("main",\
					"BITFIELD_IS_FULL_BLK")
                                ret = %BITFIELD_IS_FULL_BLK[in,,]%;


                                break;

			case %fl_c:*:NOT_CREATE_HWGRAPH_BLK[on,,]%:

				// %fl_c:*:NOT_CREATE_HWGRAPH_BLK[on,,]% - fl_dev_new:ERROR: Couldn't create hardware graph block device
				fprintf(stderr,
					"ERROR: Couldn't create hardware graph block device\n");


				FL_TEST("main",\
					"NOT_CREATE_HWGRAPH_BLK")
                                ret = %NOT_CREATE_HWGRAPH_BLK[in,,]%;


                                break;

			case %fl_c:*:NOT_CREATE_HWGRAPH_CHR_NOT_REMOVE_BLK[on,,]%:

				// %fl_c:*:NOT_CREATE_HWGRAPH_CHR_NOT_REMOVE_BLK[on,,]% - fl_dev_new:ERROR: Couldn't create hardware graph char device, nor remove hardware graph block device
				fprintf(stderr,
					"ERROR: Couldn't create hardware graph char device, nor remove hardware graph block device\n");


				FL_TEST("main",\
					"NOT_CREATE_HWGRAPH_CHR_NOT_REMOVE_BLK")
                                ret = %NOT_CREATE_HWGRAPH_CHR_NOT_REMOVE_BLK[in,,]%;


                                break;

			case %fl_c:*:NOT_CREATE_HWGRAPH_CHR[on,,]%:

				// %fl_c:*:NOT_CREATE_HWGRAPH_CHR[on,,]% - fl_dev_new:ERROR: Couldn't create hardware graph char device
				fprintf(stderr,
					"ERROR: Couldn't create hardware graph char device\n");


				FL_TEST("main",\
					"NOT_CREATE_HWGRAPH_CHR")
                                ret = %NOT_CREATE_HWGRAPH_CHR[in,,]%;


                                break;

			case %fl_c:*:NOT_CREATE_SYMLINK_TO_BLK_NOT_REMOVE_BLK[on,,]%:

				// %fl_c:*:NOT_CREATE_SYMLINK_TO_BLK_NOT_REMOVE_BLK[on,,]% - fl_dev_new:ERROR: Couldn't create symbolic link to BLK device, nor remove BLK device
				fprintf(stderr,
					"ERROR: Couldn't create symbolic link to block device, nor remove block device\n");


				FL_TEST("main",\
					"NOT_CREATE_SYMLINK_TO_BLK_NOT_REMOVE_BLK")
                                ret = %NOT_CREATE_SYMLINK_TO_BLK_NOT_REMOVE_BLK[in,,]%;


                                break;

			case %fl_c:*:NOT_CREATE_SYMLINK_TO_BLK_NOT_REMOVE_CHR[on,,]%:

				// %fl_c:*:NOT_CREATE_SYMLINK_TO_BLK_NOT_REMOVE_CHR[on,,]% - fl_dev_new:ERROR: Couldn't create symbolic link to BLK device, nor remove CHR device
				fprintf(stderr,
					"ERROR: Couldn't create symbolic link to block device, nor remove char device\n");


				FL_TEST("main",\
					"NOT_CREATE_SYMLINK_TO_BLK_NOT_REMOVE_CHR")
                                ret = %NOT_CREATE_SYMLINK_TO_BLK_NOT_REMOVE_CHR[in,,]%;


                                break;

			case %fl_c:*:NOT_CREATE_SYMLINK_TO_BLK_NOT_REMOVE_BLK_CHR[on,,]%:

				// %fl_c:*:NOT_CREATE_SYMLINK_TO_BLK_NOT_REMOVE_BLK_CHR[on,,]% - fl_dev_new:ERROR: Couldn't create symbolic link to BLK device, nor remove CHR and BLK devices
				fprintf(stderr,
					"ERROR: Couldn't create symbolic link to block device, nor remove char and block devices\n");


				FL_TEST("main",\
					"NOT_CREATE_SYMLINK_TO_BLK_NOT_REMOVE_BLK_CHR")
                                ret = %NOT_CREATE_SYMLINK_TO_BLK_NOT_REMOVE_BLK_CHR[in,,]%;


                                break;

			case %fl_c:*:NOT_CREATE_SYMLINK_BLK[on,,]%:

				// %fl_c:*:NOT_CREATE_SYMLINK_BLK[on,,]% - fl_dev_new:ERROR: Couldn't create symbolic link to BLK device
				fprintf(stderr,
					"ERROR: Couldn't create symbolic link to block device\n");


				FL_TEST("main",\
					"NOT_CREATE_SYMLINK_BLK")
                                ret = %NOT_CREATE_SYMLINK_BLK[in,,]%;


                                break;

			case %fl_c:*:NOT_CREATE_SYMLINK_TO_CHR_NOT_REMOVE_SYMLINK_BLK[on,,]%:

				// %fl_c:*:NOT_CREATE_SYMLINK_TO_CHR_NOT_REMOVE_SYMLINK_BLK[on,,]% - fl_dev_new:ERROR: Couldn't create symbolic link to CHR device, nor remove symbolic link and BLK device
				fprintf(stderr,
					"ERROR: Couldn't create symbolic link to char device, nor remove symbolic link and block device\n");


				FL_TEST("main",\
					"NOT_CREATE_SYMLINK_TO_CHR_NOT_REMOVE_SYMLINK_BLK")
                                ret = %NOT_CREATE_SYMLINK_TO_CHR_NOT_REMOVE_SYMLINK_BLK[in,,]%;


                                break;

			case %fl_c:*:NOT_CREATE_SYMLINK_TO_CHR_NOT_REMOVE_SYMLINK_CHR[on,,]%:

				// %fl_c:*:NOT_CREATE_SYMLINK_TO_CHR_NOT_REMOVE_SYMLINK_CHR[on,,]% - fl_dev_new:ERROR: Couldn't create symbolic link to CHR device, nor remove symbolic link and CHR device
				fprintf(stderr,
					"ERROR: Couldn't create symbolic link to char device, nor remove symbolic link and char device\n");


				FL_TEST("main",\
					"NOT_CREATE_SYMLINK_TO_CHR_NOT_REMOVE_SYMLINK_CHR")
                                ret = %NOT_CREATE_SYMLINK_TO_CHR_NOT_REMOVE_SYMLINK_CHR[in,,]%;


                                break;

			case %fl_c:*:NOT_CREATE_SYMLINK_TO_CHR_NOT_REMOVE_SYMLINK_BLK_CHR[on,,]%:

				// %fl_c:*:NOT_CREATE_SYMLINK_TO_CHR_NOT_REMOVE_SYMLINK_BLK_CHR[on,,]% - fl_dev_new:ERROR: Couldn't create symbolic link to CHR device, nor remove symbolic link, BLK or CHR devices
				fprintf(stderr,
					"ERROR: Couldn't create symbolic link to char device, nor remove symbolic link, block or char devices\n");


				FL_TEST("main",\
					"NOT_CREATE_SYMLINK_TO_CHR_NOT_REMOVE_SYMLINK_BLK_CHR")
                                ret = %NOT_CREATE_SYMLINK_TO_CHR_NOT_REMOVE_SYMLINK_BLK_CHR[in,,]%;


                                break;

			case %fl_c:*:NOT_CREATE_SYMLINK_TO_CHR_NOT_REMOVE_BLK[on,,]%:

				// %fl_c:*:NOT_CREATE_SYMLINK_TO_CHR_NOT_REMOVE_BLK[on,,]% - fl_dev_new:ERROR: Couldn't create symbolic link to CHR device, nor remove BLK device
				fprintf(stderr,
					"ERROR: Couldn't create symbolic link to char device, nor remove block device\n");


				FL_TEST("main",\
					"NOT_CREATE_SYMLINK_TO_CHR_NOT_REMOVE_BLK")
                                ret = %NOT_CREATE_SYMLINK_TO_CHR_NOT_REMOVE_BLK[in,,]%;


                                break;

			case %fl_c:*:NOT_CREATE_SYMLINK_TO_CHR_NOT_REMOVE_CHR[on,,]%:

				// %fl_c:*:NOT_CREATE_SYMLINK_TO_CHR_NOT_REMOVE_CHR[on,,]% - fl_dev_new:ERROR: Couldn't create symbolic link to CHR device, nor remove CHR device
				fprintf(stderr,
					"ERROR: Couldn't create symbolic link to char device, nor remove char device\n");


				FL_TEST("main",\
					"NOT_CREATE_SYMLINK_TO_CHR_NOT_REMOVE_CHR")
                                ret = %NOT_CREATE_SYMLINK_TO_CHR_NOT_REMOVE_CHR[in,,]%;


                                break;

			case %fl_c:*:NOT_CREATE_SYMLINK_TO_CHR_NOT_REMOVE_BLK_CHR[on,,]%:

				// %fl_c:*:NOT_CREATE_SYMLINK_TO_CHR_NOT_REMOVE_BLK_CHR[on,,]% - fl_dev_new:ERROR: Couldn't create symbolic link to CHR device, nor remove BLK and CHR devices
				fprintf(stderr,
					"ERROR: Couldn't create symbolic link to char device, nor remove block and char devices\n");


				FL_TEST("main",\
					"NOT_CREATE_SYMLINK_TO_CHR_NOT_REMOVE_BLK_CHR")
                                ret = %NOT_CREATE_SYMLINK_TO_CHR_NOT_REMOVE_BLK_CHR[in,,]%;


                                break;

			case %fl_c:*:NOT_CREATE_SYMLINK_CHR[on,,]%:

				// %fl_c:*:NOT_CREATE_SYMLINK_CHR[on,,]% - fl_dev_new:ERROR: Couldn't create symbolic link to CHR device
				fprintf(stderr,
					"ERROR: Couldn't create symbolic link to char device\n");


				FL_TEST("main",\
					"NOT_CREATE_SYMLINK_CHR")
                                ret = %NOT_CREATE_SYMLINK_CHR[in,,]%;


                                break;

#               endif
#       endif
#endif


			default:


				// NOTE: Unknown return code


				FL_DEBUG_PRINTF(("[ufl][main] ERROR: ioctl() errno = %i strerror = \"%s\"\n",
						 errno,
						 strerror(errno)))


				FL_DEBUG_PRINTF(("[ufl][main] INFO: FL_CMD_ATTACH I/O control error\n"))


				fprintf(stderr,
					"ERROR: Couldn't execute attach command, received unknown return code from kernel\n");


				if (packet.ret >= 1)
				{
					return %UNKNOWN_IOCTL_ATTACH_POS_RETURN_CODE[in,,]%;	// ERROR: Unknown positive return code for attach command
				}
				else
				{

					return %UNKNOWN_IOCTL_ATTACH_NEG_RETURN_CODE[in,,]%;	// ERROR: Unknown negative return code for attach command
				}


				break;
		}
	}


	// Check for return code (FL_CMD_DETACH)
	if (args.cmd == FL_CMD_DETACH)
	{
		FL_DEBUG_PRINTF(("[main] INFO: FL_CMD_DETACH packet.ret = %i\n",
				 packet.ret))


		switch (packet.ret)
                {
			case %fl_c:*:DETACH_SUCCESS[op,,]%:

				//  %fl_c:*:DETACH_SUCCESS[op,,]% - fl_cmd_detach:SUCCESS
				FL_TEST("main",\
					"DETACH_SUCCESS")
				ret = %DETACH_SUCCESS[ip,2,]%;


				break;

			case %fl_c:*:BLK_DEV_NAME_NOT_FOUND[on,,]%:

				//  %fl_c:*:BLK_DEV_NAME_NOT_FOUND[on,,]% - fl_cmd_detach:ERROR: BLK device name not found
				fprintf(stderr,
					"ERROR: Block device file not found\n");


				FL_TEST("main",\
					"BLK_DEV_NAME_NOT_FOUND")
                                ret = %BLK_DEV_NAME_NOT_FOUND[in,,]%;


                                break;

			case %fl_c:*:CHR_DEV_HAS_ACTIVE_PROCS[on,,]%:

				// %fl_c:*:CHR_DEV_HAS_ACTIVE_PROCS[on,,]% - fl_cmd_detach:ERROR: Active processes on CHR device
				fprintf(stderr,
					"ERROR: Couldn't detach loop, there are active processes using the char device file\n");


				FL_TEST("main",\
					"CHR_DEV_HAS_ACTIVE_PROCS")
                                ret = %CHR_DEV_HAS_ACTIVE_PROCS[in,,]%;


                                break;

			case %fl_c:*:BLK_DEV_HAS_ACTIVE_PROCS[on,,]%:

				// %fl_c:*:BLK_DEV_HAS_ACTIVE_PROCS[on,,]% - fl_cmd_detach:ERROR: Active processes on BLK device
				fprintf(stderr,
					"ERROR: Couldn't detach loop, there are active processes using the block device file\n");


				FL_TEST("main",\
					"BLK_DEV_HAS_ACTIVE_PROCS")
                                ret = %BLK_DEV_HAS_ACTIVE_PROCS[in,,]%;


                                break;

			default:

				// NOTE: Unknown return code


				FL_DEBUG_PRINTF(("[ufl][main] ERROR: ioctl() errno = %i strerror = \"%s\"\n",
						 errno,
						 strerror(errno)))


				FL_DEBUG_PRINTF(("[ufl][main] INFO: FL_CMD_DETACH I/O control error\n"))


				fprintf(stderr,
					"ERROR: Couldn't execute detach command, received unknown return code from kernel\n");


				if (packet.ret >= 1)
				{
					return %UNKNOWN_IOCTL_DETACH_POS_RETURN_CODE[in,,]%;	// ERROR: Unknown positive return code for detach command
				}
				else
				{

					return %UNKNOWN_IOCTL_DETACH_NEG_RETURN_CODE[in,,]%;	// ERROR: Unknown negative return code for detach command
				}


				break;
		}
	}


	// Check for return code (FL_CMD_LIST)
	if (args.cmd == FL_CMD_LIST)
	{
		FL_DEBUG_PRINTF(("[main] INFO: FL_CMD_LIST packet.ret = %i\n",
				 packet.ret))


		switch (packet.ret)
		{
			case %fl_c:*:NO_LOOP_ENTRIES[op,,]%:

				//  %fl_c:*:NO_LOOP_ENTRIES[op,,]% - fl_cmd_list:SUCCESS (NOTE: Data available)
				printf("INFO: No loops attached\n");


				FL_TEST("main",\
					"NO_LOOP_ENTRIES")
				ret = %NO_LOOP_ENTRIES[ip,3,]%;


				break;

			case %fl_c:*:LIST_TRUNCATED[op,,]%:

                                // NOTE: Truncated
                                //
                                //  There are more entries to be transferred from
                                //  the kernel module.


                               	// Parse response packet
                               	fl_packet_parse();


				while (FL_LOOP_FOREVER)
				{
					// NOTE: Next listing
					//
					//  Read consecutive listings until a return code
					//  of LIST_SUCCESS is returned.


					switch (fl_ioctl(fd,
							 args.cmd,
							 packet.data,
							 &packet.ret))
					{
						case 1:

							// NOTE: Positive return code


							// Check for valid return codes
							if ((packet.ret != %fl_c:*:LIST_TRUNCATED[op,,]%) &&
							    (packet.ret != %fl_c:*:LIST_SUCCESS[op,,]%))
							{
								FL_DEBUG_PRINTF(("[ufl][main] ERROR: ioctl() errno = %i strerror = \"%s\"\n",
										 errno,
										 strerror(errno)))


								FL_DEBUG_PRINTF(("[ufl][main] INFO: FL_CMD_LIST I/O control return code error\n"))


								fprintf(stderr,
									"ERROR: Couldn't execute list command\n");


								return %UNEXPECTED_IOCTL_LIST_POS_RETURN_CODE_AFTER_TRUNC[in,,]%;
							}


							// Parse response packet
							fl_packet_parse();


							break;

						case 2:

							// NOTE: Negative return code
							//
							//  The list command currently has no negative
							//  return codes.


							FL_DEBUG_PRINTF(("[ufl][main] ERROR: ioctl() errno = %i strerror = \"%s\"\n",
									 errno,
									 strerror(errno)))


							FL_DEBUG_PRINTF(("[ufl][main] INFO: FL_CMD_LIST I/O control error\n"))


							fprintf(stderr,
								"ERROR: Couldn't execute list command\n");


							return %UNEXPECTED_IOCTL_LIST_NEG_RETURN_VALUE_AFTER_TRUNC[in,,]%;

						case -4:

							// NOTE: Unexpected return value


							FL_DEBUG_PRINTF(("[ufl][main] ERROR: ioctl() errno = %i strerror = \"%s\"\n",
									 errno,
									 strerror(errno)))


							FL_DEBUG_PRINTF(("[ufl][main] INFO: FL_CMD_LIST I/O control error\n"))


							fprintf(stderr,
								"ERROR: Couldn't execute list command\n");


							return %UNEXPECTED_IOCTL_LIST_RETURN_VALUE_AFTER_TRUNC[in,,]%;

						case -5:

							// NOTE: Unexpected return code


							FL_DEBUG_PRINTF(("[ufl][main] INFO: ioctl() return value = %i\n",
									 errno))


							FL_DEBUG_PRINTF(("[ufl][main] INFO: FL_CMD_LIST I/O control error\n"))


							fprintf(stderr,
								"ERROR: Couldn't execute list command\n");


							return %UNEXPECTED_IOCTL_LIST_RETURN_CODE_AFTER_TRUNC[in,,]%;
					}


					// Check for LIST_SUCCESS
					if (packet.ret == %fl_c:*:LIST_SUCCESS[op,,]%)
					{
						// NOTE: List command finished
						//
						//  All listings have been read and displayed.


						break;
					}
				}


                                ret = %LIST_SUCCESS[ip,5,]%;


				break;

			case %fl_c:*:LIST_SUCCESS[op,,]%:

                                // NOTE: No more loop entries
                                //
                                //  There are no more loop entries to be transferred
                                //  from the kernel module.


                                // Parse response packet
                                fl_packet_parse();


				//  %fl_c:*:LIST_SUCCESS[op,,]% - fl_cmd_list:SUCCESS: Couldn't find matching loop entries (NOTE: No data available)
				FL_TEST("main",\
					"LIST_SUCCESS");
				ret = %LIST_SUCCESS[op,,]%;


				break;

			default:

				// NOTE: Unknown return code


				FL_DEBUG_PRINTF(("[ufl][main] ERROR: ioctl() errno = %i strerror = \"%s\"\n",
						 errno,
						 strerror(errno)))


				FL_DEBUG_PRINTF(("[ufl][main] INFO: FL_CMD_LIST I/O control error\n"))


				fprintf(stderr,
					"ERROR: Couldn't execute list command, received unknown return code from kernel\n");


				return %UNKNOWN_IOCTL_LIST_RETURN_CODE[in,,]%;	// ERROR: Unknown return code for list command


				break;
		}
	}


	// Close /dev/fl
	close(fd);


	FL_TEST_DEINIT()


	return ret;	// SUCCESS
}


%FUNC:main:STOP%


%GLOB:ufl_c:STOP%


