#ifndef __FL_H__
#define __FL_H__


// NOTE: OS header


#include "fl_conf_id.h"


// NOTE: System headers


#if FL_OS_NAME_V_IRIX
#       if FL_OS_VERSION_V_6_5
#               if FL_OS_RELEASE_V_30


#include <sys/systm.h>


#               endif 
#       endif
#endif


// NOTE: Local headers


#include "fl_common.h"
#include "fld.h"
#include "fl_list.h"
#include "fl_file.h"
#include "fl_sem.h"



/******************************************************************************
 * Forward definitions
 ******************************************************************************/


#if FL_OS_NAME_V_OSF1
#       if FL_OS_VERSION_V_5_1
#               if FL_OS_RELEASE_V_B


#               endif
#       endif
#elif FL_OS_NAME_V_IRIX
#       if FL_OS_VERSION_V_6_5
#               if FL_OS_RELEASE_V_30


struct install_cd_s
{
        // Is attached switch
        fl_uint32_t is_attached;


        // Partition offset (in bytes)
        fl_uint64_t part_offset;


        // Partition size (in bytes)
        fl_uint64_t part_size;
};


#               endif
#       endif
#endif


/******************************************************************************
 * Global definitions
 ******************************************************************************/


#define FL_DRIVER_PREFIX "fl_"
#define FL_DRIVER_NAME "fl"


// CMD (4 byte)
#define FL_PACKET_REQ_MIN 4


// NOTE: Command size
//
// CMD (4 byte)
// FILE SIZE (2 byte)
// FILE NAME (max. 1024)
// DEV SIZE (2 byte)
// DEV NAME (max. 1024)
//---------------------
// TOTAL 4 + 2 + 1024 + 2 + 1024 = 2056


#define FL_PACKET_REQ_MAX 2056
#define FL_PACKET_REQ_ATTACH FL_PACKET_REQ_MAX
#define FL_PACKET_REQ_DETACH 1030


#define FL_PROC_NULL 0
#define FL_PROC_RESET 1


// Session states
#define FL_ST_NULL 0
#define FL_ST_OPEN 1
#define FL_ST_DAEMON 2

#define FL_ERR_NULL 0
#define FL_ERR_NAMEI -1
#define FL_ERR_NO_LOOP_NEW -2
#define FL_ERR_NO_LOOP_ADD -3
#define FL_ERR_FILE_NOT_FOUND -4
#define FL_ERR_NO_NAMEIDATA_DUP -5

#define FL_CMD_NULL 0
#define FL_CMD_ATTACH 1
#define FL_CMD_DETACH 2
#define FL_CMD_LIST 3

#define FL_HWGRAPH_NULL 0
#define FL_HWGRAPH_CHR  1
#define FL_HWGRAPH_BLK  2


// NOTE: Global debug lock switch
//
//  The global debug lock switch allows denying access to any
//  driver entry point.
//
//  This feature is used for debugging purposes.


#define FL_DEBUG_LOCK_SWITCH(RET) \
if (lock_switch == 1)\
{\
        return RET;\
}


#define FL_DEBUG_LOCK_SWITCH_STRAT(RET) \
if (lock_switch == 1)\
{\
        bp->b_flags |= B_ERROR;\
        bp->b_resid = bp->b_bcount;\
        bp->b_error = RET;\
        biodone(bp);\
\
\
        return;\
}


#if defined(FL_DEBUG)


#define FL_DEBUG_DRV_GETPARM(PPID, PID) drv_getparm(PPID,\
						    PID);


#else


#define FL_DEBUG_DRV_GETPARM(PPID, PID)


#endif	// defined(FL_DEBUG)


// NOTE: Global serial lock
//
//  The global serial lock is used to serialize
//  access to the whole driver.
//
//  The following macros can be used to quickly
//  toggle this functionality.


#define FL_SERIAL_LOCK()
#define FL_SERIAL_UNLOCK()


//#define FL_SERIAL_LOCK() fl_serial_lock();
//#define FL_SERIAL_UNLOCK() fl_serial_unlock();


// NOTE: Platform independent declarations


enum fl_dst_e
{
        FL_DST_NULL,
        FL_DST_IDLE,
        FL_DST_REQ
};


enum fl_chr_io_e
{
        FL_CHR_IO_NULL  = 0x00000000,
        FL_CHR_IO_READ  = 0x00000001,
        FL_CHR_IO_WRITE = 0x00000002,
        FL_CHR_IO_LAST  = 0x00000002,
        FL_CHR_IO_MASK  = 0x00000003
};


struct fl_packet_s
{
        char dev[PATH_MAX];
        char file[PATH_MAX];
        char *data;
        char data_list[4 + 4 + 4];      // Quick buffer for list command:
                                        //  4 command code,
                                        //  4 list length,
                                        //  4 return code
        fl_uint16_t dev_size;
        fl_uint16_t file_size;
        fl_uint32_t cmd;
        fl_uint32_t size;
};


struct dsf_control_s
{
        fl_int32_t ret;
	fl_uint64_t list_index;			// Loop list start index
	fl_uint64_t list_size;			// Listing size in bytes
        struct fl_packet_s *packet;		// Request packet
        struct fl_packet_s packet_resp;		// Response packet
        struct loop_entry_s *loop_entry;	// Session loop entry
};


struct fl_proc_s
{
        fl_uint32_t state;              // Session state
        fl_uint32_t reset;              // Flag indicating session reset
        struct dsf_control_s *ctrl;     // Pointer to dsf_control
};


struct fl_disk_s
{


#if FL_OS_NAME_V_OSF1
#       if FL_OS_VERSION_V_5_1
#               if FL_OS_RELEASE_V_B


#               endif
#       endif
#elif FL_OS_NAME_V_IRIX
#       if FL_OS_VERSION_V_6_5
#               if FL_OS_RELEASE_V_30


        struct volume_header vh;


#               endif
#       endif
#endif


};


struct loop_entry_s
{
        fl_uint32_t chr_is_inuse;
        fl_list_t *pid_list_chr;
        fl_uint32_t blk_is_inuse;
        fl_list_t *pid_list_blk;
        fl_uint32_t first_open;
        struct file_s dev;
        struct file_s file;
        struct fl_packet_s packet;
        struct fl_spin_s *lock;


#if FL_OS_NAME_V_OSF1
#       if FL_OS_VERSION_V_5_1
#               if FL_OS_RELEASE_V_B


#               endif
#       endif
#elif FL_OS_NAME_V_IRIX
#       if FL_OS_VERSION_V_6_5
#               if FL_OS_RELEASE_V_30


        char *sector;
        fl_uint32_t is_exclusive;
        fl_uint32_t format_is_finished;
        fl_uint32_t format_error;
        struct install_cd_s install_cd;
        struct cred cred;
        struct file_s dev_chr;
        struct uio *uio;
        struct iovec *iov;
        fl_sem_t *format_sem;


#               endif
#       endif
#endif


        struct fl_disk_s disk;
};


/**************************** External References *****************************/


extern void             biodone(struct buf *bp);


// NOTE: OS specific declarations


#if FL_OS_NAME_V_OSF1
#       if FL_OS_VERSION_V_5_1
#               if FL_OS_RELEASE_V_B


#               endif
#       endif
#elif FL_OS_NAME_V_IRIX
#       if FL_OS_VERSION_V_6_5
#               if FL_OS_RELEASE_V_30


#define FL_PIO_KLISTIO 1
#define FL_PIO_KAIO 2


#define FL_PID_NULL 0
#define FL_PID_CHR 1
#define FL_PID_BLK 2
#define FL_PID_CHR_BLK 3


// Partition type NOT SUPPORTED
#define PTYPE_NOT_SUPPORTED 5


// Maximum number of blocks in an unsigned integer
#define FL_UINT_MAX_BLOCKS 8388608


// NOTE: NENTS
//
//  This macro is not defined anywhere in a
//  fully loaded IRIX development system.
//
//  Since the DIOCRDEFECTS ioctl command is defined and
//  needs this value, the NENTS macro is defined to
//  the value 0 (zero).
//
//  This writes back only the first defhdr structure
//  to user space which is enough to comply with
//  DIOCRDEFECTS (see man dksc).


struct defect_list
{
        struct defect_header defhdr;


        // Not defined due to missing NENTS system macro
        //struct defect_entry defentry[NENTS];
};


#ifdef FL_TESTS


// NOTE: Systune variables
//
//  Systune variables are used through the systune(1M) user command for unit tests.


//extern int test_glob;
//extern int test_func;
//extern int test_name;


#endif


#               endif
#       endif
#endif


/******************************************************************************
 * Private prototypes
 ******************************************************************************/


int fl_entry_new(struct loop_entry_s **);
int fl_entry_get_by_dev(dev_t,
                        struct loop_entry_s **,
			struct fl_args_s *);
int fl_node_get_by_dev_name(char *,
                            fl_list_node_t **,
                            fl_uint64_t *);


#if FL_OS_NAME_V_OSF1
#       if FL_OS_VERSION_V_5_1
#               if FL_OS_RELEASE_V_B


#               endif
#       endif
#endif


int fl_dev_create_loop_dir(void);
int fl_dev_compare(char *,
                   char *,
                   fl_int32_t *);
int fl_dev_map(dev_t,
               fl_uint32_t,
               struct loop_entry_s **,
               dev_t *,
               fl_uint32_t *,
               fl_uint32_t *,
	       struct fl_args_s *);
int fl_dev_check_minor_is_attached(fl_uint32_t);
int fl_dev_is_attached(char *,
                       fl_uint32_t *);
int fl_dev_get_lminor(fl_uint32_t *,
                      fl_uint32_t *);
int fl_dev_set_lminor(fl_uint32_t,
                      fl_uint32_t);
int fl_dev_check_path(char *);
int fl_dev_new(struct loop_entry_s *);
int fl_dev_get_size(dev_t,
                    fl_uint64_t *,
                    fl_uint32_t *,
                    fl_uint32_t *);
long long fl_dev_get_size_in_blocks(dev_t);
int fl_vh_check_install_cd(struct loop_entry_s *,
                           fl_uint32_t *);
int fl_loop_new(struct loop_entry_s *,
                fl_int32_t *);
int fl_loop_add(struct loop_entry_s *);
int fl_vnode_get_usecount(struct vnode *,
                          fl_int32_t *,
                          fl_int32_t *);
int fl_entry_rel(struct loop_entry_s *,
                 fl_uint32_t *);
int fl_pid_check_active(struct loop_entry_s *,
                        fl_uint32_t);
int fl_user_check(uid_t,
		  struct fl_args_s *);
int fl_cmd_attach_with_entry(struct loop_entry_s *);
int fl_cmd_attach(fl_int32_t *);
int fl_cmd_detach(void);
int fl_entry_serialize(struct loop_entry_s *,
                       char **);
int fl_packet_new_resp(void);
int fl_cmd_list(void);
int fl_session_reset(struct fl_proc_s *);
int fl_packet_read(char *,
                   struct fl_packet_s *);
int fl_data_get_cmd(char *,
                    fl_uint32_t *);


#if FL_OS_NAME_V_OSF1
#       if FL_OS_VERSION_V_5_1
#               if FL_OS_RELEASE_V_B


#               endif
#       endif
#endif


int fl_open_new_pid(dev_t,
                    fl_uint32_t);
int fl_close_empty_pids(dev_t,
                        fl_uint32_t);
int fl_int_compare_ranges(fl_uint64_t,
                          fl_uint32_t,
                          fl_uint32_t *);
int fl_int_get_digits(fl_uint32_t,
                      void *,
                      fl_uint32_t,
                      fl_uint32_t *);
int fl_ioctl_del_daemon(void);
int fl_strategy_alloc_io_node(enum fld_io_e,
                              buf_t *,
                              fl_uint64_t,
                              fl_uint64_t,
                              dev_t,
                              char *,
                              char *,
                              fl_uint64_t,
                              fl_uint32_t,
                              fl_list_node_t **,
                              struct fld_io_req_s **);
int fl_strategy_enqueue_io(enum fld_io_e,
                           buf_t *,
                           fl_uint64_t,
                           fl_uint64_t,
                           dev_t,
                           char *,
                           char *,
                           fl_uint64_t,
                           fl_uint32_t,
                           struct fld_io_req_s **,
                           fl_int32_t *);
int fl_strategy_dequeue_io(struct fld_io_req_s **);
int fl_strategy_requeue_io(void);
int fl_strategy_peek_io(struct fld_io_req_s **);


#if FL_OS_NAME_V_IRIX
#       if FL_OS_VERSION_V_6_5
#               if FL_OS_RELEASE_V_30


int fl_hwgraph_init(struct file_s *);
int fl_hwgraph_deinit(struct file_s *);
int fl_hwgraph_add_edges(char **,
                         vertex_hdl_t *,
                         vertex_hdl_t *);
int fl_hwgraph_new_dev(struct loop_entry_s *,
                       fl_uint32_t,
                       char *);
int fl_hwgraph_new_devfl(void);
int fl_hwgraph_remove_dev_by_path(char *,
                                  fl_uint32_t);
int fl_hwgraph_remove_dev(vertex_hdl_t,
                          char *,
                          vertex_hdl_t,
                          fl_uint32_t);


#               endif
#       endif
#endif


#if FL_OS_NAME_V_OSF1
#       if FL_OS_VERSION_V_5_1
#               if FL_OS_RELEASE_V_B


#               endif
#       endif
#elif FL_OS_NAME_V_IRIX
#       if FL_OS_VERSION_V_6_5
#               if FL_OS_RELEASE_V_30


int fl_init();
int fl_start();
int fl_reg();
int fl_unreg();
int fl_unload();


#               endif
#       endif
#endif


#if FL_OS_NAME_V_OSF1
#       if FL_OS_VERSION_V_5_1
#               if FL_OS_RELEASE_V_B


#               endif
#       endif
#elif FL_OS_NAME_V_IRIX
#       if FL_OS_VERSION_V_6_5
#               if FL_OS_RELEASE_V_30


int fl_open(dev_t *,
            int,
            int,
            cred_t *);
int fl_close(dev_t,
             int,
             int,
             cred_t *);
int fl_read(dev_t,
            uio_t *,
            cred_t *);
int fl_write(dev_t,
             uio_t *,
             cred_t *);
int fl_size(dev_t);
long long fl_size64(dev_t);
int fl_ioctl(dev_t,
             int,
             void *,
             int,
             cred_t *,
             int *);


#               endif
#       endif
#endif


void fl_strategy(struct buf *);


#endif	// __FL_H__


