

// NOTE: System header files


#if FL_OS_NAME_V_OSF1
#       if FL_OS_VERSION_V_5_1
#               if FL_OS_RELEASE_V_B


#               endif
#       endif
#elif FL_OS_NAME_V_IRIX
#       if FL_OS_VERSION_V_6_5
#               if FL_OS_RELEASE_V_30


#               endif
#       endif
#endif


// NOTE: Local header files


#include "fl_file.h"


%GLOB:fl_file:START%


// Import macros from fl_conf_id.h 
//  %IMPORT:GLOB:fl_conf_id:MACRO:$FL_OS_NAME_V_OSF1%
//  %IMPORT:GLOB:fl_conf_id:MACRO:$FL_OS_NAME_V_IRIX%
//  %IMPORT:GLOB:fl_conf_id:MACRO:$FL_OS_VERSION_V_5_1%
//  %IMPORT:GLOB:fl_conf_id:MACRO:$FL_OS_VERSION_V_6_5%
//  %IMPORT:GLOB:fl_conf_id:MACRO:$FL_OS_RELEASE_V_B%
//  %IMPORT:GLOB:fl_conf_id:MACRO:$FL_OS_RELEASE_V_30%


%FUNC:fl_file_init:START%


int fl_file_init(struct file_s *file)	// -- IN
{
	// NOTE: Arguments


	// IN: Check for file
	if (file == NULL)
	{
		return 0;	// ERROR: No file
	}


	// NOTE: Processing


	// Initialize file
	bzero((void *) file,
	      sizeof(struct file_s));


	return 1;	// SUCCESS
}


%FUNC:fl_file_init:STOP%
%FUNC:fl_file_deinit:START%


int fl_file_deinit(struct file_s *file)	// -- IN
{
	// NOTE: Arguments


	// IN: Check for file
	if (file == NULL)
	{
		return 0;	// ERROR: No file
	}


	// NOTE: Processing


	// Check for parent vnode
	if (file->dvp != NULL)
	{
		FL_DEBUG_UPRINTF((\
				  "[fl_file_remove] INFO: Before: file->dvp = %p file->dvp->count = %llu\n",
				  file->dvp,
				  file->dvp->v_count))


		VUNREF(file->dvp);


		FL_DEBUG_UPRINTF((\
				  "[fl_file_remove] INFO: After: file->dvp = %p file->dvp->count = %llu\n",
				  file->dvp,
				  file->dvp->v_count))
	}


	// Check for vnode
	if (file->vp != NULL)
	{
		FL_DEBUG_UPRINTF((\
				  "[fl_file_remove] INFO: Before: file->vp = %p file->vp->count = %llu\n",
				  file->vp,
				  file->vp->v_count))


		VUNREF(file->vp);


		FL_DEBUG_UPRINTF((\
				  "[fl_file_remove] INFO: After: file->vp = %p file->vp->count = %llu\n",
				  file->vp,
				  file->vp->v_count))
	}


	// Reinitialize file
	bzero((void *) file,
	      sizeof(struct file_s));


	return 1;	// SUCCESS
}


%FUNC:fl_file_deinit:STOP%
%FUNC:fl_file_lookup:START%


#if FL_OS_NAME_V_OSF1
#       if FL_OS_VERSION_V_5_1
#               if FL_OS_RELEASE_V_B


#               endif
#       endif
#elif FL_OS_NAME_V_IRIX
#       if FL_OS_VERSION_V_6_5
#               if FL_OS_RELEASE_V_30


int fl_file_lookup(struct file_s *file,		// -- IN
                   enum symfollow switches,	// -- IN
                   fl_uint32_t *is_found,	// -- OUT
                   fl_int32_t *ret)		// -- OUT


#               endif
#       endif
#endif


{
	fl_int32_t ret_lookup;


	// NOTE: Arguments


        // IN: Check for file
        if (file == NULL)
        {
		// Conditionally set return value
		if (ret != NULL)
		{
			*ret = %NO_FILE[in,0,]%;
		}


                return %NO_FILE[on,,]%;        // ERROR: No file
        }


#if FL_OS_NAME_V_OSF1
#       if FL_OS_VERSION_V_5_1
#               if FL_OS_RELEASE_V_B


#               endif
#       endif
#endif


	// OUT: Check for is found switch
        if (is_found == NULL)
        {
		// Conditionally set return value
		if (ret != NULL)
		{
			*ret = %NO_IS_FOUND[in,,]%;
		}


                return %NO_IS_FOUND[on,,]%;     // ERROR: No is found switch
        }


	// OUT: Check for return value
	if (ret == NULL)
	{
		return %NO_RETURN_VALUE[in,,]%;	// ERROR: No return value
	}


	// NOTE: Intialization


	// is found switch
        *is_found = 0;


	// NOTE: Processing


        FL_DEBUG_UPRINTF((\
			  "[fl_file_lookup] INFO: START\n"))


	// NOTE: Used file
	//
	//  When a file descriptor has any of its two vnodes set,
	//  dvp or vp, it was already used by any file library
	//  function.
	//
	//  To reuse the file descriptor, first deinitialize it,
	//  so that current vnode references get unreferenced.


	// Check for vnode
	if (file->vp != NULL)
	{
		return %VNODE_ALREADY_SET[in,,]%;	// ERROR: Vnode is already set
	}
	else
	{
		// NOTE: Parent vnode
		//
		//  The parent vnode may only be set when the (file->vp)
		//  vnode is set.


		// Check for parent vnode
		if (file->dvp != NULL)
		{
			return %PARENT_VNODE_ALREADY_SET[in,,]%;	// ERROR: Parent vnode is already set
		}
	}


        FL_DEBUG_UPRINTF((\
			  "[fl_file_lookup] INFO: File name to lookup = %s\n",
			  file->name))


#if FL_OS_NAME_V_OSF1
#       if FL_OS_VERSION_V_5_1
#               if FL_OS_RELEASE_V_B


#               endif
#       endif
#elif FL_OS_NAME_V_IRIX
#       if FL_OS_VERSION_V_6_5
#               if FL_OS_RELEASE_V_30


        ret_lookup = lookupname(file->name,
                                UIO_SYSSPACE,
                                switches,
				&file->dvp,
                                &file->vp,
                                NULL);


        FL_DEBUG_UPRINTF((\
			  "[fl_file_lookup] INFO: Lookupname(file->name=\"%s\") = %d\n",
			  file->name,
			  ret_lookup))


        FL_DEBUG_UPRINTF((\
			  "[fl_file_lookup] INFO: file->dvp = %p\n",
			  file->dvp))


        FL_DEBUG_UPRINTF((\
			  "[fl_file_lookup] INFO: file->vp = %p\n",
			  file->vp))


        // NOTE: Vnode refcount
        //
        //  On success, lookupname() returns a vnode with its refcount
        //  incremented by one (1).


#               endif
#       endif
#endif


        if (ret_lookup != 0)
        {
		// Set return value
		*ret = %ENTRY_NOT_FOUND[ip,,]%;


                return %ENTRY_NOT_FOUND[op,,]%; // SUCCESS: Couldn't find entry
	}


	// Check for found vnode
	if (file->vp == NULL)
	{
		// Check for parent vnode
		if (file->dvp != NULL)
		{
			// Unreference parent vnode
			VUNREF(file->dvp)


			// Reset parent vnode
			file->dvp = NULL;
		}


		// Set return value
		*ret = %ENTRY_NOT_FOUND_BUT_PARENT[ip,,]%;


                return %ENTRY_NOT_FOUND_BUT_PARENT[op,,]%; // SUCCESS: Couldn't find entry but parent was found
	}


	// NOTE: Filename found


	// Set is found switch
        *is_found = 1;


#if FL_OS_NAME_V_OSF1
#       if FL_OS_VERSION_V_5_1
#               if FL_OS_RELEASE_V_B


#               endif
#       endif
#elif FL_OS_NAME_V_IRIX
#       if FL_OS_VERSION_V_6_5
#               if FL_OS_RELEASE_V_30


        // Print parent vnode use count (after)
        FL_DEBUG_UPRINTF((\
			  "[fl_file_lookup] INFO: file->dvp->count = %d\n",
			  file->dvp->v_count))


        // Print vnode use count (after)
        FL_DEBUG_UPRINTF((\
			  "[fl_file_lookup] INFO: file->vp->count = %d\n",
			  file->vp->v_count))


#               endif
#       endif
#endif


        FL_DEBUG_UPRINTF((\
			  "[fl_file_lookup] INFO: STOP\n"))


	// Set return value
	*ret = %LOOKUP_SUCCESS[ip,1,]%;


        return %LOOKUP_SUCCESS[op,,]%; // SUCCESS
}


%FUNC:fl_file_lookup:STOP%


#if FL_OS_NAME_V_OSF1
#       if FL_OS_VERSION_V_5_1
#               if FL_OS_RELEASE_V_B


#               endif
#       endif
#elif FL_OS_NAME_V_IRIX
#       if FL_OS_VERSION_V_6_5
#               if FL_OS_RELEASE_V_30


int fl_file_open(struct file_s *file,	// -- IN
                 fl_int32_t mode,	// -- IN
                 struct cred *cred,	// -- IN
                 fl_uint32_t *is_open)	// -- OUT


#               endif
#       endif
#endif


{
        fl_int32_t ret;


#if FL_OS_NAME_V_IRIX
#       if FL_OS_VERSION_V_6_5
#               if FL_OS_RELEASE_V_30


        struct vnode *openvp;


#               endif
#       endif
#endif


	// NOTE: Arguments


        // IN: Check for file
        if (file == NULL)
        {
                return 0;       // ERROR: No file
        }


	// IN: Check for credentials
        if (cred == NULL)
        {
                return -1;      // ERROR: No credentials
        }


	// OUT: Check for is open switch
        if (is_open == NULL)
        {
                return -2;      // ERROR: No is open switch
        }


	// NOTE: Initialization


	// Is open switch
	*is_open = 0;


	// NOTE: Processing


        FL_DEBUG_UPRINTF((\
			  "[fl_file_open] INFO: START\n"))


        // Check for vnode
        if (file->vp == NULL)
        {
		// TODO: Similar file handling as in file_remove()


                return -3;      // ERROR: No vnode
        }


	// Check for open file
	if (file->is_open == 1)
	{
		return 2;	// SUCCESS: File already open
	}


#if FL_OS_NAME_V_OSF1
#       if FL_OS_VERSION_V_5_1
#               if FL_OS_RELEASE_V_B


#               endif
#       endif
#elif FL_OS_NAME_V_IRIX
#       if FL_OS_VERSION_V_6_5
#               if FL_OS_RELEASE_V_30


	// NOTE: VOP_OPEN
	//
	//  In IRIX, a vnode open does not change the reference
	//  count, nor the contents of file->vp on error.


        ret = 0;
        openvp = file->vp;


#if FL_VOP_OPEN_LIB == 1


	// Open vnode
        FL_VOP_OPEN(openvp,\
		    &file->vp,\
		    mode,\
		    cred,\
		    ret)


#else


	// Open vnode
	VOP_OPEN(openvp,\
		 &file->vp,\
		 mode,\
		 cred,\
		 ret)


#endif


#               endif
#       endif
#endif


        FL_DEBUG_UPRINTF((\
			  "[fl_file_open] INFO: VOP_OPEN() = %d\n",
			  ret))


	// Check VOP_OPEN() return value
	if (ret != 0)
	{
                return -4;      // ERROR: VOP_OPEN() error
        }


        // Set file open
        file->is_open = 1;


#if FL_OS_NAME_V_IRIX
#       if FL_OS_VERSION_V_6_5
#               if FL_OS_RELEASE_V_30


        // Set open flags mode
        file->open_flags = mode;


#               endif
#       endif
#endif


        // Set result variable
        *is_open = 1;


        FL_DEBUG_UPRINTF((\
			  "[fl_file_open] INFO: STOP\n"))


        return 1;       // SUCCESS
}


#if FL_OS_NAME_V_OSF1
#       if FL_OS_VERSION_V_5_1
#               if FL_OS_RELEASE_V_B


#               endif
#       endif
#elif FL_OS_NAME_V_IRIX
#       if FL_OS_VERSION_V_6_5
#               if FL_OS_RELEASE_V_30


int fl_file_create(struct file_s *file,		// -- IN
		   vtype_t type,		// -- IN
		   fl_int32_t flags,		// -- IN
		   fl_int32_t mode,		// -- IN
                   struct cred *cred,		// -- IN
                   fl_uint32_t *was_created)	// -- OUT


#               endif
#       endif
#endif


{
	char chr;
	char *file_name;
        fl_int32_t ret;
	fl_uint64_t basename_index;
        struct vattr va;


        // NOTE: Arguments


	// IN: Check for file
        if (file == NULL)
        {
                return 0;       // ERROR: No file
        }


	// IN: Check for valid file type
	if ((type != VREG) &&
	    (type != VDIR) &&
	    (type != VLNK))
	{
		return -1;	// ERROR: Invalid file type
	}


	// IN: Check for credentials
        if (cred == NULL)
        {
                return -2;      // ERROR: No credentials
        }


	// OUT: Check for was created switch
        if (was_created == NULL)
        {
                return -3;      // ERROR: No was created switch
        }


	// NOTE: Initialization


	// Was created switch
	*was_created = 0;


	// Vnode attributes
	bzero((void *) &va,
	      sizeof(struct vattr));


	// NOTE: Processing


        FL_DEBUG_UPRINTF((\
			  "[fl_file_create] INFO: START\n"))


	// Check for vnode
	if (file->vp != NULL)
	{
		return -4;	// ERROR: Vnode is set
	}


        // Check for filename
        if (file->name == NULL)
        {
                return -5;      // ERROR: No filename
        }


	// Check for empty filename
	if (file->name[0] == '\0')
	{
                return -6;      // ERROR: Empty filename
	}


	// Check for basename
	if (file->basename == NULL)
	{
		basename_index = 0;
		fl_file_set_basename(file,
				     &basename_index);
	}
	else
	{
		// NOTE: Get basename index


		// Set basename index
		basename_index = (fl_uint64_t) file->basename -
				 (fl_uint64_t) file->name;
	}


	// Check for symbolic link
	if (type == VLNK)
	{
		// Check for symbolic link target
		if (file->symlink_target[0] == '\0')
		{
			return -7;	// ERROR: No symbolic link target
		}
	}


#if FL_OS_NAME_V_OSF1
#       if FL_OS_VERSION_V_5_1
#               if FL_OS_RELEASE_V_B


#               endif
#       endif
#elif FL_OS_NAME_V_IRIX
#       if FL_OS_VERSION_V_6_5
#               if FL_OS_RELEASE_V_30


	// NOTE: Canonical pathname
	//
	//  The given filename is expected to be a canonical
	//  pathname to a file:
	//
	//   - it is an absolute pathname
	//   - it contains no . or ..
	//   - it has no consecutive slashes
	//   - it contains no symbolic links


	// NOTE: Terminate parent directory path


	if ((basename_index - 1) != 0)
	{
		file->name[basename_index - 1] = '\0';
	}
	else
	{
		// Save character after first slash
		chr = file->name[1];


		// Terminate string
		file->name[1] = '\0';
	}


        FL_DEBUG_UPRINTF((\
			  "[fl_file_create] INFO: lookup file->name = %s\n",
			  file->name))


        // Lookup vnode for last directory component
        ret = lookupname(file->name,
                         UIO_SYSSPACE,
                         NO_FOLLOW,
                         NULLVPP,
                         &file->dvp,
                         NULL);


	// NOTE: Restore '/' in filename


	if ((basename_index - 1) != 0)
	{
		file->name[basename_index - 1] = '/';
	}
	else
	{
		// Restore string
                file->name[1] = chr;
	}


	// Check for parent directory lookup
	if (ret != 0)
	{
		return -7;	// ERROR: Couldn't lookup parent directory
	}


	// Set filename
	file_name = file->basename;


	// NOTE: Fill vnode attributes


	va.va_mode = mode;
        va.va_mask |= AT_MODE;
        va.va_type = type;
        va.va_mask |= AT_TYPE;
	va.va_uid = cred->cr_uid;
        va.va_mask |= AT_UID;
        va.va_size = 0;
        va.va_mask |= AT_SIZE;
        va.va_gid = cred->cr_gid;
        va.va_mask |= AT_GID;


	// NOTE: Create file


	switch (type)
	{
		case VREG:

			// NOTE: Regular file


			FL_DEBUG_UPRINTF((\
					  "[fl_file_create] INFO: Create entry file_name = %s\n",
					  file_name))


			// Create file
			VOP_CREATE(file->dvp,\
				   file_name,\
				   &va,\
				   flags,\
				   mode,\
				   &file->vp,\
				   cred,\
				   ret)


			FL_DEBUG_UPRINTF((\
					  "[fl_file_create] INFO: VOP_CREATE() = %d\n",
					  ret))


			break;

		case VDIR:

			// NOTE: Directory


			// Create directory
			VOP_MKDIR(file->dvp,\
				  file_name,\
				  &va,\
				  &file->vp,\
				  cred,\
				  ret)


			FL_DEBUG_UPRINTF((\
					  "[fl_file_create] INFO: VOP_MKDIR() = %d\n",
					  ret))


			break;

		case VLNK:

			// NOTE: Symbolic link


			// Create symbolic link
		        VOP_SYMLINK(file->dvp,\
				    file_name,\
				    &va,\
				    file->symlink_target,\
				    cred,\
				    ret)


			FL_DEBUG_UPRINTF((\
					  "[fl_file_create] INFO: VOP_SYMLINK() = %d\n",
					  ret))


			// Check for created symbolic link
			if (ret != 0)
			{
				// Unref parent directory
				VUNREF(file->dvp)


				// Reset parent vnode
				file->dvp = NULL;


				return -7;	// ERROR: Couldn't create symbolic link
			}


			// NOTE: Symbolic link creation
			//
			//  The creation of a symbolic link does not place its
			//  vnode in file->vp, where it is expected by the file
			//  library.


			// Lookup symbolic link name
			ret = lookupname(file->name,
					 UIO_SYSSPACE,
					 NO_FOLLOW,
					 NULLVPP,
					 &file->vp,
					 NULL);


			break;

		default:

			// NOTE: Never reached


			break;
	}


#               endif
#       endif
#endif


	// Check for created entry type
	if (ret != 0)
	{
		// Unref parent directory
		VUNREF(file->dvp)


		// Reset parent vnode
		file->dvp = NULL;


                return -8;      // ERROR: Couldn't create entry type
        }


        // Set was created switch
        *was_created = 1;


        FL_DEBUG_UPRINTF((\
			  "[fl_file_create] INFO: STOP\n"))


	return 1;       // SUCCESS
}


#if FL_OS_NAME_V_OSF1
#       if FL_OS_VERSION_V_5_1
#               if FL_OS_RELEASE_V_B


#               endif
#       endif
#elif FL_OS_NAME_V_IRIX
#       if FL_OS_VERSION_V_6_5
#               if FL_OS_RELEASE_V_30


int fl_file_close(struct file_s *file,		// -- IN
                  struct cred *cred,		// -- IN
                  fl_uint32_t *is_closed)	// -- OUT


#               endif
#       endif
#endif


{
        fl_int32_t ret;


	// NOTE: Arguments


        // IN: Check for file
        if (file == NULL)
        {
                return 0;       // ERROR: No file
        }


	// IN: Check for credentials
        if (cred == NULL)
        {
                return -1;      // ERROR: No credentials
        }


	// OUT: Check for is closed switch
        if (is_closed == NULL)
        {
                return -2;      // ERROR: No is closed switch
        }


	// NOTE: Initialization


	// Is closed switch
        *is_closed = 0;


	// NOTE: Processing


        // Check for open file
        if (file->is_open != 1)
        {
                return -3;      // ERROR: File not open
        }


#if FL_OS_NAME_V_OSF1
#       if FL_OS_VERSION_V_5_1
#               if FL_OS_RELEASE_V_B


#               endif
#       endif
#elif FL_OS_NAME_V_IRIX
#       if FL_OS_VERSION_V_6_5
#               if FL_OS_RELEASE_V_30


        // Call VOP_CLOSE()
        ret = 0;
        VOP_CLOSE(file->vp,\
		  file->open_flags,\
		  L_FALSE,\
		  cred,\
		  ret)


#               endif
#       endif
#endif


        FL_DEBUG_UPRINTF((\
			  "[fl_file_close] INFO: VOP_CLOSE() = %d\n",
			  ret))


        // Check VOP_CLOSE() return value
        if (ret != 0)
	{
		return -4;      // ERROR: VOP_CLOSE() error
	}


	// NOTE: Close file
	//
	//  A closed file unreferences and resets its vnodes.


#if FL_OS_NAME_V_IRIX
#       if FL_OS_VERSION_V_6_5
#               if FL_OS_RELEASE_V_30


	// NOTE: Parent vnode
	//
	//  A parent vnode is not required to close a file.


	// Check for parent vnode
	if (file->dvp != NULL)
	{
        	// Unreference parent vnode
        	VUNREF(file->dvp)


		// Reset parent vnode
		file->dvp = NULL;
	}


	// Unreference vnode
	VUNREF(file->vp)


	// Reset vnode
	file->vp = NULL;


#               endif
#       endif
#endif


        // Set file close
        file->is_open = 0;


#if FL_OS_NAME_V_IRIX
#       if FL_OS_VERSION_V_6_5
#               if FL_OS_RELEASE_V_30


        // Reset open flags
        file->open_flags = 0;


#               endif
#       endif
#endif


        // Set is closed switch
        *is_closed = 1;


        return 1;       // SUCCESS
}


#if FL_OS_NAME_V_OSF1
#       if FL_OS_VERSION_V_5_1
#               if FL_OS_RELEASE_V_B


#               endif
#       endif
#elif FL_OS_NAME_V_IRIX
#       if FL_OS_VERSION_V_6_5
#               if FL_OS_RELEASE_V_30


int fl_file_read(struct file_s *file,		// -- IN
                 fl_int32_t ioflags,		// -- IN
                 struct cred *cred,		// -- IN
                 struct uio *uio,		// -- IN
                 fl_uint64_t *bytes_read)	// -- OUT


#               endif
#       endif
#endif


{
        fl_int32_t ret;
        fl_uint64_t bytes_size;


	// NOTE: Arguments


        // IN: Check for file
        if (file == NULL)
        {
                return 0;       // ERROR: No file
        }


        // TODO: Check ioflags


	// IN: Check for credentials
        if (cred == NULL)
        {
                return -1;      // ERROR: No credentials
        }


	// IN: Check for user I/O
        if (uio == NULL)
        {
                return -2;      // ERROR: No user I/O
        }


        // OUT: Check for bytes read
        if (bytes_read == NULL)
        {
                return -3;      // ERROR: No bytes read
        }


	// NOTE: Initialization


        // Bytes read
        *bytes_read = 0;


	// NOTE: Processing


        // Check user I/O residual count
        if (uio->uio_resid <= 0)
        {
                return -4;      // ERROR: Invalid user I/O residual count
        }


        // Save initial residual count
        bytes_size = uio->uio_resid;


        FL_DEBUG_UPRINTF((\
			  "[fl_file_read] INFO: vp = %p uio = %p uio->uio_resid = %llu from uio->uio_offset = %llu\n",
			  file->vp,
			  uio,
			  uio->uio_resid,
			  uio->uio_offset))


        FL_DEBUG_UPRINTF((\
			  "[fl_file_read] INFO: FL_CURUTHREAD = %p\n",
			  FL_CURUTHREAD))


#ifdef FL_DEBUG


        if (FL_CURUTHREAD != NULL)
        {
                FL_DEBUG_UPRINTF((\
				  "[fl_file_read] INFO: FL_CURUTHREAD->ut_flid = %p\n",
				  &FL_CURUTHREAD->ut_flid))
        }


#endif


#if FL_OS_NAME_V_OSF1
#       if FL_OS_VERSION_V_5_1
#               if FL_OS_RELEASE_V_B


#               endif
#       endif
#elif FL_OS_NAME_V_IRIX
#       if FL_OS_VERSION_V_6_5
#               if FL_OS_RELEASE_V_30


        FL_DEBUG_UPRINTF((\
			  "[fl_file_read] INFO: file = %p\n",
			  file))


        FL_DEBUG_UPRINTF((\
			  "[fl_file_read] INFO: cred = %p\n",
			  cred))


	// Print user I/O
        //fl_print_uio(uio);


        // VOP_READ()
        ret = 0;
        VOP_READ(file->vp,\
		 uio,\
		 ioflags,\
		 cred,\
		 &FL_CURUTHREAD->ut_flid,\
		 ret)


#               endif
#       endif
#endif


        FL_DEBUG_UPRINTF((\
			  "[fl_file_read] INFO: VOP_READ() = %d\n",
			  ret))


        // Check return value
        if (ret != 0)
        {
                return -5;      // ERROR: Read operation failed
        }


        // Set bytes read
        *bytes_read = bytes_size - uio->uio_resid;


        return 1;       // SUCCESS
}


#if FL_OS_NAME_V_OSF1
#       if FL_OS_VERSION_V_5_1
#               if FL_OS_RELEASE_V_B


#               endif
#       endif
#elif FL_OS_NAME_V_IRIX
#       if FL_OS_VERSION_V_6_5
#               if FL_OS_RELEASE_V_30


int fl_file_write(struct file_s *file,		// -- IN
                  fl_int32_t mode,		// -- IN
                  struct cred *cred,		// -- IN
                  struct uio *uio,		// -- IN
                  fl_uint64_t *bytes_written)	// -- OUT


#               endif
#       endif
#endif


{
        fl_int32_t ret;
        fl_uint64_t bytes_size;


	// NOTE: Arguments


        // IN: Check for file
        if (file == NULL)
        {
                return 0;       // ERROR: No file
        }


	// IN: Check for credentials
        if (cred == NULL)
        {
                return -1;      // ERROR: No credentials
        }


	// IN: Check for user I/O
        if (uio == NULL)
        {
                return -2;      // ERROR: No user I/O
        }


	// OUT: Check for bytes written
        if (bytes_written == NULL)
        {
                return -3;      // ERROR: No bytes written
        }


        // Self-assignments
        mode = mode;


	// NOTE: Initialization


	// Bytes written
	*bytes_written = 0;


	// NOTE: Processing


	// Check for vnode
	if (file->vp == NULL)
	{
		return -4;	// ERROR: No vnode
	}


        // Check size of write
        if (uio->uio_resid == 0)
        {
                return 2;       // SUCCESS: No bytes to write
        }


        // Store initial resid count
        bytes_size = uio->uio_resid;


	// Print user I/O
        //fl_print_uio(uio);


#if FL_OS_NAME_V_OSF1
#       if FL_OS_VERSION_V_5_1
#               if FL_OS_RELEASE_V_B


#               endif
#       endif
#elif FL_OS_NAME_V_IRIX
#       if FL_OS_VERSION_V_6_5
#               if FL_OS_RELEASE_V_30


        // VOP_WRITE()
        ret = 0;
        VOP_WRITE(file->vp,\
		  uio,\
		  mode,\
		  cred,\
		  &FL_CURUTHREAD->ut_flid,\
		  ret)


#               endif
#       endif
#endif


        FL_DEBUG_UPRINTF((\
			  "[fl_file_write] INFO: VOP_WRITE() = %d\n",
			  ret))


        // Check return value
        if (ret != 0)
        {
                return -5;      // ERROR: Vnode write operation failed
        }


        // Set bytes written
        *bytes_written = bytes_size - uio->uio_resid;


        return 1;       // SUCCESS
}


#if FL_OS_NAME_V_OSF1
#       if FL_OS_VERSION_V_5_1
#               if FL_OS_RELEASE_V_B


#               endif
#       endif
#elif FL_OS_NAME_V_IRIX
#       if FL_OS_VERSION_V_6_5
#               if FL_OS_RELEASE_V_30


int fl_file_stat(struct file_s *file,	// -- IN
		 fl_int32_t switches)	// -- IN


#               endif
#       endif
#endif


{
        fl_int32_t ret;


#if FL_OS_NAME_V_OSF1
#       if FL_OS_VERSION_V_5_1
#               if FL_OS_RELEASE_V_B


#               endif
#       endif
#elif FL_OS_NAME_V_IRIX
#       if FL_OS_VERSION_V_6_5
#               if FL_OS_RELEASE_V_30


        struct cred cred;
        fl_uint32_t cred_size = sizeof(struct cred);


#               endif
#       endif
#endif


	// NOTE: Arguments


        // IN: Check for file
        if (file == NULL)
        {
                return 0;       // ERROR: No file
        }


#if FL_OS_NAME_V_OSF1
#       if FL_OS_VERSION_V_5_1
#               if FL_OS_RELEASE_V_B


#               endif
#       endif
#elif FL_OS_NAME_V_IRIX
#       if FL_OS_VERSION_V_6_5
#               if FL_OS_RELEASE_V_30


        // TODO: Check switches


#               endif
#       endif
#endif


	// NOTE: Initialization


	// Credentials size
	cred_size = sizeof(struct cred);


	// NOTE: Processing


        FL_DEBUG_UPRINTF((\
			  "[fl_file_stat] INFO: START\n"))


        // Check for vnode
        if (file->vp == NULL)
        {
                return -2;      // ERROR: No vnode
        }


        // Set for compiler
        ret = 0;


        // NOTE: Credentials
        //
        //  Credentials are set to zero (root)
        //  to allow access to attributes of any
        //  vnode (i.e. file).


        // Set caller credentials
        bzero(&cred, cred_size);


        FL_DEBUG_UPRINTF((\
			  "[fl_file_stat] INFO: VOP_GETATTR()\n"))


#if FL_OS_NAME_V_OSF1
#       if FL_OS_VERSION_V_5_1
#               if FL_OS_RELEASE_V_B


#               endif
#       endif
#elif FL_OS_NAME_V_IRIX
#       if FL_OS_VERSION_V_6_5
#               if FL_OS_RELEASE_V_30


        // Get vnode attributes
        VOP_GETATTR(file->vp,\
		    file->vap,\
		    switches,\
		    &cred,\
		    ret)


#               endif
#       endif
#endif


        FL_DEBUG_UPRINTF((\
			  "[fl_file_stat] INFO: VOP_GETATTR() = %d\n",
			  ret))


        if (ret != 0)
        {
                return -3;      // ERROR: Couldn't get vnode attributes
        }


        FL_DEBUG_UPRINTF((\
			  "[fl_file_stat] INFO: STOP\n"))


        return 1;       // SUCCESS
}


%FUNC:fl_file_remove:START%


#if FL_OS_NAME_V_OSF1
#       if FL_OS_VERSION_V_5_1
#               if FL_OS_RELEASE_V_B


#               endif
#       endif
#elif FL_OS_NAME_V_IRIX
#       if FL_OS_VERSION_V_6_5
#               if FL_OS_RELEASE_V_30


int fl_file_remove(struct file_s *file,		// -- IN
		   fl_uint32_t *is_removed,	// -- OUT
		   fl_int32_t *ret)		// -- OUT


#               endif
#       endif
#endif


{
	fl_int32_t ret_remove;


#if FL_OS_NAME_V_OSF1
#       if FL_OS_VERSION_V_5_1
#               if FL_OS_RELEASE_V_B


#               endif
#       endif
#elif FL_OS_NAME_V_IRIX
#       if FL_OS_VERSION_V_6_5
#               if FL_OS_RELEASE_V_30


	fl_int32_t ret_lookup;
	fl_uint32_t has_filename;
	fl_uint32_t has_vnode;
	fl_uint32_t has_parent_vnode;
	fl_uint64_t basename_index;
        struct cred cred;


#               endif
#       endif
#endif


	// NOTE: Arguments


        // IN: Check for file
        if (file == NULL)
        {
		// Conditionally set return value
		if (ret != NULL)
		{
			*ret = %NO_FILE[in,,]%;
		}


                return %NO_FILE[on,,]%;       // ERROR: No file
        }


#if FL_OS_NAME_V_OSF1
#       if FL_OS_VERSION_V_5_1
#               if FL_OS_RELEASE_V_B


#               endif
#       endif
#endif


	// OUT: Check for is removed switch
        if (is_removed == NULL)
        {
		// Conditionally set return value
		if (ret != NULL)
		{
			*ret = %NO_IS_REMOVED_SWITCH[in,,]%;
		}


                return %NO_IS_REMOVED_SWITCH[on,,]%;   // ERROR: No is removed switch
        }


	// OUT: Check for return value
	if (ret == NULL)
	{
		return %NO_RETURN_VALUE[in,,]%;	// ERROR: No return value
	}


	// NOTE: Initialization


        // is removed switch
        *is_removed = 0;


	// NOTE: Processing


        FL_DEBUG_UPRINTF((\
			  "[fl_file_remove] INFO: START\n"))


	// NOTE: File or vnode
	//
	//  There are two possible scenarios for removing a file:
	//
	//   1. Using only a filename (file->name)
	//   2. Using a vnode with optional parent vnode (file->vp
	//      and file->dvp)


        // Check for filename
	has_filename = 1;
        if (file->name == NULL)
        {
		// NOTE: No filename


		FL_DEBUG_UPRINTF((\
				  "[fl_file_remove] INFO: No filename\n"))


		has_filename = 0;
        }
	else
	{
		// NOTE: Non-NULL filename


		// Check for empty filename
		if (file->name[0] == '\0')
		{
			FL_DEBUG_UPRINTF((\
					  "[fl_file_remove] INFO: Empty filename\n"))


			has_filename = 0;
		}
		else
		{
			FL_DEBUG_UPRINTF((\
					  "[fl_file_remove] INFO: Remove \"%s\"\n",
					  file->name))
		}
	}


	// Check for basename
	if (has_filename == 1)
	{
		if (file->basename == NULL)
		{
			// NOTE: No basename


			// Set basename
			basename_index = 0;
			fl_file_set_basename(file,
					     &basename_index);
		}
	}


	// Check for vnode
	has_vnode = 1;
	if (file->vp == NULL)
	{
		has_vnode = 0;
	}


	// Check for parent vnode
	has_parent_vnode = 1;
	if (file->dvp == NULL)
	{
		has_parent_vnode = 0;
	}


	// NOTE: Check for valid input combinations


	if ((has_vnode != 1) &&
	    (has_parent_vnode == 1))
	{
		// Set return value
		*ret = %NO_VNODE[in,,]%;


		return %NO_VNODE[on,,]%;	// ERROR: No vnode
	}


	// NOTE: File entity
	//
	//  A file entity is a description or link to a file.
	//
	//  This description can be either a filename, or a
	//  vnode.
	//
	//  At least one file description (or entity) must be
	//  given.


	if ((has_filename != 1) &&
	    (has_vnode != 1))
	{
		// Set return value
		*ret = %NO_FILE_ENTITY[in,,]%;


		return %NO_FILE_ENTITY[on,,]%;	// ERROR: No file entity
	}


#if FL_OS_NAME_V_IRIX
#       if FL_OS_VERSION_V_6_5
#               if FL_OS_RELEASE_V_30


        // NOTE: Credentials
        //
        //  Credentials are set to zero (root)
        //  to allow for creation of any file in
        //  /dev/loop .


        // Set caller credentials
        bzero(&cred,
	      sizeof(struct cred));


#               endif
#       endif
#endif


#if FL_OS_NAME_V_OSF1
#       if FL_OS_VERSION_V_5_1
#               if FL_OS_RELEASE_V_B


#               endif
#       endif
#elif FL_OS_NAME_V_IRIX
#       if FL_OS_VERSION_V_6_5
#               if FL_OS_RELEASE_V_30


	// NOTE: File removal scenarios
	//
	//  There are 3 possible scenarios at this point:
	//
	//   1. Only filename given
	//      (without any vnodes)
	//   2. Only vnode with parent vnode given
	//      (with or without filename)
	//   3. Only vnode without parent vnode given
	//      (with or without filename)


	if (has_vnode != 1)
	{
		// NOTE: Only filename given
		//
		//  Lookup the vnode and parent vnode for
		//  the file to be removed.


		// Lookup vnodes for device file deletion
		ret_lookup = lookupname(file->name,
					UIO_SYSSPACE,
					NO_FOLLOW,
					&file->dvp,
					&file->vp,
					NULL);


		// Check return value
		if (ret_lookup != 0)
		{
			FL_DEBUG_UPRINTF((\
					  "[fl_file_remove] INFO: lookupname() = %d\n",
					  ret_lookup))


			// Set return value
			*ret = %NO_VNODE_DEV_DELETION[inm,, $FL_OS_NAME_V_IRIX && $FL_OS_VERSION_V_6_5 && $FL_OS_RELEASE_V_30]%;


			return %NO_VNODE_DEV_DELETION[onm,, $FL_OS_NAME_V_IRIX && $FL_OS_VERSION_V_6_5 && $FL_OS_RELEASE_V_30]%;      // ERROR: Couldn't lookup vnodes for device file deletion
		}
	}
	else if ((has_vnode == 1) &&
		 (has_parent_vnode == 1))
	{
                // NOTE: Only vnode with parent vnode given
                //
                //  Since VOP_REMOVE needs the entry name of the
                //  file to be removed, get the entry name of the
                //  vnode, if not already given.


                // TODO: Get entry name corresponding to vnode
	}
	else
	{
		// NOTE: Only vnode without parent vnode given
		//
		//  Get parent vnode for given vnode. Since
		//  VOP_REMOVE needs the entry name of the file
		//  to be removed, get the entry name of the
		//  vnode too, if not already given.


		// TODO: Get parent vnode for vnode


		// TODO: Get entry name corresponding to vnode
	}


        FL_DEBUG_UPRINTF((\
			  "[fl_file_remove] INFO: Before: file->vp = %p file->vp->count = %llu\n",
			  file->vp,
			  file->vp->v_count))


        FL_DEBUG_UPRINTF((\
			  "[fl_file_remove] INFO: Before: file->dvp = %p file->dvp->count = %llu\n",
			  file->dvp,
			  file->dvp->v_count))


        // NOTE: Remove file
        //
        //  In IRIX 6.5, VOP_REMOVE() does not consume any reference.


	ret_remove = ESUCCESS;
        VOP_REMOVE(file->dvp,\
		   file->basename,\
		   &cred,\
		   ret_remove)


        FL_DEBUG_UPRINTF((\
			  "[fl_file_remove] INFO: After: file->vp = %p file->vp->count = %llu\n",
			  file->vp,
			  file->vp->v_count))


        FL_DEBUG_UPRINTF((\
			  "[fl_file_remove] INFO: After: file->dvp = %p file->dvp->count = %llu\n",
			  file->dvp,
			  file->dvp->v_count))


#               endif
#       endif
#endif


        FL_DEBUG_UPRINTF((\
			  "[fl_file_remove] INFO: VOP_REMOVE() = %d\n",
			  ret_remove))


        // Check return value
	if (ret_remove != 0)
	{
		// Set return value
		*ret = %VNODE_REMOVE_FAIL[in,,]%;


                return %VNODE_REMOVE_FAIL[on,,]%;      // ERROR: Vnode remove operation failed
        }


#if FL_OS_NAME_V_IRIX
#       if FL_OS_VERSION_V_6_5
#               if FL_OS_RELEASE_V_30


        // Unreference parent vnode
        VUNREF(file->dvp)


	// Reset parent vnode
	file->dvp = NULL;


	// Unreference vnode
	VUNREF(file->vp)


	// Reset parent vnode
	file->vp = NULL;


#               endif
#       endif
#endif


        // Set is removed switch
        *is_removed = 1;


        FL_DEBUG_UPRINTF((\
			  "[fl_file_remove] INFO: STOP\n"))


	// Set return value
	*ret = 1;


        return 1;       // SUCCESS
}


%FUNC:fl_file_remove:STOP%
%FUNC:fl_file_split_path:START%


int fl_file_split_path(char *path_name,		// -- IN
		       char ***path_comp,	// -- OUT
		       fl_uint32_t *comp_num)	// -- OUT
{
	char *path_buffer;
	char *path_pointer;
	fl_uint32_t num_slash;
	fl_uint32_t comp_index;
	fl_uint32_t buffer_length;


	// NOTE: Arguments


	// IN: Check for pathname
	if (path_name == NULL)
	{
		return 0;	// ERROR: No pathname
	}


	// OUT: Check for path components
	if (path_comp == NULL)
	{
		return -1;	// ERROR: No path components
	}


	// OUT: Check for number of components
	if (comp_num == NULL)
	{
		return -2;	// ERROR: No number of components
	}


	// NOTE: Processing
	//
	//  Pathname is expected to be a canonical pathname.


	FL_DEBUG_UPRINTF((\
			  "[fl_file_split_path] INFO: START\n"))


	FL_DEBUG_UPRINTF((\
			  "[fl_file_split_path] INFO: Canonical pathname \"%s\"\n",
			  path_name))


	// NOTE: Trivial case
	//
	//  A valid canonical pathname of '/' is the trivial case.


	if ((path_name[0] == '/') &&
	    (path_name[1] == '\0'))
	{
		// Set number of components
		*comp_num = 0;


		return 2;	// SUCCESS: Pathname is '/'
	}


	// Set path pointer
	path_pointer = path_name;


	// Loop through pathname
	num_slash = 0;
	while (*path_pointer != '\0')
	{
		// Check for slash
		if (*path_pointer == '/')
		{
			// Increment slash number
			num_slash++;
		}


		// Increment path pointer
		path_pointer++;
	}


	// Allocate array of component pointers
	fl_mem_alloc((void **) path_comp,
		     num_slash * sizeof(char *),
		     FL_MEM_WAITOK,
		     "fl_file_split_path",
		     "Allocate pathname component pointers");


	FL_DEBUG_UPRINTF((\
			  "[fl_file_split_path] INFO: *path_comp = %p\n",
			  (void *) *path_comp))


	// NOTE: Pathname buffer
	//
	//  The pointers to pathname components point to different
	//  indices of a single buffer, large enough to hold all
	//  components separated by terminating NULL bytes.


	FL_DEBUG_UPRINTF((\
			  "[fl_file_split_path] INFO: strlen(path_name) = %u\n",
			  strlen(path_name)))


	// Allocate a pathname buffer
	fl_mem_alloc((void **) &path_buffer,
		     strlen(path_name),
		     FL_MEM_WAITOK,
		     "fl_file_split_path",
		     "Allocate pathname buffer");


	FL_DEBUG_UPRINTF((\
			  "[fl_file_split_path] INFO: path_buffer = %p\n",
			  path_buffer))


	// NOTE: Scancopy path components
	//
	//  In the second loop, scan pathname for components,
	//  copy the component to the new buffer, and setup
	//  the corresponding component pointer.


	// Set path pointer
	path_pointer = path_name;


	// Loop through pathname
	num_slash = 0;
	buffer_length = 0;
	while (*path_pointer != '\0')
	{
		// Check for slash
		if (*path_pointer == '/')
		{
                        // Increment path pointer
                        path_pointer++;


			// Set component pointer
                        (*path_comp)[num_slash] = &path_buffer[buffer_length];


                        // Loop through component
			comp_index = 0;
                        while ((*path_pointer != '/') &&
                               (*path_pointer != '\0'))
                        {
				// Copy character to path buffer
				path_buffer[buffer_length] = *path_pointer;


				//FL_DEBUG_UPRINTF((\
			  			  //"[fl_file_split_path] INFO: path_buffer[%u] = %c\n",
			  			  //buffer_length,
						  //*path_pointer))


				// Increment buffer pointer
                                buffer_length++;


                                // Increment component
                                comp_index++;


                                // Increment path pointer
                                path_pointer++;
                        }


			// Terminate component string
			path_buffer[buffer_length] = '\0';


			//FL_DEBUG_UPRINTF((\
					  //"[fl_file_split_path] INFO: path_buffer[%u] = \\0\n",
					  //buffer_length))


			// Increment buffer pointer
                        buffer_length++;


			//FL_DEBUG_UPRINTF((\
					  //"[fl_file_split_path] INFO: Component \"%s\"\n",
					  //(*path_comp)[num_slash]))


			// Increment slash number
			num_slash++;
		}
		else
		{
			// Increment path pointer
			path_pointer++;
		}
	}


	FL_DEBUG_UPRINTF((\
			  "[fl_file_split_path] INFO: Number of components %u\n",
			  num_slash))


	// Set number of components
	*comp_num = num_slash;


	FL_DEBUG_UPRINTF((\
			  "[fl_file_split_path] INFO: STOP\n"))


	return 1;	// SUCCESS
}


%FUNC:fl_file_split_path:STOP%


int fl_file_set_basename(struct file_s *file,		// -- IN
			 fl_uint64_t *basename_index)	// -- OUT
{
	fl_uint64_t name_index;
	fl_uint64_t name_length;
	fl_uint64_t last_slash;


	// NOTE: Arguments


	// IN: Check for file
	if (file == NULL)
	{
		return 0;	// ERROR: No file
	}


	// OUT: Check for basename index
	if (basename_index == NULL)
	{
		return -1;	// ERROR: No basename index
	}


	// NOTE Initialization


	// Basename index
	*basename_index = 0;


	// NOTE: Processing


	FL_DEBUG_UPRINTF((\
			  "[fl_set_basename] INFO: START\n"))


	// Check for filename
	if (file->name == NULL)
	{
		return -2;	// ERROR: No filename
	}


	// Check for empty filename
	if (file->name[0] == '\0')
	{
		return -3;	// ERROR: Empty filename
	}


	// NOTE: Get last directory component of filename


	// Get length of filename string
	name_length = strlen(file->name);


	// Check for valid filename length
	if (name_length >= PATH_MAX)
	{
		return -4;	// ERROR: Invalid filename length
	}


	// Find last '/'
	for (name_index = 0;
	     name_index < name_length;
	     name_index++)
	{
		if (file->name[name_index] == '/')
		{
			last_slash = name_index;
		}
	}


	// Set basename
	file->basename = &file->name[last_slash + 1];


	FL_DEBUG_UPRINTF((\
			  "[fl_set_basename] INFO: file->name = \"%s\"\n",
			  file->name))


	FL_DEBUG_UPRINTF((\
			  "[fl_set_basename] INFO: file->basename = \"%s\"\n",
			  file->basename))


	// Set basename index
	*basename_index = last_slash + 1;


	FL_DEBUG_UPRINTF((\
			  "[fl_set_basename] INFO: basename_index = %llu\n",
			  *basename_index))


	FL_DEBUG_UPRINTF((\
			  "[fl_set_basename] INFO: STOP\n"))


	return 1;	// SUCCESS
}


%GLOB:fl_file:STOP%


