

#ifndef __FL_FILE_H__
#define __FL_FILE_H__


// NOTE: OS header


#include "fl_conf_id.h"


// NOTE: System header files


#if FL_OS_NAME_V_OSF1
#       if FL_OS_VERSION_V_5_1
#               if FL_OS_RELEASE_V_B


#               endif
#       endif
#elif FL_OS_NAME_V_IRIX
#       if FL_OS_VERSION_V_6_5
#               if FL_OS_RELEASE_V_30


#include <sys/uio.h>
#include <sys/cred.h>
#include <sys/vnode.h>


#               endif
#       endif
#endif


// NOTE: Local header files


#include "fl_common.h"
#include "fl_proc.h"


// NOTE: Macro definitions


#define FL_FILE_BLOCK 512


#if FL_OS_NAME_V_OSF1
#       if FL_OS_VERSION_V_5_1
#               if FL_OS_RELEASE_V_B


#               endif
#       endif
#elif FL_OS_NAME_V_IRIX
#       if FL_OS_VERSION_V_6_5
#               if FL_OS_RELEASE_V_30


#define FL_CURUTHREAD		(FL_PRIVATE.p_curuthread)
#define VREF(VP)                VN_HOLD(VP);
#define VUNREF(VP)              VN_RELE(VP);


// NOTE: CPU specific vnode ops
//
//  To avoid compiler WARNINGs, some vnode operations are implemented
//  adhoc for some platforms.


#ifdef FL_CPU_PLATFORM_V_32


#define FL_VOP_OPEN_LIB 1


#define FL_VOP_OPEN(VP, DVP, MODE, CRED, RET) \
{\
	RET = (*((vnodeops_t *) (VP)->v_bh.bh_first->bd_ops)->vop_open) ((VP)->v_bh.bh_first, DVP, MODE, CRED);\
}


#else


#define FL_VOP_OPEN_LIB 0


#endif	// FL_CPU_PLATFORM_V_32


#               endif
#       endif
#endif


struct file_size_s
{
        fl_uint64_t bytes;
        fl_uint64_t sectors;
};


struct file_s
{
        char *name;
	struct vnode *dvp;
	struct vnode *vp;
        struct vattr *vap;
        struct file_size_s size;
	dev_t num;
        unsigned int minor;
        fl_int64_t open_mode;
        fl_uint64_t refcount;
        fl_uint32_t is_open;
        fl_uint32_t is_created;


#if FL_OS_NAME_V_IRIX
#       if FL_OS_VERSION_V_6_5
#               if FL_OS_RELEASE_V_30


	char *basename;
	char symlink_target[PATH_MAX];
	void *iov_base_aligned;
        fl_uint32_t open_flags;
        vertex_hdl_t vertex;


#               endif
#       endif
#endif


};


#if FL_OS_NAME_V_OSF1
#       if FL_OS_VERSION_V_5_1
#               if FL_OS_RELEASE_V_B
 
 
#               endif
#       endif
#elif FL_OS_NAME_V_IRIX
#       if FL_OS_VERSION_V_6_5
#               if FL_OS_RELEASE_V_30
 
 
int fl_file_init(	 struct file_s *);

int fl_file_deinit(	 struct file_s *);

int fl_file_lookup(	 struct file_s *,
                   	 enum symfollow,
                   	 fl_uint32_t *,
                   	 fl_int32_t *);

int fl_file_open(  	 struct file_s *,
                   	 fl_int32_t,
                   	 struct cred *,
                   	 fl_uint32_t *);

int fl_file_create(	 struct file_s *,
                   	 vtype_t,
                   	 fl_int32_t,
                   	 fl_int32_t,
                   	 struct cred *,
                   	 fl_uint32_t *);

int fl_file_close( 	 struct file_s *,
                   	 struct cred *,
                   	 fl_uint32_t *);

int fl_file_read(  	 struct file_s *,
                   	 fl_int32_t,
                   	 struct cred *,
                   	 struct uio *,
                   	 fl_uint64_t *);

int fl_file_write( 	 struct file_s *,
                   	 fl_int32_t,
                   	 struct cred *,
                   	 struct uio *,
                   	 fl_uint64_t *);

int fl_file_stat(  	 struct file_s *,
		   	 fl_int32_t);

int fl_file_remove(	 struct file_s *,
		   	 fl_uint32_t *,
		   	 fl_int32_t *);

int fl_file_split_path(	 char *,
                       	 char ***,
                       	 fl_uint32_t *);

int fl_file_set_basename(struct file_s *,
                         fl_uint64_t *);


#               endif
#       endif
#endif


#endif	// __FL_FILE_H__


