

#include "fl_list.h"

#if FL_COMP_USER


#include <stdlib.h>


#elif FL_KERNEL_OSF1_5_1_B


#elif FL_KERNEL_IRIX_6_5_30


#include <sys/types.h>
#include <sys/kmem.h>
#include <sys/ddi.h>


#endif



/* Global configuration */
#ifdef FL_LIST_DEBUG
fl_list_conf_t conf_g;
#else
static fl_list_conf_t conf_g;
#endif


/*** fl_list_init() ***/
int fl_list_init(fl_list_conf_t *conf)
{
	/*** Declare variables ***/


	/*** Set variables ***/


        /*** Initialize configuration ***/
        if (conf != NULL)
        {
                conf_g.malloc = conf->malloc;
                conf_g.free = conf->free;
        }
        else
        {


#if FL_MEM_MALLOC


                conf_g.malloc = malloc;


#elif FL_MEM_KMALLOC_OSF1


#endif


#if FL_MEM_FREE


                conf_g.free = free;


#elif FL_MEM_KFREE_OSF1


#endif


#if FL_KERNEL_IRIX_6_5_30


		conf_g.malloc = kmem_alloc;
		conf_g.free = kmem_free;


#endif


        }


	/*** Return ***/
	return 1;	/* Success */
}


/*** fl_list_deinit() ***/
int fl_list_deinit(void)
{
	/*** Declare variables ***/


	/*** Set variables ***/


	/*** Deinitialize configuration ***/
	/* Reinitialize with system functions */
#if FL_MEM_MALLOC


	conf_g.malloc = malloc;


#elif FL_MEM_KMALLOC_OSF1


#elif FL_KERNEL_IRIX_6_5_30


	conf_g.malloc = kmem_alloc;


#endif


#if FL_MEM_FREE


	conf_g.free = free;


#elif FL_MEM_KFREE_OSF1


#elif FL_KERNEL_IRIX_6_5_30


	conf_g.free = kmem_free;


#endif


	/*** Return ***/
	return 1;	/* Success */
}


/*** fl_list_new() ***/
int __fl_list_new_proc(fl_list_t **l)
{


#if FL_MEM_MALLOC


        /*** Declare variables ***/
        size_t sl;


#elif FL_MEM_KMALLOC_OSF1


#elif FL_KERNEL_IRIX_6_5_30


        /*** Declare variables ***/
        size_t sl;


#endif
	char *mp;
	unsigned long long m;


        /*** Create list ***/
        sl = sizeof(fl_list_t);


#if FL_MEM_MALLOC


        if (((*l) = (fl_list_t *) (*conf_g.malloc)(sl)) == NULL)


#elif FL_MEM_KMALLOC_OSF1


#elif FL_KERNEL_IRIX_6_5_30


	// kmem_alloc(size_t size, int flags)
	if (((*l) = (fl_list_t *) (*conf_g.malloc)(sl, KM_SLEEP)) == NULL)


#endif
	{
		return -2;	/* malloc() error */
	}

	/* Clear struct */
	mp = (char *) *l;
	for (m = 0; m < (unsigned long long) sl; m++)
		mp[m] = 0;

	/* Copy configuration */
	(*l)->conf.malloc = conf_g.malloc;
	(*l)->conf.free = conf_g.free;


	/*** Return ***/
	return 1;	/* Success */
}


int __fl_list_new_check(fl_list_t **l)
{
	/*** Check parameters ***/
	if (l == NULL)
	{
		return 0;	/* No list */
	}


	/*** Return ***/
	return __fl_list_new_proc(l);
}


int fl_list_new(fl_list_t **list)
{
	/*** Return ***/
	return __fl_list_new_check(list);
}


/*** fl_list_free() ***/
int __fl_list_free_proc(fl_list_t **l)
{


#if FL_MEM_FREE


        /*** Declare variables ***/
        size_t sl;
	void (*freep)(void *);


#elif FL_MEM_KFREE_OSF1


#elif FL_KERNEL_IRIX_6_5_30


	size_t sl;
	void (*freep)(void *, size_t);


#endif


	char *mp;
	unsigned long long m;

	/* Set variables */
        freep = (*l)->conf.free;


	/*** Free list ***/
	/* Clear struct */
	sl = sizeof(fl_list_t);
	mp = (char *) *l;
        for (m = 0; m < (unsigned long long) sl; m++)
                mp[m] = 0;


#if FL_MEM_FREE


        /* Free */
        (*freep)(*l);


#elif FL_MEM_KFREE_OSF1


#elif FL_KERNEL_IRIX_6_5_30


	(*freep)((void *) *l, sizeof(fl_list_t));


#endif

	/* Reset */
	*l = NULL;


	/*** Return ***/
	return 1;	/* Success */
}


int __fl_list_free_check(fl_list_t **l)
{
	/*** Check parameters ***/
	if (l == NULL)
	{
		return -1;	/* No list */
	}


	/*** Return ***/
	return __fl_list_free_proc(l);
}


int fl_list_free(fl_list_t **list)
{
	/*** Return ***/
	return __fl_list_free_check(list);
}


/*** fl_list_new_node() ***/
int __fl_list_new_node_proc(fl_list_t *l, fl_list_node_t **n)
{


#if FL_MEM_MALLOC


        /*** Declare variables ***/
        size_t sn;


#elif FL_MEM_KMALLOC_OSF1


#elif FL_KERNEL_IRIX_6_5_30


	size_t sn;


#endif


	char *mp;
	unsigned long long m;


        /*** Create new node ***/
        sn = sizeof(fl_list_node_t);


#if FL_MEM_MALLOC


        if (((*n) = (fl_list_node_t *) (*(l->conf.malloc))(sn)) == NULL)


#elif FL_MEM_KMALLOC_OSF1


#elif FL_KERNEL_IRIX_6_5_30


	if (((*n) = (fl_list_node_t *) (*(l->conf.malloc))(sn, KM_SLEEP)) == NULL)


#endif
	{
		return -2;	/* malloc() error */
	}


	/* Clear struct */
	mp = (char *) *n;
        for (m = 0; m < (unsigned long long) sn; m++)
                mp[m] = 0;



	/*** Return ***/
	return 1;	/* Success */
}


int __fl_list_new_node_check(fl_list_t *l, fl_list_node_t **n)
{
	/*** Check parameters ***/
	if (l == NULL)
	{
		return 0;	/* No list */
	}

	if (n == NULL)
	{
		return -1;	/* No node */
	}


	/*** Return ***/
	return __fl_list_new_node_proc(l, n);
}


int fl_list_new_node(fl_list_t *list, fl_list_node_t **node)
{
	/*** Return ***/
	return __fl_list_new_node_check(list, node);
}


/*** fl_list_free_node() ***/
int __fl_list_free_node_proc(fl_list_t *l, fl_list_node_t **n)
{


#if FL_MEM_FREE


	/*** Declare variables ***/
        size_t sn;


#elif FL_MEM_KFREE_OSF1


#elif FL_KERNEL_IRIX_6_5_30


	size_t sn;


#endif


	char *mp;
	unsigned long long m;


	/*** Free node ***/
	/* Clear struct */
	sn = sizeof(fl_list_node_t);
	mp = (char *) *n;
        for (m = 0; m < (unsigned long long) sn; m++)
                mp[m] = 0;


#if FL_MEM_FREE


        /* Free */
        (*(l->conf.free))(*n);


#elif FL_MEM_KFREE_OSF1


#elif FL_KERNEL_IRIX_6_5_30


	(*(l->conf.free))((void *) *n, sizeof(fl_list_node_t));


#endif

	/* Reset */
	*n = NULL;


	/*** Return ***/
	return 1;	/* Success */
}


int __fl_list_free_node_check(fl_list_t *l, fl_list_node_t **n)
{
	/*** Check parameters ***/
	if (l == NULL)
	{
		return 0;	/* No list */
	}

	if (n == NULL)
	{
		return -1;	/* No node */
	}


	/*** Return ***/
	return __fl_list_free_node_proc(l, n);
}


int fl_list_free_node(fl_list_t *list, fl_list_node_t **node)
{
	/*** Return ***/
	return __fl_list_free_node_check(list, node);
}


/*** fl_list_insert() ***/
int __fl_list_insert_proc(fl_list_t *l, fl_list_node_t *n, unsigned long long p)
{
	/*** Declare variables ***/
	unsigned long long le;
	unsigned long long i;
	fl_list_node_t *fn;
	fl_list_node_t *ln;
	fl_list_node_t *pn;
	fl_list_node_t *prn;


	/*** Set variables ***/
	le = l->length;


	/*** Insert node ***/
	if (le == 0)
	{
		/* List is empty */

		/* Update descriptor */
		l->first = n;
		l->last = n;

		/* Update new node */
		n->next = n;
		n->prev = n;

		/* Update length */
		l->length = 1;
	}
	else	/* le > 0 */
	{
		/* List is not empty */

		if (p == le)
		{
			/* Append */

			/* Update descriptor */
			ln = l->last;
			l->last = n;

			/* Update new node */
			fn = l->first;
			n->next = fn;
			n->prev = ln;

			/* Update first node */
			fn->prev = n;

			/* Update previous node */
			ln->next = n;
		}
		else	/* p >= 0 && p <= (le - 1) */
		{
			/* Insert */

			/* Get node at position */
			pn = l->first;
			for (i = 0; i < le; i++)
			{
				if (i == p)
				{
					break;
				}

				pn = pn->next;
			}

			/* Set node pointers */
			if (p == 0)
			{
				/* Update descriptor */
				l->first = n;

				/* Update new node */
				ln = l->last;
				n->next = pn;
				n->prev = ln;

				/* Update positional node */
				pn->prev = n;

				/* Update last node */
				ln->next = n;
			}
			else	/* p > 0 && p <= (le - 1) */
			{
				/* NOTE: list has minimum size of 2 nodes at this point */

				/* Update descriptor */
				/* - */

				/* Update new node */
				prn = pn->prev;
				n->next = pn;
				n->prev = prn;

				/* Update positional node */
				pn->prev = n;

				/* Update previous node */
				prn->next = n;
			}
		}

		/* Update length */
		l->length += 1;
	}


	/*** Return ***/
	return 1;	/* Success */
}


int __fl_list_insert_check(fl_list_t *list, fl_list_node_t *node, unsigned long long position)
{
	/*** Check parameters ***/
	if (list == NULL)
	{
		return 0;	/* No list */
	}

	if (node == NULL)
	{
		return -1;	/* No node */
	}

	if (position > list->length)
	{
		return -2;	/* Position out of range */
	}


	/*** Return ***/
	return __fl_list_insert_proc(list, node, position);
}


/*** fl_list_delete() ***/
int __fl_list_delete_proc(fl_list_t *l, fl_list_node_t **n, unsigned long long p)
{
        /*** Declare variables ***/
        unsigned long long le;
        unsigned long long i;
        fl_list_node_t *fn;
        fl_list_node_t *ln;
        fl_list_node_t *pn;
        fl_list_node_t *prn;
	fl_list_node_t *nen;


	/*** Set variables ***/
	le = l->length;


	/*** Delete ***/
	if (le == 1)
	{
		/* Update descriptor */
		(*n) = l->first;
		l->first = NULL;
		l->last = NULL;

		/* Update length */
		l->length = 0;
	}
	else if (le == 2)
	{
		if (p == 0)
		{
			/* Update descriptor */
			(*n) = l->first;
			fn = l->last;
			l->first = fn;

			/* Update new first node */
			fn->next = fn;
			fn->prev = fn;
		}
		else /* p == 1 */
		{
			/* Update descriptor */
			(*n) = l->last;
			ln = l->first;
			l->last = ln;

			/* Update new last node */
			ln->next = ln;
			ln->prev = ln;
		}

		/* Update length */
		l->length = 1;
	}
	else /* le >= 3 */
	{
		if (p == 0)
		{
			/* Update descriptor */
			(*n) = l->first;
			fn = (*n)->next;
			l->first = fn;

			/* Update new first node */
			ln = l->last;
			fn->prev = ln;

			/* Update last node */
			ln->next = fn;
		}
		else if (p == (le - 1))
		{
			/* Update descriptor */
			(*n) = l->last;
			ln = (*n)->prev;
			l->last = ln;

			/* Update new last node */
			fn = l->first;
			ln->next = fn;

			/* Update first node */
			fn->prev = ln;
		}
		else
		{
			/* Get node at position */
			pn = l->first;
			pn = pn->next;
			for (i = 1; i < (le - 1); i++)
			{
				if (i == p)
				{
					break;
				}

				pn = pn->next;
			}

			/* Get node */
			(*n) = pn;

			/* Update previous node */
			prn = pn->prev;
			nen = pn->next;
			prn->next = nen;

			/* Update next node */
			nen->prev = prn;
		}

		/* Update length */
		l->length -= 1;
	}


	/*** Return ***/
	return 1;	/* Success */
}


int __fl_list_delete_check(fl_list_t *l, fl_list_node_t **n, unsigned long long p)
{
	/*** Check parameters ***/
	if (l == NULL)
	{
		return 0;	/* No list */
	}

	if (n == NULL)
	{
		return -1;	/* No node */
	}

	if (p >= l->length)
	{
		return -2;	/* Position out of range */
	}


	/*** Return ***/
	return __fl_list_delete_proc(l, n, p);
}


/*** fl_list_push() ***/
int __fl_list_push_proc(fl_list_t *l, fl_list_node_t *n)
{
	/*** Insert ***/
	if (__fl_list_insert_proc(l, n, l->length) != 1)
	{
		return -4;	/* __fl_list_insert_proc() error */
	}


	/*** Return ***/
	return 1;	/* Success */
}


int __fl_list_push_check(fl_list_t *l, fl_list_node_t *n)
{
	/*** Check parameters ***/
	if (l == NULL)
	{
		return -2;	/* No list */
	}

	if (n == NULL)
	{
		return -3;	/* No node */
	}


	/*** Return ***/
	return __fl_list_push_proc(l, n);
}


int fl_list_push(fl_list_t *list, fl_list_node_t *node)
{
	/*** Return ***/
	return __fl_list_push_check(list, node);
}


/*** fl_list_pop() ***/
int __fl_list_pop_proc(fl_list_t *l, fl_list_node_t **n)
{
	/*** Delete ***/
	if (__fl_list_delete_proc(l, n, l->length - 1) != 1)
	{
		return -3;	/* __fl_list_delete_proc() error */
	}


	/*** Return ***/
	return 1;	/* Success */
}


int __fl_list_pop_check(fl_list_t *l, fl_list_node_t **n)
{
	/*** Check parameters ***/
	if (l == NULL)
	{
		return 0;	/* No list */
	}

	if (n == NULL)
	{
		return -1;	/* No node */
	}

	if (l->length == 0)
        {
                return -2;      // ERROR: List is empty
        }


	/*** Return ***/
	return __fl_list_pop_proc(l, n);
}


int fl_list_pop(fl_list_t *list, fl_list_node_t **node)
{
	/*** Return ***/
	return __fl_list_pop_check(list, node);
}


/*** fl_list_unshift() ***/
int __fl_list_unshift_proc(fl_list_t *l, fl_list_node_t *n)
{
	/*** Insert ***/
	if (__fl_list_insert_proc(l, n, 0) != 1)
	{
		return -4;	/* __fl_list_insert_proc() error */
	}


	/*** Return ***/
	return 1;	/* Success */
}


int __fl_list_unshift_check(fl_list_t *l, fl_list_node_t *n)
{
	/*** Check parameters ***/
	if (l == NULL)
	{
		return -2;	/* No list */
	}

	if (n == NULL)
	{
		return -3;	/* No node */
	}


	/*** Return ***/
	return __fl_list_unshift_proc(l, n);
}


int fl_list_unshift(fl_list_t *list, fl_list_node_t *node)
{
	/*** Return ***/
	return __fl_list_unshift_check(list, node);
}


/*** fl_list_shift() ***/
int __fl_list_shift_proc(fl_list_t *l, fl_list_node_t **n)
{
	/*** Delete ***/
	if (__fl_list_delete_proc(l, n, 0) != 1)
	{
		return -3;	/* __fl_list_delete_proc() error */
	}


	/*** Return ***/
	return 1;	/* Success */
}


int __fl_list_shift_check(fl_list_t *l, fl_list_node_t **n)
{
	/*** Check parameters ***/
	if (l == NULL)
	{
		return 0;	/* No list */
	}

	if (n == NULL)
	{
		return -1;	/* No node */
	}

	if (l->length == 0)
	{
		return -2;	// ERROR: List empty
	}


	/*** Return ***/
	return __fl_list_shift_proc(l, n);
}


int fl_list_shift(fl_list_t *list, fl_list_node_t **node)
{
	/*** Return ***/
	return __fl_list_shift_check(list, node);
}


/*** fl_list_radd() ***/
int __fl_list_radd_proc(fl_list_t *l, fl_list_node_t *n, unsigned long long p)
{
	/*** Insert ***/
	if (__fl_list_insert_proc(l, n, l->length - p) != 1)
	{
		return -5;	/* __fl_list_insert_proc() error */
	}


	/*** Return ***/
	return 1;	/* Success */
}


int __fl_list_radd_check(fl_list_t *l, fl_list_node_t *n, unsigned long long p)
{
	/*** Check parameters ***/
	if (l == NULL)
	{
		return -2;	/* No list */
	}

	if (n == NULL)
	{
		return -3;	/* No node */
	}

	if (p > l->length)
	{
		return -4;	/* Position out of range */
	}


	/*** Return ***/
	return __fl_list_radd_proc(l, n, p);
}


int fl_list_radd(fl_list_t *list, fl_list_node_t *node, unsigned long long position)
{
	/*** Return ***/
	return __fl_list_radd_check(list, node, position);
}


/*** fl_list_add() ***/
int __fl_list_add_proc(fl_list_t *l, fl_list_node_t *n, unsigned long long p)
{
	/*** Insert ***/
	if (__fl_list_insert_proc(l, n, p) != 1)
	{
		return -3;	/* __fl_list_insert_proc() error */
	}


	/*** Return ***/
	return 1;	/* Success */
}


int __fl_list_add_check(fl_list_t *l, fl_list_node_t *n, unsigned long long p)
{
	/*** Check parameters ***/
	if (l == NULL)
	{
		return -2;	/* No list */
	}

	if (n == NULL)
	{
		return -3;	/* No node */
	}

	if (p > l->length)
	{
		return -4;	/* Position out of range */
	}


	/*** Return ***/
	return __fl_list_add_proc(l, n, p);
}


int fl_list_add(fl_list_t *list, fl_list_node_t *node, unsigned long long position)
{
	/*** Return ***/
	return __fl_list_add_check(list, node, position);
}


/*** fl_list_rdel() ***/
int __fl_list_rdel_proc(fl_list_t *l, fl_list_node_t **n, unsigned long long p)
{
	/*** Delete ***/
	if (__fl_list_delete_proc(l, n, (l->length - 1) - p) != 1)
	{
		return -5;	/* __fl_list_delete_proc() error */
	}


	/*** Return ***/
	return 1;	/* Success */
}


int __fl_list_rdel_check(fl_list_t *l, fl_list_node_t **n, unsigned long long p)
{
	/*** Check parameters ***/
	if (l == NULL)
	{
		return -2;	/* No list */
	}

	if (n == NULL)
	{
		return -3;	/* No node */
	}

	if (p > (l->length - 1))
	{
		return -4;	/* Position out of range */
	}


	/*** Return ***/
	return __fl_list_rdel_proc(l, n, p);
}


int fl_list_rdel(fl_list_t *list, fl_list_node_t **node, unsigned long long position)
{
	/*** Return ***/
	return __fl_list_rdel_check(list, node, position);
}


/*** fl_list_del() ***/
int __fl_list_del_proc(fl_list_t *l, fl_list_node_t **n, unsigned long long p)
{
	/*** Delete ***/
	if (__fl_list_delete_proc(l, n, p) != 1)
	{
		return -5;	/* __fl_list_delete_proc() error */
	}


	/*** Return ***/
	return 1;	/* Success */
}


int __fl_list_del_check(fl_list_t *l, fl_list_node_t **n, unsigned long long p)
{
	/*** Check parameters ***/
	if (l == NULL)
	{
		return -2;	/* No list */
	}

	if (n == NULL)
	{
		return -3;	/* No node */
	}

	if (p > (l->length - 1))
	{
		return -4;	/* Position out of range */
	}


	/*** Return ***/
	return __fl_list_del_proc(l, n, p);
}


int fl_list_del(fl_list_t *list, fl_list_node_t **node, unsigned long long position)
{
	/*** Return ***/
	return __fl_list_del_check(list, node, position);
}


/*** fl_list_length() ***/
int __fl_list_length_proc(fl_list_t *l, unsigned long long *le)
{
	/*** Get length ***/
	(*le) = l->length;


	/*** Return ***/
	return 1;	/* Success */
}


int __fl_list_length_check(fl_list_t *l, unsigned long long *le)
{
	/*** Check parameters ***/
	if (l == NULL)
	{
		return -2;	/* No list */
	}

	if (le == NULL)
	{
		return -3;	/* No result variable */
	}


	/*** Return ***/
	return __fl_list_length_proc(l, le);
}


int fl_list_length(fl_list_t *list, unsigned long long *length)
{
	/*** Return ***/
	return __fl_list_length_check(list, length);
}


/*** fl_list_rget_node() ***/
int __fl_list_rget_node_proc(fl_list_t *l, fl_list_node_t **n, unsigned long long p)
{
        /*** Declare variables ***/
        fl_list_node_t *pn;
        unsigned long long i;


        /*** Get node ***/
        pn = l->last;
        for (i = 0; i < p; i++)
        {
                pn = pn->prev;
        }

        (*n) = pn;


        /*** Return ***/
        return 1;       /* Success */
}


int __fl_list_rget_node_check(fl_list_t *l, fl_list_node_t **n, unsigned long long p)
{
        /*** Check parameters ***/
        if (l == NULL)
        {
                return -2;      /* No list */
        }

        if (n == NULL)
        {
                return -3;      /* No node */
        }

        if (p >= l->length)
        {
                return -4;      /* Position out of range */
        }


        /*** Return ***/
        return __fl_list_rget_node_proc(l, n, p);
}


int fl_list_rget_node(fl_list_t *list, fl_list_node_t **node, unsigned long long position)
{
        /*** Return ***/
        return __fl_list_rget_node_check(list, node, position);
}


/*** fl_list_get_node() ***/
int __fl_list_get_node_proc(fl_list_t *l, fl_list_node_t **n, unsigned long long p)
{
        /*** Declare variables ***/
	fl_list_node_t *pn;
	unsigned long long i;


	/*** Get node ***/
        pn = l->first;
        for (i = 0; i < p; i++)
        {
                pn = pn->next;
        }

        (*n) = pn;


	/*** Return ***/
	return 1;	/* Success */
}


int __fl_list_get_node_check(fl_list_t *l, fl_list_node_t **n, unsigned long long p)
{
	/*** Check parameters ***/
	if (l == NULL)
	{
		return -2;	/* No list */
	}

	if (n == NULL)
	{
		return -3;	/* No node */
	}

	if (p >= l->length)
	{
		return -4;	/* Position out of range */
	}


	/*** Return ***/
	return __fl_list_get_node_proc(l, n, p);
}


int fl_list_get_node(fl_list_t *list, fl_list_node_t **node, unsigned long long position)
{
	/*** Return ***/
	return __fl_list_get_node_check(list, node, position);
}


/*** fl_list_set_node() ***/
int __fl_list_set_node_proc(fl_list_t *l, fl_list_node_t *n, unsigned long long p)
{
        /*** Declare variables ***/
        fl_list_node_t *pn;


        /*** Set node ***/
        /* Get positional node */
        if (__fl_list_get_node_proc(l, &pn, p) != 1)
        {
                return -5;      /* __fl_list_get_node_proc() error */
        }

        /* Set node */
        pn->data = n->data;
        pn->prev = n->prev;
        pn->next = n->next;


        /*** Return ***/
        return 1;       /* Success */
}


int __fl_list_set_node_check(fl_list_t *l, fl_list_node_t *n, unsigned long long p)
{
        /*** Check parameters ***/
        if (l == NULL)
        {
                return -2;      /* No list */
        }

        if (n == NULL)
        {
                return -3;      /* No node */
        }

        if (p >= l->length)
        {
                return -4;      /* Position out of range */
        }


        /*** Return ***/
        return __fl_list_set_node_proc(l, n, p);
}


int fl_list_set_node(fl_list_t *list, fl_list_node_t *node, unsigned long long position)
{
        /*** Return ***/
        return __fl_list_set_node_check(list, node, position);
}


/*** fl_list_is_empty() ***/
int __fl_list_is_empty_proc(fl_list_t *l, unsigned int *i)
{
	/*** Set result variable ***/
	*i = (l->length == 0);


        /*** Return ***/
        return 1;       /* Success */
}


int __fl_list_is_empty_check(fl_list_t *l, unsigned int *i)
{
	/*** Check parameters ***/
        if (l == NULL)
        {
                return 0;      /* No list */
        }

	if (i == NULL)
	{
		return -1;	/* No result variable */
	}


        /*** Return ***/
        return __fl_list_is_empty_proc(l, i);
}


int fl_list_is_empty(fl_list_t *list, unsigned int *is_empty)
{
        /*** Return ***/
        return __fl_list_is_empty_check(list, is_empty);
}


/*** fl_list_empty() ***/
int __fl_list_empty_proc(fl_list_t *l)
{
        /*** Declare variables ***/
	unsigned long long le;
	unsigned long long i;
        fl_list_node_t *n;


        /*** Empty list ***/
	le = l->length;
	for (i = 0; i < le; i++)
	{
		/* Delete node */
		if (__fl_list_delete_proc(l, &n, 0) != 1)
			return -1;


		/* Free node */
		if (__fl_list_free_node_proc(l, &n) != 1)
			return -2;
	}


        /*** Return ***/
        return 1;       /* Success */
}


int __fl_list_empty_check(fl_list_t *l)
{
        /*** Check parameters ***/
        if (l == NULL)
        {
                return 0;      /* No list */
        }


        /*** Return ***/
        return __fl_list_empty_proc(l);
}


int fl_list_empty(fl_list_t *list)
{
        /*** Return ***/
        return __fl_list_empty_check(list);
}


/*** fl_list_fill() ***/
int __fl_list_fill_proc(fl_list_t *l, unsigned long long n)
{
        /*** Declare variables ***/
	unsigned long long le;
	unsigned long long i;
        fl_list_node_t *no;


	/*** Get length ***/
	le = l->length;


	/*** Fill list ***/
	for (i = 0; i < (n - le); i++)
	{
		if (__fl_list_new_node_proc(l, &no) != 1)
			return -3;

		if (__fl_list_push_proc(l, no) != 1)
			return -4;
	}


        /*** Return ***/
        return 1;       /* Success */
}


int __fl_list_fill_check(fl_list_t *l, unsigned long long n)
{
        /*** Check parameters ***/
        if (l == NULL)
        {
                return 0;	/* No list */
        }

        if (n == 0)
        {
                return -1;	/* No nodes to insert */
        }

	if (n <= l->length)
	{
		return -2;	/* Already filled */
	}


        /*** Return ***/
        return __fl_list_fill_proc(l, n);
}


int fl_list_fill(fl_list_t *list, unsigned long long number)
{
        /*** Return ***/
        return __fl_list_fill_check(list, number);
}


/*** fl_list_is_null() ***/
int __fl_list_is_null_proc(fl_list_t *l, unsigned int *i)
{
        /*** Set result variable ***/
        *i = (l == NULL);


        /*** Return ***/
        return 1;       /* Success */
}


int __fl_list_is_null_check(fl_list_t *l, unsigned int *i)
{
        /*** Check parameters ***/
        if (i == NULL)
        {
                return 0;      /* No result variable */
        }


        /*** Return ***/
        return __fl_list_is_null_proc(l, i);
}


int fl_list_is_null(fl_list_t *list, unsigned int *is_null)
{
        /*** Return ***/
        return __fl_list_is_null_check(list, is_null);
}


/*** fl_list_null() ***/
int __fl_list_null_proc(fl_list_t **l)
{
        /*** Empty list ***/
	switch (__fl_list_empty_proc(*l))
	{
		case -1:
			return -1;
		case -2:
			return -2;
		default:
			break;	/* Success */
	}


	/*** Free list ***/
	if (__fl_list_free_proc(l) != 1)
		return -3;

	*l = NULL;


        /*** Return ***/
        return 1;       /* Success */
}


int __fl_list_null_check(fl_list_t **l)
{
        /*** Check parameters ***/
        if (l == NULL)
        {
                return 0;      /* No list pointer */
        }

	if (*l == NULL)
        {
                return 1;      /* No list (success) */
        }


        /*** Return ***/
        return __fl_list_null_proc(l);
}


int fl_list_null(fl_list_t **list)
{
        /*** Return ***/
        return __fl_list_null_check(list);
}


/*** fl_list_find() ***/
int __fl_list_find_proc(fl_list_t *l,
			    fl_list_node_t *n,
			    char f,
			    unsigned long long p,
			    unsigned long long *pf,
			    unsigned int *i)
{
        /*** Declare variables ***/
	fl_list_node_t *pn;
	unsigned long long j;
        unsigned long long le;
	unsigned int fo;


	/*** Set variables ***/
	*i = 0;
	le = l->length;


	/*** Find ***/
	/* Get positional node */
	if (__fl_list_get_node_proc(l, &pn, p) != 1)
	{
		return -6;	/* __fl_list_get_node_proc() error */
	}

	/* Find */
	switch (f)
	{
		case 'a':
			fo = 0;
			for (j = p; j < le; j++)
			{
				if ((n->data == pn->data) &&
				    (n->prev == pn->prev) &&
				    (n->next == pn->next))
				{
					(*pf) = j;
					fo = 1;
					break;
				}

				pn = pn->next;
			}

			break;

		case 'd':
			fo = 0;
			for (j = p; j < le; j++)
			{
				if (n->data == pn->data)
				{
					(*pf) = j;
					fo = 1;
					break;
				}

				pn = pn->next;
			}

			break;

		case 'p':
			fo = 0;
			for (j = p; j < le; j++)
			{
				if (n->prev == pn->prev)
				{
					(*pf) = j;
					fo = 1;
					break;
				}

				pn = pn->next;
			}

			break;

		case 'n':
			fo = 0;
			for (j = p; j < le; j++)
			{
				if (n->next == pn->next)
				{
					(*pf) = j;
					fo = 1;
					break;
				}

				pn = pn->next;
			}

			break;

		case 'm':
			fo = 0;
			for (j = p; j < le; j++)
			{
				if (n == pn)
				{
					(*pf) = j;
					fo = 1;
					break;
				}

				pn = pn->next;
			}

			break;

		default:
			return -7;	/* Invalid field */
	}


	/*** Set return variable ***/
	*i = fo;


	/*** Return ***/
	return 1;	/* Success */
}


int __fl_list_find_check(fl_list_t *l,
			     fl_list_node_t *n,
			     char f,
			     unsigned long long p,
			     unsigned long long *pf,
			     unsigned int *i)
{
	/*** Check parameters ***/
	if (l == NULL)
	{
		return 0;	/* No list */
	}

	if (n == NULL)
	{
		return -1;	/* No node */
	}

	if ((f != 'a') &&
	    (f != 'd') &&
	    (f != 'p') &&
	    (f != 'n') &&
	    (f != 'm'))
	{
		return -2;	/* Invalid field */
	}

	if (p >= l->length)
	{
		return -3;	/* Position out of range */
	}

	if (pf == NULL)
	{
		return -4;	/* No position result variable */
	}

	if (i == NULL)
	{
		return -5;	/* No find result variable */
	}


	/*** Return ***/
	return __fl_list_find_proc(l, n, f, p, pf, i);
}


int fl_list_find(fl_list_t *list,
		     fl_list_node_t *node,
		     char field,
		     unsigned long long position,
		     unsigned long long *position_found,
		     unsigned int *is_found)
{
	/*** Return ***/
	return __fl_list_find_check(list, node, field, position, position_found, is_found);
}


/*** fl_list_find_node() ***/
int __fl_list_find_node_proc(fl_list_t *l,
				 void *n,
				 unsigned long long p,
				 unsigned long long *pf,
				 unsigned int *i)
{
	int ret;


	ret = __fl_list_find_proc(l, (fl_list_node_t *) n, 'm', p, pf, i);
	if (ret != 1)
		ret = ret + 2;	/* Translate return value */


	/*** Return ***/
	return ret;
}


int __fl_list_find_node_check(fl_list_t *l,
				  void *n,
				  unsigned long long p,
				  unsigned long long *pf,
				  unsigned int *i)
{
        /*** Check parameters ***/
        if (l == NULL)
        {
                return 0;      /* No list */
        }

        if (n == NULL)
        {
                return -1;      /* No node */
        }

        if (p >= l->length)
        {
                return -2;      /* Position out of range */
        }

        if (pf == NULL)
        {
                return -3;      /* No position result variable */
        }

	if (i == NULL)
	{
		return -4;	/* No find result variable */
	}


        /*** Return ***/
        return __fl_list_find_node_proc(l, n, p, pf, i);
}


int fl_list_find_node(fl_list_t *list,
			  void *node,
			  unsigned long long position,
			  unsigned long long *position_found,
			  unsigned int *is_found)
{
	/*** Return ***/
        return __fl_list_find_node_check(list, node, position, position_found, is_found);
}


/*** fl_list_find_node_addr() ***/
int __fl_list_find_node_addr_proc(fl_list_t *l,
                                 void *d,
				 fl_list_node_t *n,
                                 fl_list_node_t *p,
                                 unsigned long long po,
                                 unsigned long long *pf,
				 unsigned int *i)
{
        int ret;
	fl_list_node_t fn;


	fn.data = d;
	fn.next = n;
	fn.prev = p;
        ret = __fl_list_find_proc(l, &fn, 'a', po, pf, i);
        if (ret != 1)
                ret = ret + 2;  /* Translate return value */


        /*** Return ***/
        return ret;
}


int __fl_list_find_node_addr_check(fl_list_t *l,
                                  void *d,
                          	  fl_list_node_t *n,
                          	  fl_list_node_t *p,
                                  unsigned long long po,
                                  unsigned long long *pf,
				  unsigned int *i)
{
        /*** Check parameters ***/
        if (l == NULL)
        {
                return 0;      /* No list */
        }

        if (po >= l->length)
        {
                return -1;      /* Position out of range */
        }

        if (pf == NULL)
        {
                return -2;      /* No position result variable */
        }

	if (i == NULL)
	{
		return -3;	/* No find result variable */
	}


        /*** Return ***/
        return __fl_list_find_node_addr_proc(l, d, n, p, po, pf, i);
}


int fl_list_find_node_addr(fl_list_t *list,
			       void *data,
			       fl_list_node_t *next,
			       fl_list_node_t *prev,
                               unsigned long long position,
                               unsigned long long *position_found,
			       unsigned int *is_found)
{
        /*** Return ***/
        return __fl_list_find_node_addr_check(list, data, next, prev, position, position_found, is_found);
}


/*** fl_list_find_node_data() ***/
int __fl_list_find_node_data_proc(fl_list_t *l,
				      void *d,
				      unsigned long long po,
				      unsigned long long *pf,
				      unsigned int *i)
{
        int ret;
        fl_list_node_t fn;


        fn.data = d;
        fn.next = NULL;
        fn.prev = NULL;
        ret = __fl_list_find_proc(l, &fn, 'd', po, pf, i);
        if (ret != 1)
                ret = ret + 2;  /* Translate return value */


        /*** Return ***/
        return ret;
}


int __fl_list_find_node_data_check(fl_list_t *l,
				       void *d,
				       unsigned long long po,
				       unsigned long long *pf,
				       unsigned int *i)
{
        /*** Check parameters ***/
        if (l == NULL)
        {
                return 0;      /* No list */
        }

        if (po >= l->length)
        {
                return -1;      /* Position out of range */
        }

        if (pf == NULL)
        {
                return -2;      /* No position result variable */
        }

	if (i == NULL)
	{
		return -3;	/* No find result variable */
	}


        /*** Return ***/
        return __fl_list_find_node_data_proc(l, d, po, pf, i);
}


int fl_list_find_node_data(fl_list_t *list,
                               void *data,
                               unsigned long long position,
                               unsigned long long *position_found,
			       unsigned int *is_found)
{
        /*** Return ***/
        return __fl_list_find_node_data_check(list, data, position, position_found, is_found);
}



/*** fl_list_find_node_next() ***/
int __fl_list_find_node_next_proc(fl_list_t *l,
                                      fl_list_node_t *n,
                                      unsigned long long po,
                                      unsigned long long *pf,
				      unsigned int *i)
{
        int ret;
        fl_list_node_t fn;


        fn.data = NULL;
        fn.next = n;
        fn.prev = NULL;
        ret = __fl_list_find_proc(l, &fn, 'n', po, pf, i);
        if (ret != 1)
                ret = ret + 2;  /* Translate return value */


        /*** Return ***/
        return ret;
}


int __fl_list_find_node_next_check(fl_list_t *l,
                                       fl_list_node_t *n,
                                       unsigned long long po,
                                       unsigned long long *pf,
				       unsigned int *i)
{
        /*** Check parameters ***/
        if (l == NULL)
        {
                return 0;      /* No list */
        }

        if (po >= l->length)
        {
                return -1;      /* Position out of range */
        }

        if (pf == NULL)
        {
                return -2;      /* No position result variable */
        }

	if (i == NULL)
	{
		return -3;	/* No find result variable */
	}


        /*** Return ***/
        return __fl_list_find_node_next_proc(l, n, po, pf, i);
}


int fl_list_find_node_next(fl_list_t *list,
                               fl_list_node_t *next,
                               unsigned long long position,
                               unsigned long long *position_found,
			       unsigned int *is_found)
{
        /*** Return ***/
        return __fl_list_find_node_next_check(list, next, position, position_found, is_found);
}


/*** fl_list_find_node_prev() ***/
int __fl_list_find_node_prev_proc(fl_list_t *l,
                                      fl_list_node_t *p,
                                      unsigned long long po,
                                      unsigned long long *pf,
				      unsigned int *i)
{
        int ret;
        fl_list_node_t fn;


        fn.data = NULL;
        fn.next = NULL;
        fn.prev = p;
        ret = __fl_list_find_proc(l, &fn, 'p', po, pf, i);
        if (ret != 1)
                ret = ret + 2;  /* Translate return value */


        /*** Return ***/
        return ret;
}


int __fl_list_find_node_prev_check(fl_list_t *l,
                                       fl_list_node_t *p,
                                       unsigned long long po,
                                       unsigned long long *pf,
				       unsigned int *i)
{
        /*** Check parameters ***/
        if (l == NULL)
        {
                return 0;      /* No list */
        }

        if (po >= l->length)
        {
                return -1;      /* Position out of range */
        }

        if (pf == NULL)
        {
                return -2;      /* No position result variable */
        }

	if (i == NULL)
	{
		return -3;	/* No find result variable */
	}


        /*** Return ***/
        return __fl_list_find_node_prev_proc(l, p, po, pf, i);
}


int fl_list_find_node_prev(fl_list_t *list,
                               fl_list_node_t *prev,
                               unsigned long long position,
                               unsigned long long *position_found,
			       unsigned int *is_found)
{
        /*** Return ***/
        return __fl_list_find_node_prev_check(list, prev, position, position_found, is_found);
}


