

#ifndef __FL_LIST_H__
#define __FL_LIST_H__

/* First include */
#include "fl_conf_id.h"

#if FL_COMP_USER
#include <stdio.h>
#endif


#define FL_LIST_VERSION "0.2.6"
#define FL_LIST_MAXNODES ULONG_MAX


// NOTE: Set mem allocation macros


/* Helper defines */
#define FL_MEM_MALLOC 0
#define FL_MEM_FREE 0
#define FL_MEM_KMALLOC 0
#define FL_MEM_KFREE 0
#define FL_MEM_KMALLOC_OSF1 0
#define FL_MEM_KFREE_OSF1 0
#define FL_OSF1_5_1_B 0
#define FL_USER_OSF1_5_1_B 0
#define FL_KERNEL_OSF1_5_1_B 0
#define FL_IRIX_6_5_30 0
#define FL_USER_IRIX_6_5_30 0
#define FL_KERNEL_IRIX_6_5_30 0


// DEC/COMPAQ/HP OSF1
#if FL_OS_NAME_V_OSF1
#	if FL_OS_VERSION_V_5_1
#		if FL_OS_RELEASE_V_B


#		endif /* FL_OS_RELEASE_V_B */
#	endif /* FL_OS_VERSION_V_5_1 */
#endif /* FL_OS_NAME_V_OSF1 */


// SGI IRIX
#if FL_OS_NAME_V_IRIX
#	if FL_OS_VERSION_V_6_5
#		if FL_OS_RELEASE_V_30

#			if FL_COMP_USER
#				undef FL_MEM_MALLOC
#                               define FL_MEM_MALLOC 1
#				undef FL_MEM_FREE
#                               define FL_MEM_FREE 1
#				undef FL_IRIX_6_5_30
#				define FL_IRIX_6_5_30 1
#				undef FL_USER_IRIX_6_5_30
#				define FL_USER_IRIX_6_5_30 1
#			elif FL_COMP_KERNEL
#				undef FL_MEM_KMALLOC
#                               define FL_MEM_KMALLOC 1
#				undef FL_MEM_KFREE
#                               define FL_MEM_KFREE 1
#				undef FL_IRIX_6_5_30
#				define FL_IRIX_6_5_30 1
#				undef FL_KERNEL_IRIX_6_5_30
#				define FL_KERNEL_IRIX_6_5_30 1
#			endif

#		endif /* FL_OS_RELEASE_V_30 */
#	endif /* FL_OS_VERSION_V_6_5 */
#endif /* FL_OS_NAME_V_IRIX */


// NOTE: OS specific include headers


#if FL_KERNEL_IRIX_6_5_30


#include <sys/types.h>


#endif


struct fl_list_conf_s
{
	// NOTE: Pointer to allocator/deallocator


#if FL_MEM_MALLOC


	void *(*malloc)(size_t);


#endif
#if FL_MEM_FREE


	void (*free)(void *);


#endif
#if FL_MEM_KMALLOC_OSF1


#endif
#if FL_MEM_KFREE_OSF1


#endif
#if FL_KERNEL_IRIX_6_5_30


	void *(*malloc)(size_t, int);
	void (*free)(void *, size_t);


#endif


};

typedef struct fl_list_conf_s fl_list_conf_t;


struct fl_list_node_s
{
	void *data;
	struct fl_list_node_s *next;
	struct fl_list_node_s *prev;
};

typedef struct fl_list_node_s fl_list_node_t;


struct fl_list_s
{
	unsigned long long length;
	fl_list_conf_t conf;
	fl_list_node_t *first;
	fl_list_node_t *last;
	void *private_data;
};

typedef struct fl_list_s fl_list_t;


/*
	PUBLIC PROTOTYPES
*/


int fl_list_init(fl_list_conf_t *conf);
int fl_list_deinit(void);
int fl_list_new(fl_list_t **list);
int fl_list_free(fl_list_t **list);
int fl_list_new_node(fl_list_t *list, fl_list_node_t **node);
int fl_list_free_node(fl_list_t *list, fl_list_node_t **node);
int fl_list_push(fl_list_t *list, fl_list_node_t *node);
int fl_list_pop(fl_list_t *list, fl_list_node_t **node);
int fl_list_unshift(fl_list_t *list, fl_list_node_t *node);
int fl_list_shift(fl_list_t *list, fl_list_node_t **node);
int fl_list_radd(fl_list_t *list, fl_list_node_t *node, unsigned long long position);
int fl_list_add(fl_list_t *list, fl_list_node_t *node, unsigned long long position);
int fl_list_rdel(fl_list_t *list, fl_list_node_t **node, unsigned long long position);
int fl_list_del(fl_list_t *list, fl_list_node_t **node, unsigned long long position);
int fl_list_length(fl_list_t *list, unsigned long long *length);
int fl_list_rget_node(fl_list_t *list, fl_list_node_t **node, unsigned long long position);
int fl_list_get_node(fl_list_t *list, fl_list_node_t **node, unsigned long long position);
int fl_list_set_node(fl_list_t *list, fl_list_node_t *node, unsigned long long position);
int fl_list_is_empty(fl_list_t *list, unsigned int *is_empty);
int fl_list_empty(fl_list_t *list);
int fl_list_fill(fl_list_t *list, unsigned long long number);
int fl_list_is_null(fl_list_t *list, unsigned int *is_null);
int fl_list_null(fl_list_t **list);

int fl_list_find(fl_list_t *list,
                     fl_list_node_t *node,
                     char field,
                     unsigned long long position,
                     unsigned long long *position_found,
                     unsigned int *is_found);

int fl_list_find_node(fl_list_t *list,
                          void *node,
                          unsigned long long position,
                          unsigned long long *position_found,
                          unsigned int *is_found);

int fl_list_find_node_addr(fl_list_t *list,
                               void *data,
                               fl_list_node_t *next,
                               fl_list_node_t *prev,
                               unsigned long long position,
                               unsigned long long *position_found,
                               unsigned int *is_found);

int fl_list_find_node_data(fl_list_t *list,
                               void *data,
                               unsigned long long position,
                               unsigned long long *position_found,
                               unsigned int *is_found);

int fl_list_find_node_next(fl_list_t *list,
                               fl_list_node_t *next,
                               unsigned long long position,
                               unsigned long long *position_found,
                               unsigned int *is_found);

int fl_list_find_node_prev(fl_list_t *list,
                               fl_list_node_t *prev,
                               unsigned long long position,
                               unsigned long long *position_found,
                               unsigned int *is_found);


#endif /* __FL_LIST_H__ */


