

// NOTE: System header files


#if FL_OS_NAME_V_OSF1
#       if FL_OS_VERSION_V_5_1
#               if FL_OS_RELEASE_V_B


#               endif
#       endif
#elif FL_OS_NAME_V_IRIX
#       if FL_OS_VERSION_V_6_5
#               if FL_OS_RELEASE_V_30





#               endif
#       endif
#endif


// NOTE: Local header files


#include "fl_mem.h"


// NOTE: External definitions


// NOTE: Global variables


// Mem list
static fl_list_t *mem_list;
 
 
// Mem counter
static fl_int64_t malloc_count = 0;


// NOTE: Locking
//
//  Some platforms (as for example IRIX), can't hold spinlocks when
//  calling the system memory allocator. In this case, a mutex is used.
//
//  This also means that locking throughout the mem library must be
//  done directly.


#if FL_OS_NAME_V_OSF1
#       if FL_OS_VERSION_V_5_1
#               if FL_OS_RELEASE_V_B


#               endif
#       endif
#elif FL_OS_NAME_V_IRIX
#       if FL_OS_VERSION_V_6_5
#               if FL_OS_RELEASE_V_30


// Mem list lock
mutex_t *mem_list_lock;


#               endif
#       endif
#endif


// NOTE: Public functions


int fl_mem_init(void)
{
        // NOTE: Mem list
        //
        //  The mem list contains an entry for each memory allocation done
	//  using fl_mem_alloc().


        // Create mem list
	fl_list_new(&mem_list);


#if FL_OS_NAME_V_OSF1
#       if FL_OS_VERSION_V_5_1
#               if FL_OS_RELEASE_V_B


#               endif
#       endif   
#elif FL_OS_NAME_V_IRIX
#       if FL_OS_VERSION_V_6_5
#               if FL_OS_RELEASE_V_30


	// Initialize lock for mem list
	mem_list_lock = MUTEX_ALLOC(MUTEX_DEFAULT,
                                    KM_SLEEP,
                                    NULL);


#               endif
#       endif   
#endif  


        return 1;       // SUCCESS
}


int fl_mem_deinit(void)
{
        // Free mem list
        fl_list_free(&mem_list);


#if FL_OS_NAME_V_OSF1
#       if FL_OS_VERSION_V_5_1
#               if FL_OS_RELEASE_V_B


#               endif
#       endif   
#elif FL_OS_NAME_V_IRIX
#       if FL_OS_VERSION_V_6_5
#               if FL_OS_RELEASE_V_30


	// Terminate lock for mem list
	MUTEX_DESTROY(mem_list_lock);


#               endif
#       endif   
#endif  


        return 1;       // SUCCESS
}


int fl_mem_lock(void)
{


#if FL_OS_NAME_V_OSF1
#       if FL_OS_VERSION_V_5_1
#               if FL_OS_RELEASE_V_B

        
#               endif
#       endif
#elif FL_OS_NAME_V_IRIX 
#       if FL_OS_VERSION_V_6_5
#               if FL_OS_RELEASE_V_30


        MUTEX_LOCK(mem_list_lock,
		   -1);
        

#               endif
#       endif
#endif


        return 1;       // SUCCESS
}


int fl_mem_unlock(void)
{


#if FL_OS_NAME_V_OSF1
#       if FL_OS_VERSION_V_5_1
#               if FL_OS_RELEASE_V_B


#               endif
#       endif
#elif FL_OS_NAME_V_IRIX
#       if FL_OS_VERSION_V_6_5
#               if FL_OS_RELEASE_V_30


	MUTEX_UNLOCK(mem_list_lock);


#               endif
#       endif
#endif


        return 1;       // SUCCESS
}


int fl_mem_alloc(void **data,		// -- OUT
                 fl_uint3264_t size,	// -- IN
                 fl_uint32_t flags,	// -- IN
                 char *function,	// -- IN
                 char *purpose)		// -- IN
{
        char *data_pointer;
        fl_uint3264_t data_index;
        fl_uint64_t *data_size;
        fl_list_node_t *mem_node;
        struct mem_track_s *mem_track;
        fl_uint32_t malloc_flags;


	// NOTE: Arguments


        // OUT: Check for data
        if (data == NULL)
        {
                return 0;       // ERROR: No data
        }


        // IN: Check for non-zero size
        if (size == 0)
        {
                return -1;      // ERROR: No non-zero size
        }


	// TODO: IN: Check for valid switches


	// IN: Check for function name
	if (function == NULL)
	{
		return -2;	// ERROR: No function name
	}


	// IN: Check for purpose string
	if (purpose == NULL)
	{
		return -3;	// ERROR: No purpose name
	}


	// NOTE: Processing


        // Reset memory flags
        malloc_flags = 0;


#if FL_OS_NAME_V_OSF1
#       if FL_OS_VERSION_V_5_1
#               if FL_OS_RELEASE_V_B


#               endif
#       endif
#endif


        if ((flags & FL_MEM_WAITOK) > 0)
        {


#if FL_OS_NAME_V_OSF1
#       if FL_OS_VERSION_V_5_1
#               if FL_OS_RELEASE_V_B


#               endif
#       endif
#elif FL_OS_NAME_V_IRIX
#       if FL_OS_VERSION_V_6_5
#               if FL_OS_RELEASE_V_30


                malloc_flags |= KM_SLEEP;


#               endif
#       endif
#endif


        }

        if ((flags & FL_MEM_NOWAIT) > 0)
        {


#if FL_OS_NAME_V_OSF1
#       if FL_OS_VERSION_V_5_1
#               if FL_OS_RELEASE_V_B


#               endif
#       endif
#elif FL_OS_NAME_V_IRIX
#       if FL_OS_VERSION_V_6_5
#               if FL_OS_RELEASE_V_30


                malloc_flags |= KM_NOSLEEP;


#               endif
#       endif
#endif


        }


        if ((flags & FL_MEM_CACHEALIGN) > 0)
        {


#if FL_OS_NAME_V_IRIX
#       if FL_OS_VERSION_V_6_5
#               if FL_OS_RELEASE_V_30


                malloc_flags |= KM_CACHEALIGN;


#               endif
#       endif
#endif


        }


#if FL_OS_NAME_V_OSF1
#       if FL_OS_VERSION_V_5_1
#               if FL_OS_RELEASE_V_B


#               endif
#       endif
#elif FL_OS_NAME_V_IRIX
#       if FL_OS_VERSION_V_6_5
#               if FL_OS_RELEASE_V_30


        // NOTE: IRIX memory freeing
        //
        //  In IRIX kmem_free(), a second parameter must be specified,
        //  indicating the exact size of the data segment pointed to
        //  by the first parameter.
        //
        //  This requires carrying around the data size for each allocated
        //  data segment until freeing the data.
        //
        //  The approach taken to provide this data size, is to enlarge the
        //  allocated buffer by 8 bytes, store the size (as a ulong), and
        //  displacing by 8 bytes to get the final data address.


        // Allocate memory
	*data = kmem_alloc(size + 8,
			   malloc_flags);
	if (*data == NULL)


#               endif
#       endif
#endif
        {
                return -2;      // ERROR: Couldn't allocate memory
        }


        // NOTE: Buffer postprocessing


#if FL_OS_NAME_V_IRIX
#       if FL_OS_VERSION_V_6_5
#               if FL_OS_RELEASE_V_30


        // Set data pointer
        data_pointer = (char *) *data;


        //FL_DEBUG_UPRINTF((
			  //"[fl_mem_alloc] INFO: Allocated data_pointer = %p\n",
			  //data_pointer))


        // Store data size
        data_size = (fl_uint64_t *) data_pointer;
        *data_size = size + 8;


        //FL_DEBUG_UPRINTF((\
			  //"[fl_mem_alloc] INFO: Data_size = %p -> %llu\n",
			  //data_size,
			  //*data_size))


        // Displace data pointer
        data_pointer = data_pointer + 8;


        // Set displaced data
        *data = (void *) data_pointer;


        //FL_DEBUG_UPRINTF((\
			  //"[fl_mem_alloc] INFO: Displaced data_pointer = %p\n",
			  //data_pointer))


        // Check for zero flag
        if ((flags & FL_MEM_ZERO) > 0)
        {
		//FL_DEBUG_UPRINTF((\
				  //"[fl_mem_alloc] INFO: Data_index = 0 to data_size = %llu\n",
				  //size))


                for (data_index = 0;
		     data_index < size;
		     data_index++)
                {
                        data_pointer[data_index] = '\0';
                }
        }


#               endif
#       endif
#endif


#ifdef FL_MEM_TRACK


        // Increment allocation counter
        malloc_count++;


        // Acquire lock for mem list
	fl_mem_lock();


        // Allocate mem node
        fl_list_new_node(mem_list,
			 &mem_node);


        // Allocate mem track
        mem_track = NULL;
        if (fl_mem_alloc_notrack((void **) &mem_track,
                                 sizeof(struct mem_track_s),
                                 FL_MEM_NULL) < 1)
        {
                // Free mem node
                fl_list_free_node(mem_list, &mem_node);


        	// Release lock for mem list
		fl_mem_unlock();


                return -3;      // ERROR: Couldn't allocate mem tracker
        }


        // Populate mem tracker
        mem_track->data = data_pointer;
        mem_track->function = function;
        mem_track->purpose = purpose;


        // Set mem node data field
        mem_node->data = (void *) mem_track;


        // Insert mem tracker
        fl_list_push(mem_list,
		     mem_node);


       	// Release lock for mem list
	fl_mem_unlock();


#endif  // FL_MEM_TRACK


        return 1;       // SUCCESS
}


int fl_mem_alloc_notrack(void **data,
                         fl_uint3264_t size,
                         fl_uint32_t flags)
{
        char *data_pointer;
        fl_uint64_t *data_size;
        fl_uint3264_t data_index;
        fl_uint32_t malloc_flags;


	// NOTE: Arguments


        // OUT: Check for data
        if (data == NULL)
        {
                return 0;       // ERROR: No data
        }


        // IN: Check for non-zero size
        if (size == 0)
        {
                return -1;      // ERROR: No non-zero size
        }


        // Set memory flags
        //if (flags == FL_MEM_ZERO)
        //{
                //flags = M_WAITOK | M_ZERO;
        //}
        //else
        //{
                //flags = M_WAITOK;
        //}


	// NOTE: Processing


        // Reset memory flags
        malloc_flags = 0;


#if FL_OS_NAME_V_OSF1
#       if FL_OS_VERSION_V_5_1
#               if FL_OS_RELEASE_V_B


#               endif
#       endif
#endif


        if ((flags & FL_MEM_WAITOK) > 0)
        {


#if FL_OS_NAME_V_OSF1
#       if FL_OS_VERSION_V_5_1
#               if FL_OS_RELEASE_V_B


#               endif
#       endif
#elif FL_OS_NAME_V_IRIX
#       if FL_OS_VERSION_V_6_5
#               if FL_OS_RELEASE_V_30

                malloc_flags |= KM_SLEEP;


#               endif
#       endif
#endif


        }


        if ((flags & FL_MEM_NOWAIT) > 0)
        {


#if FL_OS_NAME_V_OSF1
#       if FL_OS_VERSION_V_5_1
#               if FL_OS_RELEASE_V_B


#               endif
#       endif
#elif FL_OS_NAME_V_IRIX
#       if FL_OS_VERSION_V_6_5
#               if FL_OS_RELEASE_V_30


                malloc_flags |= KM_NOSLEEP;


#               endif
#       endif
#endif


        }


#if FL_OS_NAME_V_OSF1
#       if FL_OS_VERSION_V_5_1
#               if FL_OS_RELEASE_V_B


#               endif
#       endif
#elif FL_OS_NAME_V_IRIX
#       if FL_OS_VERSION_V_6_5
#               if FL_OS_RELEASE_V_30


        // Allocate memory
        if ((*data = kmem_alloc(size + 8,
				malloc_flags)) == NULL)


#               endif
#       endif
#endif
        {
                return -2;      // ERROR: Couldn't allocate memory
        }


        // NOTE: Buffer postprocessing


#if FL_OS_NAME_V_IRIX
#       if FL_OS_VERSION_V_6_5
#               if FL_OS_RELEASE_V_30


        // Set data pointer
        data_pointer = (char *) *data;


        //FL_DEBUG_UPRINTF((\
			  //"[fl_mem_alloc_notrack] INFO: Allocated data_pointer = %p\n",
			  //data_pointer))


        // Store data size
        data_size = (fl_uint64_t *) data_pointer;
        *data_size = size + 8;


        //FL_DEBUG_UPRINTF((\
			  //"[fl_mem_alloc_notrack] INFO: Data_size = %p -> %llu\n",
			  //data_size,
			  //*data_size))


        // Displace data pointer
        data_pointer = data_pointer + 8;


        // Set displaced data
        *data = (void *) data_pointer;


        //FL_DEBUG_UPRINTF((\
			  //"[fl_mem_alloc_notrack] INFO: Displaced data_pointer = %p\n",
			  //data_pointer))


        // Check for zero flag
        if ((flags & FL_MEM_ZERO) > 0)
        {
                //FL_DEBUG_UPRINTF((\
				  //"[fl_mem_alloc_notrack] INFO: Data_index = 0 to data_size = %llu\n",
				  //size))


                for (data_index = 0;
		     data_index < size;
		     data_index++)
                {
                        data_pointer[data_index] = '\0';
                }
        }


#               endif
#       endif
#endif


        return 1;       // SUCCESS
}


int fl_mem_free(void *data)	// -- IN
{
        char *data_pointer;
        fl_uint32_t is_found;
        fl_uint64_t *data_size;
        fl_list_node_t *mem_node;
        struct mem_track_s *mem_track;
        fl_uint64_t mem_index;
        fl_uint64_t mem_length;


	// NOTE: Arguments


        // OUT: Check for data
        if (data == NULL)
        {
                return 0;       // ERROR: No data
        }


#if FL_OS_NAME_V_OSF1
#       if FL_OS_VERSION_V_5_1
#               if FL_OS_RELEASE_V_B


#               endif
#       endif
#elif FL_OS_NAME_V_IRIX
#       if FL_OS_VERSION_V_6_5
#               if FL_OS_RELEASE_V_30


	// Set data pointer
	data_pointer = (char *) data;


	//FL_DEBUG_UPRINTF((\
			  //"[fl_mem_free] INFO: Displaced data_pointer = %p\n",
			  //data_pointer))


	// Displace data pointer
	data_pointer = data_pointer - 8;


	//FL_DEBUG_UPRINTF((\
			  //"[fl_mem_free] INFO: Allocated data_pointer = %p\n",
			  //data_pointer))


	// Get data size
	data_size = (fl_uint64_t *) data_pointer;


	//FL_DEBUG_UPRINTF((\
			  //"[fl_mem_free] INFO: Data_size = %p -> %llu\n",
			  //data_size,
			  //*data_size))


	kmem_free((void *) data_pointer,
		  *data_size);


#               endif
#       endif
#endif


#ifdef FL_MEM_TRACK


        malloc_count--;


        // Acquire lock for mem list
	fl_mem_lock();


        // Get length of mem list
        fl_list_length(mem_list,
		       &mem_length);


        // Traverse mem list
        is_found = 0;
        for (mem_index = 0;
	     mem_index < mem_length;
	     mem_index++)
        {
                // Get next node
                fl_list_get_node(mem_list,
				 &mem_node,
				 mem_index);


                // Get mem track
                mem_track = (struct mem_track_s *) mem_node->data;


                // Check for data
                if (mem_track->data == data)
                {
                        is_found = 1;
                        break;
                }
        }


        // Check if found
        if (is_found == 1)
        {
                //FL_DEBUG_UPRINTF((\
				  //"[fl_mem_free] INFO: Mem_track found at %llu\n",
				  //mem_index))


                // Unlink mem node
                fl_list_del(mem_list,
			    &mem_node,
			    mem_index);


                // Free mem node
                fl_list_free_node(mem_list,
				  &mem_node);


                // Free mem track
                if (fl_mem_free_notrack(mem_track) < 1)
                {
                        // NOTE: Release lock for mem list


        		// Release lock for mem list
			fl_mem_unlock();


                        return -1;      // ERROR: Couldn't free mem track
                }
        }


       	// Release lock for mem list
	fl_mem_unlock();


#endif  // FL_MEM_TRACK


        return 1;       // SUCCESS
}


int fl_mem_free_notrack(void *data)	// -- IN
{
        char *data_pointer;
        fl_uint64_t *data_size;


	// NOTE: Arguments


        // IN: Check for data
        if (data == NULL)
        {
                return 0;       // ERROR: No data
        }


	// NOTE: Processing


#if FL_OS_NAME_V_OSF1
#       if FL_OS_VERSION_V_5_1
#               if FL_OS_RELEASE_V_B


#               endif
#       endif
#elif FL_OS_NAME_V_IRIX
#       if FL_OS_VERSION_V_6_5
#               if FL_OS_RELEASE_V_30


	// Set data pointer
	data_pointer = (char *) data;


	//FL_DEBUG_UPRINTF((\
			  //"[fl_mem_free_notrack] INFO: Displaced data_pointer = %p\n",
			  //data_pointer))


	// Displace data pointer
	data_pointer = data_pointer - 8;


	//FL_DEBUG_UPRINTF((\
			  //"[fl_mem_free_notrack] INFO: Allocated data_pointer = %p\n",
			  //data_pointer))


	// Get data size
	data_size = (fl_uint64_t *) data_pointer;


	//FL_DEBUG_UPRINTF((\
			  //"[fl_mem_free_notrack] INFO: Data_size = %p -> %llu\n",
			  //data_size,
			  //*data_size))


	kmem_free((void *) data_pointer,
		  *data_size);


#               endif
#       endif
#endif


        return 1;       // SUCCESS
}


int fl_mem_dump()
{
        fl_uint64_t mem_index;
        fl_uint64_t mem_length;
        fl_list_node_t *mem_node;
        struct mem_track_s *mem_track;


        // Acquire lock for mem list
	fl_mem_lock();


        // Get length of mem list
        fl_list_length(mem_list,
		       &mem_length);


        FL_DEBUG_UPRINTF((\
			  "[fl_mem_dump] INFO: Mem_list length = %llu\n",
			  mem_length))


        // Traverse mem list
        for (mem_index = 0;
	     mem_index < mem_length;
	     mem_index++)
        {
                // Get next node
                fl_list_get_node(mem_list,
				 &mem_node,
				 mem_index);


                // Get mem track
                mem_track = (struct mem_track_s *) mem_node->data;


                // NOTE: Print mem track


                FL_UPRINTF((\
			    "[fl_mem_dump] INFO: Function: %s\n",
			    mem_track->function))


                FL_UPRINTF((\
			    "[fl_mem_dump] INFO: Purpose: %s\n",
			    mem_track->purpose))
        }


        // Release lock for mem list
	fl_mem_unlock();


        return 1;       // SUCCESS
}


int fl_mem_dump_count()
{
	// Print mem allocation counter
	FL_UPRINTF((\
		    "[fl_mem_dump] INFO: malloc_count = %lld\n",
		    malloc_count))


	return 1;	// SUCCESS
}


