

// NOTE: System header files


#if FL_OS_NAME_V_OSF1
#       if FL_OS_VERSION_V_5_1
#               if FL_OS_RELEASE_V_B


#               endif
#       endif
#elif FL_OS_NAME_V_IRIX
#       if FL_OS_VERSION_V_6_5
#               if FL_OS_RELEASE_V_30


#               endif
#       endif
#endif


// NOTE: Local header files


#include "fl_proc.h"
#include "fl_spin.h"
#include "fl_test.h"


// NOTE: Global declarations


FL_TEST_DECLS


// NOTE: External variables


#if FL_OS_NAME_V_OSF1
#       if FL_OS_VERSION_V_5_1
#               if FL_OS_RELEASE_V_B


#               endif
#       endif
#elif FL_OS_NAME_V_IRIX
#       if FL_OS_VERSION_V_6_5
#               if FL_OS_RELEASE_V_30


extern pid_t            pid_base;
extern int              pidtabsz;
extern pid_slot_t       *pidtab;
extern sv_t             pid_dealloc_sv;
extern pid_active_t     pid_active_list;


extern int vproc_reference(vproc_t *, int);
extern void vproc_release(vproc_t *);
extern int sv_broadcast(sv_t *);


#               endif
#       endif
#endif


// NOTE: Global variables


#if FL_OS_NAME_V_IRIX
#       if FL_OS_VERSION_V_6_5
#               if FL_OS_RELEASE_V_30
 
 
fl_int32_t proc_pl;


#               endif
#       endif
#endif


// Proc list
static fl_list_t *proc_list;


// Lock for proc list
struct fl_spin_s *proc_list_lock;


// NOTE: Public functions


int fl_proc_init()
{
	// NOTE: Proc list
	//
	//  The proc list contains the pids of user processes that entered
	//  the kernel module by calling an entry point.


	// Create proc list
	fl_list_new(&proc_list);


	// Create lock for proc list
	fl_spin_new(&proc_list_lock);


	return 1;	// SUCCESS
}


int fl_proc_deinit()
{
        // Free proc list
	fl_list_free(&proc_list);


	// Terminate lock for proc list
        fl_spin_free(proc_list_lock);


        return 1;       // SUCCESS
}


int fl_proc_lock()
{
	fl_spin_lock(proc_list_lock);


	return 1;	// SUCCESS
}


int fl_proc_unlock()
{
	fl_spin_unlock(proc_list_lock);


	return 1;	// SUCCESS
}


int fl_proc_new(struct proc_s **proc)	// -- OUT
{
        pid_t pid;
        struct proc_s *proc_loc;


        // NOTE: Arguments


	// OUT: Check for proc
        if (proc == NULL)
        {
                return 0;       // ERROR: No proc
        }


	// NOTE: Processing


        // Get pid of current process
	fl_proc_get_data(FL_DATA_PID,
			 &pid);


        // Allocate proc
        fl_mem_alloc((void **) &proc_loc,
                     sizeof(struct proc_s),
                     FL_MEM_WAITOK,
                     "fl_proc_new",
                     "Allocate proc");


        // Set default proc fields
        proc_loc->pid = pid;


        // Set result variable
        *proc = proc_loc;


        return 1;       // SUCCESS
}


int fl_proc_free(struct proc_s *proc)	// -- IN
{
        // NOTE: Arguments


	// IN: Check for proc
        if (proc == NULL)
        {
                return 0;       // ERROR: No proc
        }


	// NOTE: Processing


	// Free proc
	fl_mem_free(proc);


	return 1;	// SUCCESS
}


int fl_proc_get_data(fl_uint32_t data_bitmap,	// -- IN
		     pid_t *pid)		// -- OUT
{


#if FL_OS_NAME_V_IRIX
#       if FL_OS_VERSION_V_6_5
#               if FL_OS_RELEASE_V_30


	fl_uint3264_t pid_1;


#               endif
#       endif
#endif


	// NOTE: Arguments


	// IN: Check for valid data bitmap
	if ((data_bitmap & ~FL_DATA_MASK) != 0)
	{
		return 0;	// ERROR: Invalid data bitmap
	}


	// OUT: Check for pid
	if (pid == NULL)
	{
		return -1;	// ERROR: No pid
	}


	// NOTE: Initialization


	// Pid
	*pid = 0;


	// NOTE: Processing


#if FL_OS_NAME_V_OSF1
#       if FL_OS_VERSION_V_5_1
#               if FL_OS_RELEASE_V_B


#               endif
#       endif
#elif FL_OS_NAME_V_IRIX
#       if FL_OS_VERSION_V_6_5
#               if FL_OS_RELEASE_V_30



#               endif
#       endif
#endif


#if FL_OS_NAME_V_OSF1
#       if FL_OS_VERSION_V_5_1
#               if FL_OS_RELEASE_V_B


#               endif
#       endif
#elif FL_OS_NAME_V_IRIX
#       if FL_OS_VERSION_V_6_5
#               if FL_OS_RELEASE_V_30


        // Check for pid request
        if ((data_bitmap & FL_DATA_PID) != 0)
        {
		// NOTE: drv_getparm()
		//
		//  The drv_getparm() function only fails if
		//  a wrong parameter was passed in.


		// Get pid of current user context
		drv_getparm(PPID,
			    &pid_1);


		// Set pid
		*pid = (pid_t) pid_1;
        }


#               endif
#       endif
#endif



#if FL_OS_NAME_V_OSF1
#       if FL_OS_VERSION_V_5_1
#               if FL_OS_RELEASE_V_B


#               endif
#       endif
#elif FL_OS_NAME_V_IRIX
#       if FL_OS_VERSION_V_6_5
#               if FL_OS_RELEASE_V_30


#               endif
#       endif
#endif


        return 1;       // SUCCESS
}


int fl_proc_get(struct proc_s **proc)	// -- OUT
{
        // NOTE: Arguments


	// IN: Check for proc
        if (proc == NULL)
        {
                return 0;       // ERROR: No proc
        }


	// NOTE: Processing


	// Acquire lock for proc list
	fl_proc_lock();


	if (fl_proc_get_unlocked(proc) < 1)
	{
		// Release lock for proc list
		fl_proc_unlock();


                return -1;      // ERROR: Couldn't get proc
        }


	// Release lock for proc list
	fl_proc_unlock();


        return 1;       // SUCCESS
}


int fl_proc_get_unlocked(struct proc_s **proc)	// -- OUT
{
        pid_t pid;
        fl_uint32_t is_found;
        fl_uint64_t proc_index;
        fl_uint64_t proc_length;
        fl_list_node_t *proc_node;


        FL_TEST_TYPE_DECL


	// NOTE: Arguments


	// OUT: Check for proc
        if (proc == NULL)
        {
                return 0;       // ERROR: No proc
        }


        // NOTE: Initialization


        FL_TEST_TYPE_INIT


	// NOTE: Processing


	// Check for proc list
        if (proc_list == NULL)
        {
                return -1;       // ERROR: No proc list
        }


        // Get pid of current process
	fl_proc_get_data(FL_DATA_PID,
			 &pid);


	FL_DEBUG_UPRINTF((\
			  "[fl_proc_get_unlocked] INFO: pid = %u\n",
			  pid))


        // Get length of proc list
        fl_list_length(proc_list,
		       &proc_length);


        // Traverse proc list
        is_found = 0;
        for (proc_index = 0;
	     proc_index < proc_length;
	     proc_index++)
        {
                // Get next proc node
                fl_list_get_node(proc_list,
				 &proc_node,
				 proc_index);


                // Get proc
                *proc = (struct proc_s *) proc_node->data;


		FL_DEBUG_UPRINTF((\
				  "[fl_proc_get_unlocked] INFO: (*proc)->pid = %u == pid = %u\n",
				  (*proc)->pid,
				  pid))


                // Compare pids
                if ((*proc)->pid == pid)
                {
                        is_found = 1;


                        break;
                }
        }


        FL_TEST_F_fl_proc_get_N_NOT_FOUND_PID(0,\
					      1)


        // Check if pid was not found
        if (FL_TEST_COND((fl_test_ret == 1), ||)
            (is_found != 1))
        {
                FL_TEST_F_fl_proc_get_N_NOT_FOUND_PID(1, 1)


                FL_TEST("fl_proc_get", "NOT_FOUND_PID")
                return -1;      // ERROR: Couldn't find pid
        }


        return 1;       // SUCCESS
}


int fl_proc_push(struct proc_s *proc)	// -- IN
{
        // NOTE: Arguments


	// IN: Check for proc
        if (proc == NULL)
        {
                return 0;       // ERROR: No proc
        }


	// NOTE: Processing


        // Acquire lock for proc list
	fl_proc_lock();


	if (fl_proc_push_unlocked(proc) < 1)
        {
		// Release lock for proc list
		fl_proc_unlock();


                return -1;      // ERROR: Couldn't get proc
        }


	// Release lock for proc list
	fl_proc_unlock();


        return 1;       // SUCCESS
}


int fl_proc_push_unlocked(struct proc_s *proc)	// -- IN
{
        fl_list_node_t *proc_node;


        // NOTE: Arguments


	// IN: Check for proc
        if (proc == NULL)
        {
                return 0;       // ERROR: No proc
        }


	// NOTE: Processing


        // Create new proc node
        proc_node = NULL;
	fl_list_new_node(proc_list,
			 &proc_node);


        // Set node data
        proc_node->data = (void *) proc;


        // Push node into list
        fl_list_push(proc_list,
		     proc_node);


        return 1;       // SUCCESS
}


int fl_proc_shift(struct proc_s **proc)	// -- OUT
{
        // NOTE: Arguments


	// OUT: Check for proc
        if (proc == NULL)
        {
                return 0;       // ERROR: No proc
        }


	// NOTE: Processing


	// Acquire lock for proc list
	fl_proc_lock();


	if (fl_proc_shift_unlocked(proc) < 1)
	{
		// Release lock for proc list
		fl_proc_unlock();


		return -1;      // ERROR: Couldn't get proc
	}


	// Release lock for proc list
	fl_proc_unlock();


	return 1;       // SUCCESS
}


int fl_proc_shift_unlocked(struct proc_s **proc)	// -- OUT
{
	fl_int32_t ret;
	fl_list_node_t *proc_node;


	// NOTE: Arguments


	// OUT: Check for proc
	if (proc == NULL)
	{
		return 0;	// ERROR: No proc
	}


	// NOTE: Initialization


	// Proc
	*proc = NULL;


	// NOTE: Processing


	// Shift proc list
	ret = fl_list_shift(proc_list,
			    &proc_node);
	if (ret >= 1)
	{
		// Set proc
		*proc = (struct proc_s *) proc_node->data;


		// Free proc node
		fl_list_free_node(proc_list,
				  &proc_node);
	}


	return ret;
}


int fl_proc_get_index(struct proc_s **proc,	// -- OUT
		      fl_uint64_t proc_index)	// -- IN
{
	// NOTE: Arguments


	// OUT: Check for proc
	if (proc == NULL)
	{
		return 0;	// ERROR: No proc
	}


	// NOTE: Initialization


	// Proc
	*proc = NULL;


	// NOTE: Processing


        // Acquire lock for proc list
	fl_proc_lock();


	switch (fl_proc_get_index_unlocked(proc,
					   proc_index))
	{
		case 0:

			// NOTE: Never reached


			break;

		case -1:

			// Release lock for proc list
			fl_proc_unlock();


			return -1;	// ERROR: Position out of range

		default:

			break;
	}


       	// Release lock for proc list
	fl_proc_unlock();


	return 1;       // SUCCESS
}


int fl_proc_get_index_unlocked(struct proc_s **proc,	// -- OUT
			       fl_uint64_t proc_index)	// -- IN
{
        fl_list_node_t *proc_node;


	// NOTE: Arguments


	// OUT: Check for proc
	if (proc == NULL)
	{
		return 0;	// ERROR: No proc
	}


	// NOTE: Initialization


	// Proc
	*proc = NULL;


	// NOTE: Processing


	// Get requested proc node
	switch (fl_list_get_node(proc_list,
			         &proc_node,
			         proc_index))
	{
		case -2:

			// NOTE: Never reached


			break;

		case -3:

			// NOTE: Never reached


			break;

		case -4:

			return -1;	// ERROR: Position out of range

		default:

			// Set proc
			*proc = (struct proc_s *) proc_node->data;


			break;
	}


	return 1;	// SUCCESS
}


int fl_proc_get_length(fl_uint64_t *proc_length)	// -- IN
{
        // NOTE: Arguments


	// IN: Check for proc length
        if (proc_length == NULL)
        {
                return 0;       // ERROR: No proc length
        }


	// NOTE: Processing


        // Acquire lock for proc list
	fl_proc_lock();


	if (fl_proc_get_length_unlocked(proc_length) < 1)
	{
		// Release lock for proc list
		fl_proc_unlock();


		return -1;      // ERROR: Couldn't get proc
	}


       	// Release lock for proc list
	fl_proc_unlock();


	return 1;       // SUCCESS
}


int fl_proc_get_length_unlocked(fl_uint64_t *proc_length)	// -- IN
{
	// NOTE: Processing


	return fl_list_length(proc_list, proc_length);
}


int fl_proc_del_index(struct proc_s **proc,	// -- OUT
		      fl_uint64_t proc_index)	// -- IN
{
        // NOTE: Arguments


	// OUT: Check for proc
        if (proc == NULL)
        {
                return 0;       // ERROR: No proc
        }


	// NOTE: Processing


        // Acquire lock for proc list
	fl_proc_lock();


	switch (fl_proc_del_index_unlocked(proc,
					   proc_index))
	{
		case -2:

			// NOTE: Never reached


			break;

		case -3:

			// NOTE: Never reached


			break;

		case -4:

			// NOTE: Position out of range


       			// Release lock for proc list
			fl_proc_unlock();


			return -1;	// ERROR: Position out of range

		default:

			break;
	}


       	// Release lock for proc list
	fl_proc_unlock();


	return 1;       // SUCCESS
}


int fl_proc_del_index_unlocked(struct proc_s **proc,	// -- OUT
			       fl_uint64_t proc_index)	// -- IN
{
	fl_list_node_t *proc_node;


	// NOTE: Arguments


	// OUT: Check for proc
	if (proc == NULL)
	{
		return 0;       // ERROR: No proc
	}


	// NOTE: Initialization


	// Proc
	*proc = NULL;


	// NOTE: Processing


	switch (fl_list_del(proc_list,
			    &proc_node,
			    proc_index))
	{
		case -2:

			// NOTE: Never reached


			break;

		case -3:

			// NOTE: Never reached


			break;

		case -4:

			// NOTE: Position out of range


			return -1;	// ERROR: Position out of range

		default:

			// Set proc
			*proc = (struct proc_s *) proc_node->data;


			// Delete proc node
			fl_list_free_node(proc_list,
					  &proc_node);


			break;
	}


	return 1;       // SUCCESS
}


#if FL_OS_NAME_V_OSF1
#       if FL_OS_VERSION_V_5_1
#               if FL_OS_RELEASE_V_B


#               endif
#       endif
#elif FL_OS_NAME_V_IRIX
#       if FL_OS_VERSION_V_6_5
#               if FL_OS_RELEASE_V_30


int fl_proc_get_cred(struct cred *cred)	// -- IN


#               endif
#       endif
#endif
{
        char *dp;
        char *tp;
        fl_uint32_t bytes_index;


#if FL_OS_NAME_V_IRIX
#       if FL_OS_VERSION_V_6_5
#               if FL_OS_RELEASE_V_30


        struct cred *cr;


#               endif
#       endif
#endif


	// NOTE: Arguments


        // IN: Check for credentials
        if (cred == NULL)
        {
                return 0;       // ERROR: No credentials
        }


	// NOTE: Processing


#if FL_OS_NAME_V_OSF1
#       if FL_OS_VERSION_V_5_1
#               if FL_OS_RELEASE_V_B


#               endif
#       endif
#elif FL_OS_NAME_V_IRIX
#       if FL_OS_VERSION_V_6_5
#               if FL_OS_RELEASE_V_30


        // Get credentials of current process
        cr = get_current_cred();


        // Set data pointer
        dp = (char *) cr;


#               endif
#       endif
#endif


        // Set target pointer
        tp = (char *) cred;


#if FL_OS_NAME_V_OSF1
#       if FL_OS_VERSION_V_5_1
#               if FL_OS_RELEASE_V_B


#               endif
#       endif
#elif FL_OS_NAME_V_IRIX
#       if FL_OS_VERSION_V_6_5
#               if FL_OS_RELEASE_V_30


        // Copy credentials
        for (bytes_index = 0;
	     bytes_index < sizeof(struct cred);
	     bytes_index++)
        {
                *tp++ = *dp++;
        }


#               endif
#       endif
#endif


        //FL_DEBUG_UPRINTF(("[fl_proc_get_cred] cred.cr_uid = %u\n\
			  //cred.cr_gid = %u\n\
			  //cred._cr_ruid = %u\n\
			  //cred._cr_rgid = %u\n",
			  //cred.cr_uid,
			  //cred.cr_gid,
			  //cred._cr_ruid,
			  //cred._cr_rgid))


        return 1;       // SUCCESS
}


