

#ifndef __FL_PROC_H__
#define __FL_PROC_H__


// NOTE: OS header


#include "fl_conf_id.h"


// NOTE: System header files


#if FL_OS_NAME_V_OSF1
#       if FL_OS_VERSION_V_5_1
#               if FL_OS_RELEASE_V_B


#               endif
#       endif
#elif FL_OS_NAME_V_IRIX
#       if FL_OS_VERSION_V_6_5
#               if FL_OS_RELEASE_V_30


#include <sys/types.h>
#include <sys/cred.h>


#               endif
#       endif
#endif


// NOTE: Local header files


#include "fl_common.h"
#include "fl_list.h"


///////////////////////
// NOTE: Definitions //
///////////////////////


// NOTE: Macro definitions


#define FL_DATA_MASK      0x000000ff
#define FL_DATA_PID       0x00000001
#define FL_DATA_EXEC_NAME 0x00000002


// NOTE: OS specific macro definitions


#if FL_OS_NAME_V_IRIX
#       if FL_OS_VERSION_V_6_5
#               if FL_OS_RELEASE_V_30


#if FL_CPU_BITS_V_32

#define FL_SIGN_EXT(N) (N)

#elif FL_CPU_BITS_V_64

#define FL_SIGN_EXT(N) ((N) | 0xffffffff00000000LL)

#endif


#if NBPP == 4096        
#define FL_PDAPAGE	FL_SIGN_EXT(0xffffa000)     /* private data area page */
#elif NBPP == 16384
#define FL_PDAPAGE	FL_SIGN_EXT(0xffffc000)     /* private data area */
#endif

#define FL_PDAADDR	FL_PDAPAGE
#define FL_PRIVATE	(*((pda_t *) FL_PDAADDR))
#define FL_CURPROCP	(FL_PRIVATE.p_curuthread ? UT_TO_PROC(FL_PRIVATE.p_curuthread) : 0)


//#define curprocp                (private.p_curuthread ? UT_TO_PROC(private.p_curuthread) : 0)
#define PLOCK(p)                mutex_bitlock(&(p)->p_flag, STATELOCK);
#define PUNLOCK(p,rv)           mutex_bitunlock(&(p)->p_flag, STATELOCK, rv);


#define VPROC_NULL	((vproc_t *) 0)
#define PID_INDEX(pid)	(((pid) - pid_base) % pidtabsz)
#define PID_SLOT(pid)	(&pidtab[PID_INDEX(pid)])

#define pidactive_lock()		mutex_spinlock(&pid_active_list.pa_lock)
#define pidactive_unlock(X)		mutex_spinunlock(&pid_active_list.pa_lock, X)
#define pidslot_nested_lock(ps)		nested_spinlock(&(ps)->ps_lock)
#define pidslot_nested_unlock(ps)	nested_spinunlock(&(ps)->ps_lock)
#define VPROC_NESTED_REFCNT_LOCK(v)     nested_spinlock(&(v)->vp_refcnt_lock)
#define VPROC_REFCNT_UNLOCK(v, s)       mutex_spinunlock(&(v)->vp_refcnt_lock,s)
#define VPROC_RELE(v)                   vproc_release(v)


#define VPROC_BHV_HEAD(v)       (&(v)->vp_bhvh)
#define VPROC_BHV_FIRST(v)      BHV_HEAD_FIRST(VPROC_BHV_HEAD(v))
#define _VPOP_(op, v)   (*((vproc_ops_t *)BHV_OPS(VPROC_BHV_FIRST(v)))->op)


#define _VPROC_(vpop, v, parms)                                         \
        {                                                               \
        BHV_READ_LOCK(VPROC_BHV_HEAD(v));                               \
        _VPOP_(vpop, v) parms;                                          \
        BHV_READ_UNLOCK(VPROC_BHV_HEAD(v));                             \
        }


#define VPROC_GET_XATTR(v, _flags, _xattr) \
        _VPROC_(vpop_get_xattr, v, (VPROC_BHV_FIRST(v), _flags, _xattr))


// NOTE: vproc attributes


#define VGXATTR_NICE            0x00000200      /* process nice value */
#define VGXATTR_WCHAN           0x00000400      /* thread sleep chan */
#define VGXATTR_PRI             0x00000800      /* thread priority */
#define VGXATTR_UID             0x00001000      /* real/eff/saved uid */
#define VGXATTR_LCPU            0x00002000      /* thread last cpu */
#define VGXATTR_RSS             0x00004000      /* thread rss */
#define VGXATTR_SIZE            0x00008000      /* thread as size */
#define VGXATTR_UST             0x00010000      /* thread aggregrate time */
#define VGXATTR_NAME            0x00020000      /* process name/args */
#define VGXATTR_STAT            0x00040000      /* process/thread flags */


typedef struct proc {
        proc_proxy_t    p_proxy;        // the proxy structure //

        uint            p_flag;         // generic process flags //
        char            p_acflag;       // accounting flags? //
        char            p_stat;         // process state //
        pid_t           p_pgid;         // name of process group leader //
        pid_t           p_pid;          // process id //
        pid_t           p_ppid;         // process id of parent//
        pid_t           p_sid;          // session id //
        
        struct vnode    *p_script;      // script executed by #!/path/intp //
        
        struct fdt      *p_fdt;         // file descriptor table //

        // Signal information //
        k_sigset_t      p_sigtrace;     // tracing signal mask for /proc //
        struct sigvec_s p_sigvec;       // pointer to signal vector info //
        
        struct child_pidlist *p_childpids; // ptr to first child process //
        mutex_t         p_childlock;    // protects process child list //
        sv_t            p_wait;         // waiting for something to happen //
        sv_t            p_wexit;        // Unused cf: 576784 & 656452 //
        time_t          p_start;        // Time of process creation //
        time_t          p_ticks;        // lbolt at process creation //
        struct rusage   p_cru;          // sum of stats for reaped children //
        struct itimer_info {
                struct  itimerval realtimer; // time to next real alarm //
                lock_t  itimerlock;     // ITIMER_REAL handling lock //
                __uint64_t interval_tick;
                __uint64_t next_timeout; 
                int     interval_unit;
                toid_t  next_toid; 
        } p_ii;


#if FL_CPU_BITS_V_64


	fl_uint64_t filler;


#endif	// FL_CPU_BITS_V_64


        struct user_info {      
                char            u_comm[PSCOMSIZ];
                char            u_psargs[PSARGSZ];       // args from exec //
                struct rlimit   u_rlimit[RLIM_NLIMITS]; // resource limits //
        } p_ui;

#define p_comm          p_ui.u_comm
#define p_rlimit        p_ui.u_rlimit
#define p_psargs        p_ui.u_psargs

        struct proc     *p_slink;       // share group link //
        struct shaddr_s *p_shaddr;      // pointer to shared group desc //
        char            p_exitsig;      // signal to send on exit of sharegrp //

        mrlock_t        p_who;          // process personality lock -- //
                                        // protects p_vpgrp, etc//
        struct proc     *p_pgflink;     // process group chain //
        struct proc     *p_pgblink;     // process group chain //
        struct vpgrp    *p_vpgrp;       // virtual process group structure //
        mrlock_t        p_credlock;     // process credentials lock //
        struct cred     *p_cred;        // process credentials //
        struct prnode   *p_trace;       // pointer to /proc prnode //
        struct prnode   *p_pstrace;     // pointer to /proc/pinfo prnode //
#ifdef PRCLEAR
        int              p_prnode_refcnt; // prnode reference count     //
                                          // for p_trace and p_pstrace  //
#else
        int              p_notused;     // make sure that we do not change //
                                        // the size of the structure       //
#endif
        uint            p_exec_cnt;     // exec generation number //
        struct vnode    *p_exec;        // vnode for original exec //
        struct vpagg_s  *p_vpagg;       // process aggregate info //
        struct cipso_proc       *p_cipso_proc;  // T-irix cipso_proc ptr //
        struct pr_tracemasks    *p_trmasks;     // Procfs syscall/fault trace //
        u_int           p_dbgflags;     // flags for low-freq /proc events //

        pid_t           p_unblkonexecpid; // pid to unblock on exec //

#ifdef CKPT
        mrlock_t        p_ckptlck;
        struct avltree_desc *p_ckptshm;
#endif
        struct exit_callback *p_ecblist; // funcs to call on exit //
        ptimer_info_t   *p_ptimers;     // array of posix timers //
        bhv_desc_t      p_bhv;          // Base behavior //

        struct prof     *p_profp;       // ptr to list of prof structures //
        short           p_profn;        // number of prof structures in list //
        inst_t          *p_fastprof[PROF_FAST_CNT+1]; // space to save pc's //
        int             p_fastprofcnt;  // index into p_fastprof //

        struct kaiocbhd *p_kaiocbhd;    // KAIO //
        // More miscellany //
        struct shacct   *p_shacct;      // Shadow accounting data //
        __int64_t       p_parcookie;    // par syscall access token //
        void            *p_migrate;     // what to send when migrated/rexec'd //
#ifdef _SHAREII
        struct ShadProc_t      *p_shareP;       // shadow proc entry //
#endif // _SHAREII //
} proc_t;


// NOTE: vproc


typedef struct vproc {
        kqueue_t                vp_queue;       /* Hash queue link */
        pid_t                   vp_pid;         /* vproc process identifier */
        int                     vp_refcnt;      /* vproc reference count */
        int                     vp_state;       /* State of process */
        int                     vp_needwakeup;  /* if someone is waiting */
        lock_t                  vp_refcnt_lock; /* lock for ref. count */
        sv_t                    vp_cond;        /* condition variable */
        bhv_head_t              vp_bhvh;        /* Behavior head */
} vproc_t;


typedef struct vp_get_attr_s {
        pid_t   va_ppid;                /* Parents pid */
        pid_t   va_pgid;                /* process group id */
        pid_t   va_sid;                 /* Session id */
        int     va_has_ctty;            /* proc has controlling tty */
        uid_t   va_uid;                 /* process uid */
        uid_t   va_gid;                 /* process gid */
        time_t  va_signature;           /* Process signature */
        asid_t  va_asid;                /* Process address space id */
        cell_t  va_cellid;              /* Process cell id */
} vp_get_attr_t;


typedef struct vp_get_xattr_s {
        vp_get_attr_t   xa_basic;
        char            xa_name[PSCOMSIZ];
        char            xa_args[PSARGSZ];
        struct timeval  xa_utime;
        struct timeval  xa_stime;
        char            xa_stat;
        char            xa_nice;
        cpuid_t         xa_lcpu;
        uint            xa_pflag;
        uint            xa_uflag;
        long            xa_pri;
        caddr_t         xa_slpchan;
        uid_t           xa_realuid;
        uid_t           xa_saveuid;
        pgno_t          xa_rss;
        size_t          xa_stksz;
        pgno_t          xa_size;
} vp_get_xattr_t;


typedef struct vproc_ops {
        bhv_position_t  vpr_position;   /* position within behavior chain */
        void (*vpop_destroy)
                (bhv_desc_t *b);        /* behavior */

        void (*vpop_get_proc)           /* XXX temporary interface */
                (bhv_desc_t *b,         /* behavior */
                struct proc **p,        /* returned proc pointer */
                int flag);              /* 1 = local only */

        int (*vpop_get_nice)
                (bhv_desc_t *b,         /* behavior */
                 int *nice,             /* returned nice value */
                 struct cred *);        /* Callers credentials */

        int (*vpop_set_nice)
                (bhv_desc_t *b,         /* behavior */
                 int *nice,             /* nice value */
                 struct cred *scred,    /* calling process information */
                 int flag);             /* VNICE_INCR, when called */
                                        /* to implement 'nice' semantics. */
        void (*vpop_proc_set_dir)
                (bhv_desc_t *b,         /* behavior */
                 struct vnode *vp,      /* directory vnode */
                 struct vnode **dir,    /* address of vnode pointer */
                 int flag);             /* VPROC_DIR_ROOT for root directory, */
                                        /* VPROC_DIR_CURRENT otherwise */
        int (*vpop_proc_umask)
                (bhv_desc_t *b,         /* behavior */
                 short  cmask,          /* new file creation mask */
                 short *omask);         /* address to store old mask */

        void (*vpop_get_attr)
                (bhv_desc_t *b,         /* behavior */
                 int flags,             /* mask of attributes wanted */
                 vp_get_attr_t *attr);  /* returned attributes */

        int (*vpop_parent_notify)
                (bhv_desc_t *b,         /* behavior */
                 pid_t pid,             /* pid that is exiting */
                 int wcode,             /* wcode of exiting proc. */
                 int wdata,             /* wdata of exiting proc. */
                 struct timeval utime,  /* current u and s time of process */
                 struct timeval stime,
                 pid_t pgid,            /* Current pgrp id of process */
                 //sequence_num_t pgseq,  /* Current pgrp signal sequence */
                 int pgseq,
                 short xstat);          /* exit status of child */

        void (*vpop_reparent)
                (bhv_desc_t *b,         /* behavior */
                 int to_detach);        /* perform orphan pgrp jazz */

        void (*vpop_reap)
                (bhv_desc_t *b,         /* behavior */
                 int flags,             /* flags */
                 struct rusage *,       /* rusage stats from child */
                 struct cpu_mon *,      /* R10000 only - hw perf counters */
                 int *rtflags);         /* Returned flags from server */

        int (*vpop_add_exit_callback)
                (bhv_desc_t *b,         /* behavior */
                 int flags,             /* flags */
                 void (* func)(void *), /* function to call */
                 void *arg);            /* argument to call func with */

        void (*vpop_run_exitfuncs)
                (bhv_desc_t *b);        /* behavior */

        void (*vpop_set_state)
                (bhv_desc_t *b,         /* behavior */
                 int state);            /* state */

        int (*vpop_thread_state)
                (bhv_desc_t *b,         /* behavior */
                 int state);            /* state */

        int (*vpop_sendsig)
                (bhv_desc_t *b,         /* behavior */
                 int sig,               /* signal to post */
                 int flags,             /* input flags */
                 pid_t sid,             /* session id of caller */
                 struct cred *,         /* credentials of caller */
                 struct k_siginfo *);   /* Associated siginfo struct to post */

        void (*vpop_sig_threads)
                (bhv_desc_t *b,         /* behavior */
                 int sig,               /* signal to post */
                 struct k_siginfo *);   /* associated siginfo struct to post */

        int (*vpop_sig_thread)
                (bhv_desc_t *b,         /* behavior */
                 ushort_t id,           /* thread id */
                 int sig);              /* signal to post */

        void (*vpop_prstop_threads)
                (bhv_desc_t *b);        /* behavior */

        int (*vpop_prwstop_threads)
                (bhv_desc_t *b,         /* behavior */
                 struct prnode *);      /* prnode */

        void (*vpop_prstart_threads)
                (bhv_desc_t *b);        /* behavior */

        int (*vpop_setpgid)
                (bhv_desc_t *b,         /* behavior */
                 pid_t pgid,            /* pgrp id to set to */
                 pid_t,                 /* Callers pid */
                 pid_t sid);            /* Callers session id */

        void (*vpop_pgrp_linkage)
                (bhv_desc_t *b,         /* behavior */
                 pid_t parent_pgid,     /* parent's new pgrp */
                 pid_t parent_sid);     /* parent's new session id */

        int (*vpop_setuid)
                (bhv_desc_t *b,         /* behavior */
                 uid_t ruid,            /* real uid */
                 uid_t euid,            /* eff uid */
                 int flags);            /* flags */

        int (*vpop_setgid)
                (bhv_desc_t *b,         /* behavior */
                 gid_t rgid,            /* real gid */
                 gid_t egid,            /* eff gid */
                 int flags);            /* flags */

        void (*vpop_setgroups)
                (bhv_desc_t *b,         /* behavior */
                 int setsize,           /* # entries of new set */
                 gid_t *newgids);       /* new group set */

        int (*vpop_setsid)
                (bhv_desc_t *b,         /* behavior */
                 pid_t *new_pgid);      /* output parameter - proc's new pgid */

        int (*vpop_setpgrp)
                (bhv_desc_t *b,         /* behavior */
                 int flags,             /* flags */
                 pid_t *new_pgid);      /* output parameter - proc's new pgid */

        int (*vpop_ctty_access)
                (bhv_desc_t *b,         /* behavior */
                 int acc);		/* access type */
                 //enum jobcontrol acc);  /* access type */

        int (*vpop_ctty_clear)
                (bhv_desc_t *b);        /* behavior */

        void (*vpop_ctty_hangup)
                (bhv_desc_t *b);        /* behavior */

        void (*vpop_get_sigact)
                (bhv_desc_t *b,         /* behavior */
                int sig,                /* signal */
                k_sigaction_t *oact);   /* old action, returned */

        void (*vpop_set_sigact)
                (bhv_desc_t *b,         /* behavior */
                int sig,                /* signal */
                void (*disp)(),         /* signal disposition */
                k_sigset_t *mask,       /* signal mask */
                int flags,              /* guess */
                int (*sigtramp)());     /* user trampoline function */

        struct sigvec_s *(*vpop_get_sigvec)
                (bhv_desc_t *b,         /* behavior */
                int flags);             /* access mode */

        void (*vpop_put_sigvec)
                (bhv_desc_t *b);        /* behavior */

        int (*vpop_trysig_thread)
                (bhv_desc_t *b);        /* behavior */

        void (*vpop_flag_threads)
                (bhv_desc_t *b,         /* behavior */
                 uint_t flag,           /* flag bits to set/clear */
                 int op);               /* set or clear */

        void (*vpop_getvpagg)
                (bhv_desc_t *b,         /* behavior */
                 struct vpagg_s **);    /* pointer to store result in */

        void (*vpop_setvpagg)
                (bhv_desc_t *b,         /* behavior */
                 struct vpagg_s *);     /* vpagg_t to set */

        struct kaiocbhd *(*vpop_getkaiocbhd)
                (bhv_desc_t *b);        /* behavior */

        struct kaiocbhd *(*vpop_setkaiocbhd)
                (bhv_desc_t *b,         /* behavior */
                 struct kaiocbhd *);    /* new kaiocbhd to set */

        void (*vpop_getrlimit)
                (bhv_desc_t *b,         /* behavior */
                 int which,             /* which limit to retrieve */
                 struct rlimit *rlim);  /* result */

        int (*vpop_setrlimit)
                (bhv_desc_t *b,         /* behavior */
                 int which,             /* which limit to set */
                 struct rlimit *rlim);  /* new value */

        int (*vpop_exec)
                (bhv_desc_t *b,         /* behavior */
                 void *);		/* exec uarg */
                 //struct uarg *);        /* exec uarg */

        int (*vpop_set_fpflags)
                (bhv_desc_t *b,         /* behavior */
                 u_char fpflag,         /* fp flag to set */
                 int flags);            /* directives */

        void (*vpop_set_proxy)
                (bhv_desc_t *b,         /* behavior */
                 int,                   /* attribute */
                 __psint_t,             /* arg1 */
                 __psint_t);            /* arg1 */

        void (*vpop_times)
                (bhv_desc_t *b,         /* behavior */
                 void *);		/* times structure */
                 //struct tms *);         /* times structure */

        void (*vpop_read_us_timers)
                (bhv_desc_t *b,         /* behavior */
                 struct timespec *,     /* timespec for utime */
                 struct timespec *);    /* timespec for stime */

        void (*vpop_getrusage)
                (bhv_desc_t *b,         /* behavior */
                 int flags,             /* childs rusage or own rusage? */
                 struct rusage *);

        void (*vpop_get_extacct)
                (bhv_desc_t *b,         /* behavior */
                 struct proc_acct_s *); /* counters, etc, returned here */

        void (*vpop_get_acct_timers)
                (bhv_desc_t *b,         /* behavior */
                 struct acct_timers *); /* The timers that ext. acct. wants */

        void (*vpop_sched_setparam)
                (bhv_desc_t *b,         /* behavior */
                 int pri,
                 int *error);

        void (*vpop_sched_getparam)
                (bhv_desc_t *b,         /* behavior */
                 struct cred *,         /* callers creds */
                 int *pri,              /* return value */
                 int *error);

        void (*vpop_sched_setscheduler)
                (bhv_desc_t *b,         /* behavior */
                 int pri,
                 int policy,
                 __int64_t *rval,
                 int *error);

        void (*vpop_sched_getscheduler)
                (bhv_desc_t *b,         /* behavior */
                 struct cred *,         /* callers creds */
                 __int64_t *sched,      /* return value */
                 int *error);

        void (*vpop_sched_rr_get_interval)
                (bhv_desc_t *b,         /* behavior */
                 struct timespec *);    /* returned timeslice */

        void (*vpop_setinfo_runq)
                (bhv_desc_t *b,         /* behavior */
                 int rqtype,
                 int arg,
                 __int64_t *rval);

        void (*vpop_getrtpri)
                (bhv_desc_t *b,         /* behavior */
                 __int64_t *rval);

        void (*vpop_setmaster)
                (bhv_desc_t *b,         /* behavior */
                 pid_t callers_pid,
                 int *error);

        void (*vpop_schedmode)
                (bhv_desc_t *b,         /* behavior */
                 int arg,
                 __int64_t *rval,
                 int *error);

        void (*vpop_sched_aff)
                (bhv_desc_t *b,         /* behavior */
                 int flag,
                 __int64_t *rval);

        void (*vpop_prnode)
                (bhv_desc_t *b,         /* behavior */
                 int flag,
                 struct vnode **vpp);

        void (*vpop_exitwake)           /* op deleted */
                (void);

        int (*vpop_procblk)
                (bhv_desc_t *b,
                 int action,
                 int count,
                 struct cred *cr,
                 int isself);

        int (*vpop_prctl)
                (bhv_desc_t *b,
                 int option,
                 sysarg_t arg,
                 int isself,
                 struct cred *cr,
                 pid_t callers_pid,
                 union rval *rvp);

        int (*vpop_set_unblkonexecpid)
                (bhv_desc_t *b,
                 pid_t unblkpid);

        void (*vpop_unblkpid)
                (bhv_desc_t *b);

        int (*vpop_unblock_thread)
                (bhv_desc_t *b,         /* behavior */
                 ushort_t id);          /* exec uarg */

        int (*vpop_fdt_dup)             /* all this for attachsema */
                (bhv_desc_t *,          /* behavior */
                 struct vfile *,        /* open file */
                 int *);                /* new file descriptor */

#ifdef CKPT
        void (*vpop_ckpt_shmget)        /* so we can log shmids */
                (bhv_desc_t *,          /* behavior */
                 int );                 /* shmid */

        int (*vpop_get_ckpt)            /* so we can retrieve proc ckpt info */
                (bhv_desc_t *,          /* behavior */
                 int,                   /* code */
                 void *arg);
#endif
        void (*vpop_poll_wakeup)        /* thread part of pollwakeup */
                (bhv_desc_t *,          /* behavior */
                 uthreadid_t,           /* tid */
                 ushort_t);             /* rotor hint */

        int (*vpop_getcomm)             /* fetch command line */
                (bhv_desc_t *,          /* behavior */
                 struct cred *,         /* callers creds */
                 char *,                /* string buffer */
                 size_t);               /* buffer length */

        void (*vpop_get_xattr)
                (bhv_desc_t *b,         /* behavior */
                 int flags,             /* mask of xattributes wanted */
                 vp_get_xattr_t *xattr);/* returned xattributes */

        void (*vpop_read_all_timers)
                (bhv_desc_t *b,         /* behavior */
                 struct timespec *);    /* timespec for stime */

        /* ADD all new entries before vpop_noop.
         * Null routine, to inhibit compilation due to
         * prototype mismatch if op-tables do not match.
         */
        void (*vpop_noop)
                (void);
} vproc_ops_t;


// NOTE: pid_entry_t


typedef struct pid_entry {
        kqueue_t        pe_queue;               /* active list */
        pid_t           pe_pid;                 /* pid busy in this entry */
        union {
                struct {
                        uint    peu_nodealloc : 8, /* can't dealocate */
                                peu_dealloc : 1, /* want to dealocate */
                                peu_pad : 19,
                                peu_sess : 1,   /* used as session id? */
                                peu_pgrp : 1,   /* used as pgrp id? */
                                peu_batch : 1,  /* used as batch id? */
                                peu_proc : 1;   /* used as proc id? */
                } peu_bit;
                struct {
                        uint    peu_pad : 28,
                                peu_busy : 4;
                } peu_busybit;
        } pe_ubusy;

        struct vproc            *pe_vproc;      /* vproc ptr when active */
        struct pid_entry        *pe_next;       /* busy pid link */

} pid_entry_t;


#define pe_nodealloc pe_ubusy.peu_bit.peu_nodealloc
#define pe_dealloc   pe_ubusy.peu_bit.peu_dealloc


// NOTE: pid_slot_t


typedef struct pid_slot {
        pid_t   ps_pid;
        lock_t  ps_lock;
        union {
                struct {
                        int     psu_forw;       /* next index in free list */
                        int     psu_back;       /* prev index in free list */
                } psu_free;
                struct {
                        int     psu_active;     /* Has active proc? */
                        int     psu_busycnt;    /* How many active procs */
                } psu_busy;
        } psu;
        struct pid_entry        *ps_chain;      /* busy pid link */
} pid_slot_t;


typedef struct pid_active {
        kqueue_t        pa_queue;               /* active list */
        lock_t          pa_lock;
} pid_active_t;


#               endif
#       endif
#endif


struct proc_s
{
	pid_t pid;                      // Pid of proc
	void *priv;			// Private data pointer
};


////////////////////////
// NOTE: Declarations //
////////////////////////


// NOTE: Function prototypes


int fl_proc_init(void);
int fl_proc_deinit(void);
int fl_proc_new(struct proc_s **);
int fl_proc_free(struct proc_s *);
int fl_proc_get_data(fl_uint32_t,
		     pid_t *);
int fl_proc_lock(void);
int fl_proc_unlock(void);
int fl_proc_get(struct proc_s **);
int fl_proc_get_unlocked(struct proc_s **);
int fl_proc_push(struct proc_s *);
int fl_proc_push_unlocked(struct proc_s *);
int fl_proc_shift(struct proc_s **);
int fl_proc_shift_unlocked(struct proc_s **);
int fl_proc_get_index(struct proc_s **,
		      fl_uint64_t);
int fl_proc_get_index_unlocked(struct proc_s **,
			       fl_uint64_t);
int fl_proc_get_length(fl_uint64_t *);
int fl_proc_get_length_unlocked(fl_uint64_t *);
int fl_proc_del_index(struct proc_s **,
		      fl_uint64_t);
int fl_proc_del_index_unlocked(struct proc_s **,
			       fl_uint64_t);


#if FL_OS_NAME_V_OSF1
#       if FL_OS_VERSION_V_5_1
#               if FL_OS_RELEASE_V_B


#               endif
#       endif
#elif FL_OS_NAME_V_IRIX
#       if FL_OS_VERSION_V_6_5
#               if FL_OS_RELEASE_V_30


int fl_proc_get_cred(struct cred *);


#               endif
#       endif
#endif


#endif	// __FL_PROC_H__


