

// NOTE: System header files


// NOTE: Local header files


#include "fl_scsi.h"


// NOTE: Private macros


#define FL_MIN(A, B) (A < B ? A : B)


// SCSI inquiry
static struct scsi_cmd_inquiry_s scsi_inquiry;
static struct scsi_cmd_prevent_medium_removal_s scsi_pmr;


// NOTE: Public functions


int fl_scsi_check_cmd(unsigned char *cmdbuf,		// -- IN
                      fl_uint32_t cmdlen,		// -- IN
                      struct scsi_cmd_s *scsi_cmd)	// -- IN
{
        // NOTE: Check command
        //
        //  Check for implemented SCSI commands and convert command buffer (data)
        //  to usable struct representation (cmd).


        // NOTE: Arguments


        // IN: Check for command buffer
        if (cmdbuf == NULL)
        {
                return 0;       // ERROR: No command buffer
        }


        // IN: Check for command length of zero
        if (cmdlen == 0)
        {
                return -1;      // ERROR: Command length is zero
        }


        // IN: Check for SCSI command
        if (scsi_cmd == NULL)
        {
                return -2;      // ERROR: No SCSI command
        }


	// NOTE: Processing


        // NOTE: Task command buffer
        //
        //  The task command buffer is 16 bytes long.


        // Check for SCSI command
        switch (cmdbuf[0])
        {
                case FL_SCSI_OP_TEST_UNIT_READY:

                        FL_DEBUG_UPRINTF((\
					  "[fl_scsi_check_cmd] INFO: TEST UNIT READY\n"))


                        // Set SCSI command fields
                        scsi_cmd->op_code = FL_SCSI_OP_TEST_UNIT_READY;


                        break;

                case FL_SCSI_OP_INQUIRY:

                        FL_DEBUG_UPRINTF((\
					  "[fl_scsi_check_cmd] INFO: INQUIRY\n"))


                        // Set SCSI command fields
                        scsi_cmd->op_code = FL_SCSI_OP_INQUIRY;
                        scsi_cmd->cmd = (void *) &scsi_inquiry;
                        scsi_cmd->data = (void *) cmdbuf;
                        scsi_cmd->data_length = cmdlen;


                        // Convert data to cmd
                        if (fl_scsi_convert(FL_SCSI_CONV_DATA_TO_CMD,
					    scsi_cmd) < 1)
                        {
                                return -3;      // ERROR: Couldn't convert data to command (INQUIRY)
                        }


                        break;

                case FL_SCSI_OP_PREVENT_MEDIUM_REMOVAL:

                        FL_DEBUG_UPRINTF((\
					  "[fl_scsi_check_cmd] INFO: PREVENT MEDIUM REMOVAL\n"))


                        // Set SCSI command fields
                        scsi_cmd->op_code = FL_SCSI_OP_PREVENT_MEDIUM_REMOVAL;
                        scsi_cmd->cmd = (void *) &scsi_pmr;
                        scsi_cmd->data = (void *) cmdbuf;
                        scsi_cmd->data_length = cmdlen;


                        // Convert data to cmd
                        if (fl_scsi_convert(FL_SCSI_CONV_DATA_TO_CMD,
					    scsi_cmd) < 1)
                        {
                                return -4;      // ERROR: Couldn't convert data to command (INQUIRY)
                        }


                        break;

                case FL_SCSI_OP_READ6:

                        FL_DEBUG_UPRINTF((\
					  "[fl_scsi_check_cmd] INFO: READ(6)\n"))


                        //read6 = (struct scsi_read6 *) cmdbuf;


                        // NOTE: Compose lba


                        //read6->lba = ((struct scsi_read6_lba *) cmdbuf)->lba;
                        //read6->lba = read6->lba & 0x00ffffff;


                        //FL_DEBUG_UPRINTF(("[fl_scsi_check_cmd] INFO: lba = %u\n",
					  //read6->lba))


			//FL_DEBUG_UPRINTF(("[fl_scsi_check_cmd] INFO: length = %u\n",
					  //read6->transfer_length))


                        //break;

                default:

                        FL_DEBUG_UPRINTF((\
					  "[fl_scsi_check_cmd] Unimplemented SCSI command = 0x%x\n",
					  cmdbuf[0]))


                        return -5;      // ERROR: Unimplemented SCSI command

        }


        return 1;       // SUCCESS
}


int fl_scsi_convert(fl_uint8_t dir,			// -- IN
		    struct scsi_cmd_s *scsi_cmd)	// -- IN
{
        struct scsi_cmd_inquiry_s *cmd_inq;
        struct scsi_data_6bytes_s *data_inq;
        struct scsi_cmd_prevent_medium_removal_s *cmd_pmr;
        struct scsi_data_6bytes_s *data_pmr;
        struct scsi_cmd_inquiry1_s *cmd_inq1;
        struct scsi_data_inquiry1_s *data_inq1;
        struct ushort_to_bytes_s *ushort_to_bytes;


        // NOTE: Arguments


        // IN: Check for valid conversion direction
        if ((dir != FL_SCSI_CONV_DATA_TO_CMD) &&
            (dir != FL_SCSI_CONV_CMD_TO_DATA))
        {
                return 0;       // ERROR: Invalid conversion direction
        }


        // IN: Check for SCSI command
        if (scsi_cmd == NULL)
        {
                return -1;      // ERROR: No SCSI command
        }


        // Switch SCSI commands
        switch (scsi_cmd->op_code)
        {
                case FL_SCSI_OP_INQUIRY:

                        // Check for correct length
                        if (scsi_cmd->data_length != 6)
                        {
                                return -2;      // ERROR: Wrong command length for INQUIRY
                        }


                        // Get variables
                        cmd_inq = (struct scsi_cmd_inquiry_s *) scsi_cmd->cmd;
                        data_inq = (struct scsi_data_6bytes_s *) scsi_cmd->data;


                        FL_DEBUG_UPRINTF((\
					  "[fl_scsi_convert] INFO: cmd_inq = %p\n",
					  cmd_inq))


                        FL_DEBUG_UPRINTF((\
					  "[fl_scsi_convert] INFO: data_inq = %p\n",
					  data_inq))


                        // Convert
                        if (dir == FL_SCSI_CONV_DATA_TO_CMD)
                        {
                                // Decompose data bytes
                                cmd_inq->op_code = FL_SCSI_OP_INQUIRY;
                                cmd_inq->CMDDT = data_inq->bytes[1] & 0x2;
                                cmd_inq->EVPD = data_inq->bytes[1] & 0x1;
                                cmd_inq->page_code = (fl_uint8_t) data_inq->bytes[2];
                                bcopy((void *) &data_inq->bytes[3],
                                      (void *) &cmd_inq->alloc_length,
                                      2);
                                cmd_inq->control = data_inq->bytes[5];
                        }
                        else
                        {
                                // Compose data bytes
                                data_inq->bytes[0] = FL_SCSI_OP_INQUIRY;
                                data_inq->bytes[1] = 0x0 |
						     (cmd_inq->CMDDT & 0x1) |
						     (cmd_inq->EVPD & 0x1);
                                data_inq->bytes[2] = cmd_inq->page_code;
                                ushort_to_bytes = (struct ushort_to_bytes_s *) &cmd_inq->alloc_length;
                                data_inq->bytes[3] = ushort_to_bytes->bytes[0];
                                data_inq->bytes[4] = ushort_to_bytes->bytes[1];
                                data_inq->bytes[5] = cmd_inq->control;
                        }


                        break;

                case FL_SCSI_OP_PREVENT_MEDIUM_REMOVAL:

                        // Check for correct length
                        if (scsi_cmd->data_length != 6)
                        {
                                return -3;      // ERROR: Wrong command length for PREVENT_MEDIUM_REMOVAL
                        }


                        // Get variables
                        cmd_pmr = (struct scsi_cmd_prevent_medium_removal_s *) scsi_cmd->cmd;
                        data_pmr = (struct scsi_data_6bytes_s *) scsi_cmd->data;


                        FL_DEBUG_UPRINTF((\
					  "[fl_scsi_convert] INFO: cmd_pmr = %p\n",
					  cmd_pmr))


                        FL_DEBUG_UPRINTF((\
					  "[fl_scsi_convert] INFO: data_pmr = %p\n",
					  data_pmr))


                        // NOTE: Convert


                        if (dir == FL_SCSI_CONV_DATA_TO_CMD)
                        {
                                // Decompose data bytes
                                cmd_pmr->op_code = FL_SCSI_OP_PREVENT_MEDIUM_REMOVAL;
                                cmd_pmr->prevent = data_pmr->bytes[4] & 0x3;
                                cmd_pmr->control = data_pmr->bytes[5];
                        }
                        else
                        {
                                // NOTE: Not implemented


                        }


                        break;

                case FL_SCSI_OP_INQUIRY1:

                        // Check for correct length
                        if ((scsi_cmd->data_length < 36) ||
                            (scsi_cmd->data_length > 96))
                        {
                                return -4;      // ERROR: Wrong command length for INQUIRY1
                        }


                        // Get variables
                        cmd_inq1 = (struct scsi_cmd_inquiry1_s *) scsi_cmd->cmd;
                        data_inq1 = (struct scsi_data_inquiry1_s *) scsi_cmd->data;


                        FL_DEBUG_UPRINTF((\
					  "[fl_scsi_convert] INFO: cmd_inq1 = %p\n",
					  cmd_inq1))


                        FL_DEBUG_UPRINTF((\
					  "[fl_scsi_convert] INFO: data_inq1 = %p\n",
					  data_inq1))


                        // Convert
                        if (dir == FL_SCSI_CONV_DATA_TO_CMD)
                        {
                                // NOTE: Not implemented


                        }
                        else
                        {
                                // Initialize data buffer
                                bzero(data_inq1,
				      96);


                                // Compose data bytes
                                data_inq1->bytes[0] = (cmd_inq1->peripheral_qualifier << 5) |
                                                      cmd_inq1->peripheral_devicetype;
                                data_inq1->bytes[1] = cmd_inq1->RMB << 7;
                                data_inq1->bytes[2] = cmd_inq1->VERSION;
                                data_inq1->bytes[3] = (cmd_inq1->NORMACA << 5) |
                                                      (cmd_inq1->HISUP << 4) |
                                                      cmd_inq1->response_data_format;
                                data_inq1->bytes[4] = cmd_inq1->additional_length;
                                data_inq1->bytes[5] = (cmd_inq1->SCCS << 7) |
                                                      (cmd_inq1->ACC << 6) |
                                                      (cmd_inq1->TPGS << 4) |
                                                      (cmd_inq1->THIRDPC << 3) |
                                                      (cmd_inq1->PROTECT);
                                data_inq1->bytes[6] = (cmd_inq1->ENCSERV << 6) |
                                                      (cmd_inq1->VS << 5) |
                                                      (cmd_inq1->MULTIP << 4);
                                data_inq1->bytes[7] = (cmd_inq1->CMDQUE << 1) |
                                                      (cmd_inq1->VS);
                                bcopy((void *) cmd_inq1->vendor_id,
                                      (void *) &data_inq1->bytes[8],
                                      8);
                                bcopy((void *) cmd_inq1->product_id,
                                      (void *) &data_inq1->bytes[16],
                                      16);
                                bcopy((void *) cmd_inq1->product_rev,
                                      (void *) &data_inq1->bytes[32],
                                      4);
                        }


                        break;

                default:

                        FL_DEBUG_UPRINTF((\
					  "[fl_scsi_convert] INFO: Unimplemented SCSI command = 0x%x\n",
					  scsi_cmd->op_code))


                        return -5;      // ERROR: Unimplemented SCSI command

        }


        return 1;       // SUCCESS
}


int fl_scsi_process(struct scsi_cmd_s *scsi_cmd,	// -- IN
		    struct task_desc *task)		// -- IN
{
        char inq1_vendor_id[8] = "SONY    ";
        char inq1_product_id[16] = "                ";
        char inq1_product_rev[4] = "    ";
        struct scsi_cmd_s scsi_cmd_inq1;
        struct scsi_cmd_inquiry1_s cmd_inq1;
        struct scsi_data_inquiry1_s data_inq1;


#ifdef FL_DEBUG


        struct scsi_cmd_prevent_medium_removal_s *cmd_pmr;


#endif


        // NOTE: Arguments


        // IN: Check for SCSI command
        if (scsi_cmd == NULL)
        {
                return 0;       // ERROR: No SCSI command
        }


        // IN: Check for task
        if (task == NULL)
        {
                return -1;      // ERROR: No task
        }


	// NOTE: Prcoessing


        // Switch SCSI commands
        switch (scsi_cmd->op_code)
        {
                case FL_SCSI_OP_TEST_UNIT_READY:

                        task->td_status = FL_SCSI_STS_GOOD;


                        break;

                case FL_SCSI_OP_INQUIRY:

                        FL_DEBUG_UPRINTF((\
					  "[fl_scsi_process] INFO: INQUIRY\n"))


                        // NOTE: INQUIRY
                        //
                        //  Fill in a per loop entry static inquiry response buffer
                        //  identifying a CD-ROM drive device.


                        // TODO: Make these a static buffer initialized once upon fl.o load


                        // Initialize variables
                        bzero(&scsi_cmd_inq1,
			      sizeof(struct scsi_cmd_s));
                        bzero(&cmd_inq1,
			      sizeof(struct scsi_cmd_inquiry1_s));
                        bzero(&data_inq1,
			      sizeof(struct scsi_data_inquiry1_s));


                        // Fill in inquiry buffer
                        cmd_inq1.peripheral_qualifier = 0x00;
                        cmd_inq1.peripheral_devicetype = 0x05;
                        cmd_inq1.RMB = 1;
                        cmd_inq1.VERSION = 0x03;
                        cmd_inq1.NORMACA = 0x00;
                        cmd_inq1.HISUP = 0x00;
                        cmd_inq1.response_data_format = 0x02;
                        cmd_inq1.additional_length = 0x20;
                        cmd_inq1.SCCS = 0x00;
                        cmd_inq1.ACC = 0x00;
                        cmd_inq1.TPGS = 0x00;
                        cmd_inq1.THIRDPC = 0x00;
                        cmd_inq1.PROTECT = 0x00;
                        cmd_inq1.ENCSERV = 0x00;
                        cmd_inq1.VS = 0x00;
                        cmd_inq1.MULTIP = 0x00;
                        cmd_inq1.CMDQUE = 0x01;
                        cmd_inq1.vendor_id = inq1_vendor_id;
                        cmd_inq1.product_id = inq1_product_id;
                        cmd_inq1.product_rev = inq1_product_rev;


                        // Set SCSI command fields
                        scsi_cmd_inq1.op_code = FL_SCSI_OP_INQUIRY1;
                        scsi_cmd_inq1.cmd = (void *) &cmd_inq1;
                        scsi_cmd_inq1.data = (void *) &data_inq1;
                        scsi_cmd_inq1.data_length = sizeof(struct scsi_data_inquiry1_s);


                        // Convert cmd to data
                        fl_scsi_convert(FL_SCSI_CONV_CMD_TO_DATA,
					&scsi_cmd_inq1);


                        // Write inquiry1 data
                        if (copyout(data_inq1.bytes,
                                    task->td_iovbuf[0].iov_base,
                                    FL_MIN(scsi_cmd_inq1.data_length,
					   task->td_iovbuf[0].iov_len)) != 0)
                        {
                                return -2;      // ERROR: Couldn't copy out inquiry1 data
                        }


                        break;

                case FL_SCSI_OP_PREVENT_MEDIUM_REMOVAL:

                        task->td_status = FL_SCSI_STS_GOOD;


#ifdef FL_DEBUG


                        // Get variables
                        cmd_pmr = (struct scsi_cmd_prevent_medium_removal_s *) scsi_cmd->cmd;


#endif


                        FL_DEBUG_UPRINTF((\
					  "[fl_scsi_process] INFO: cmd_pmr->prevent = %u\n",
					  cmd_pmr->prevent))


                        break;

                default:

                        task->td_status = FL_SCSI_STS_UNIMPL;


                        break;

        }


        return 1;       // SUCCESS
}


