

#ifndef __FL_SCSI_H__
#define __FL_SCSI_H__


// NOTE: OS header


#include "fl_conf_id.h"


// NOTE: System header files


#if FL_OS_NAME_V_OSF1
#       if FL_OS_VERSION_V_5_1
#               if FL_OS_RELEASE_V_B


#               endif
#       endif
#elif FL_OS_NAME_V_IRIX
#       if FL_OS_VERSION_V_6_5
#               if FL_OS_RELEASE_V_30


#               endif
#       endif
#endif


// NOTE: Local header files


#include "fl_common.h"


///////////////////////
// NOTE: Definitions //
///////////////////////


// NOTE: Macro definitions


// SCSI command opcodes
#define FL_SCSI_OP_TEST_UNIT_READY 0x00
#define FL_SCSI_OP_READ6 0x08
#define FL_SCSI_OP_INQUIRY 0x12
#define FL_SCSI_OP_PREVENT_MEDIUM_REMOVAL 0x1e
#define FL_SCSI_OP_INQUIRY1 0xff


// SCSI command status
#define FL_SCSI_STS_GOOD 0x00
#define FL_SCSI_STS_UNIMPL 0xff


// NOTE: Conversion
//
//  Structures needed for converting from data to cmd representation
//  are declared here.
// 
//  scsi_cmd_*_s structures are a SCSI command's cmd representation
//  and scsi_data_*_s are its data equivalent.


#define FL_SCSI_CONV_NULL 0
#define FL_SCSI_CONV_DATA_TO_CMD 1
#define FL_SCSI_CONV_CMD_TO_DATA 2


struct scsi_read6
{
        fl_uint8_t op_code;
        fl_uint8_t lba0;
        fl_uint8_t lba1;
        fl_uint8_t lba2;
        fl_uint8_t transfer_length;
        fl_uint8_t control;
        fl_uint32_t lba;
};              
                        
                        
struct scsi_read6_lba    
{       
        fl_uint32_t lba;
};
 
                        
struct bytes_to_ushort_s
{                       
        fl_uint16_t value;
}; 
 
 
struct ushort_to_bytes_s
{
        fl_uint8_t bytes[2];
};           
 
 
struct bytes_to_uint_s
{       
        fl_uint32_t value;
};      
 

struct scsi_cmd_s
{
        fl_uint8_t op_code;
        void *cmd;
        void *data;
        fl_uint32_t data_length;
};


// NOTE: SCSI command structures


struct scsi_cmd_inquiry_s
{
        fl_uint8_t op_code;
        fl_uint8_t CMDDT : 1;
        fl_uint8_t EVPD : 1;
        fl_uint8_t page_code;
        fl_uint16_t alloc_length;
        fl_uint8_t control;
};


struct scsi_cmd_prevent_medium_removal_s
{
        fl_uint8_t op_code;
        fl_uint8_t prevent;
        fl_uint8_t control;
};


struct scsi_cmd_inquiry1_s
{
        fl_uint8_t peripheral_qualifier : 3;
        fl_uint8_t peripheral_devicetype : 5;
        fl_uint8_t RMB : 1;
        fl_uint8_t VERSION;
        fl_uint8_t NORMACA : 1;
        fl_uint8_t HISUP : 1;
        fl_uint8_t response_data_format : 4;
        fl_uint8_t additional_length;
        fl_uint8_t SCCS : 1;
        fl_uint8_t ACC : 1;
        fl_uint8_t TPGS : 2;
        fl_uint8_t THIRDPC : 1;
        fl_uint8_t PROTECT : 1;
        fl_uint8_t ENCSERV : 1;
        fl_uint8_t VS : 1;
        fl_uint8_t MULTIP : 1;
        fl_uint8_t CMDQUE : 1;
        char *vendor_id;
        char *product_id;
        char *product_rev;
};


// NOTE: SCSI data structures


struct scsi_data_6bytes_s
{
        fl_uint8_t bytes[6];
};


struct scsi_data_inquiry1_s
{
        fl_uint8_t bytes[96];
};


////////////////////////
// NOTE: Declarations //
////////////////////////


// NOTE: Function prototypes


int fl_scsi_check_cmd(unsigned char *,
                      fl_uint32_t,
                      struct scsi_cmd_s *);
int fl_scsi_convert(fl_uint8_t,
		    struct scsi_cmd_s *);
int fl_scsi_process(struct scsi_cmd_s *,
		    struct task_desc *);


#endif	// __FL_SCSI_H__


