

// NOTE: System header files


#if FL_OS_NAME_V_OSF1
#       if FL_OS_VERSION_V_5_1
#               if FL_OS_RELEASE_V_B


#               endif
#       endif
#elif FL_OS_NAME_V_IRIX
#       if FL_OS_VERSION_V_6_5
#               if FL_OS_RELEASE_V_30


#               endif
#       endif
#endif


// NOTE: Local header files


#include "fl_sem.h"


int fl_sem_init(fl_sem_t **sem,		// -- OUT
		fl_uint64_t value)	// -- IN
{
        // NOTE: Arguments


	// OUT: Check for semaphore
        if (sem == NULL)
        {
                return 0;	// ERROR: No semaphore
        }


	// NOTE: Initialization


	// Semaphore
        *sem = NULL;


	// NOTE: Processing


        // Allocate semaphore
        fl_mem_alloc((void **) sem,
                     sizeof(fl_sem_t),
                     FL_MEM_ZERO |
		     FL_MEM_WAITOK,
                     "fl_sem_init",
                     "Allocate semaphore");


        // Populate fields
        (*sem)->inf = FL_SEM_INF_NULL;
        (*sem)->value = value;


#if FL_OS_NAME_V_OSF1
#       if FL_OS_VERSION_V_5_1
#               if FL_OS_RELEASE_V_B


#               endif
#       endif
#elif FL_OS_NAME_V_IRIX
#       if FL_OS_VERSION_V_6_5
#               if FL_OS_RELEASE_V_30


	// Allocate semaphore lock
        (*sem)->slock = LOCK_ALLOC(-1,
                                   plhi,
                                   (lkinfo_t *) -1,
                                   KM_SLEEP);


#               endif
#       endif
#endif


        return 1;       // SUCCESS
}


int fl_sem_deinit(fl_sem_t **sem)	// -- IN/OUT
{
        // NOTE: Arguments


	// IN/OUT: Check for semaphore
        if (sem == NULL)
        {
                return 0;	// ERROR: No semaphore
        }


	// NOTE: Processing


        // Free space of semaphore
        fl_mem_free(*sem);


#if FL_OS_NAME_V_OSF1
#       if FL_OS_VERSION_V_5_1
#               if FL_OS_RELEASE_V_B


#               endif
#       endif
#elif FL_OS_NAME_V_IRIX
#       if FL_OS_VERSION_V_6_5
#               if FL_OS_RELEASE_V_30


	// Deallocate semaphore lock
        LOCK_DEALLOC((*sem)->slock);


#               endif
#       endif
#endif


        // Reset semaphore
        *sem = NULL;


        return 1;       // SUCCESS
}


int fl_sem_get_value(fl_sem_t *sem,		// -- IN
		     fl_uint64_t *value)	// -- OUT
{


#if FL_OS_NAME_V_IRIX
#       if FL_OS_VERSION_V_6_5
#               if FL_OS_RELEASE_V_30


        int pl;


#               endif
#       endif
#endif


        // NOTE: Arguments


	// IN: Check for semaphore
        if (sem == NULL)
        {
                return 0;       // ERROR: No semaphore
        }


	// OUT: Check for value
        if (value == NULL)
        {
                return -1;      // ERROR: No value
        }


	// NOTE: Processing


#if FL_OS_NAME_V_OSF1
#       if FL_OS_VERSION_V_5_1
#               if FL_OS_RELEASE_V_B


#               endif
#       endif
#elif FL_OS_NAME_V_IRIX
#       if FL_OS_VERSION_V_6_5
#               if FL_OS_RELEASE_V_30


	// Lock semaphore
        pl = LOCK(sem->slock,
		  plhi);


#               endif
#       endif
#endif


        // Set value
        *value = sem->value;


        // NOTE: Release semaphore lock


#if FL_OS_NAME_V_OSF1
#       if FL_OS_VERSION_V_5_1
#               if FL_OS_RELEASE_V_B


#               endif
#       endif
#elif FL_OS_NAME_V_IRIX
#       if FL_OS_VERSION_V_6_5
#               if FL_OS_RELEASE_V_30


	// Unlock semaphore
        UNLOCK(sem->slock,
	       pl);


#               endif
#       endif
#endif


        return 1;       // SUCCESS
}


int fl_sem_get_inf(fl_sem_t *sem,	// -- IN
		   fl_uint32_t *inf)	// -- OUT
{


#if FL_OS_NAME_V_IRIX
#       if FL_OS_VERSION_V_6_5
#               if FL_OS_RELEASE_V_30


        int pl;


#               endif
#       endif
#endif


	// NOTE: Arguments


	// IN: Check for semaphore
        if (sem == NULL)
        {
                return 0;       // ERROR: No semaphore
        }


	// OUT: Check for infinite switch
        if (inf == NULL)
        {
                return -1;      // ERROR: No infinite switch
        }


        // NOTE: Processing


#if FL_OS_NAME_V_OSF1
#       if FL_OS_VERSION_V_5_1
#               if FL_OS_RELEASE_V_B


#               endif
#       endif
#elif FL_OS_NAME_V_IRIX
#       if FL_OS_VERSION_V_6_5
#               if FL_OS_RELEASE_V_30


	// Lock semaphore
        pl = LOCK(sem->slock,
		  plhi);


#               endif
#       endif
#endif


        // Set infinite switch
        *inf = sem->inf;


#if FL_OS_NAME_V_OSF1
#       if FL_OS_VERSION_V_5_1
#               if FL_OS_RELEASE_V_B


#               endif
#       endif
#elif FL_OS_NAME_V_IRIX
#       if FL_OS_VERSION_V_6_5
#               if FL_OS_RELEASE_V_30


	// Unlock semaphore
        UNLOCK(sem->slock,
	       pl);


#               endif
#       endif
#endif


        return 1;       // SUCCESS
}


#if FL_OS_NAME_V_IRIX
#       if FL_OS_VERSION_V_6_5
#               if FL_OS_RELEASE_V_30


// NOTE: Timer callback function
//
//  In IRIX, the way to implement interruptible short
//  sleeping periods through kernel threads is using a callback
//  mechanism through the itimeout() interface.
//
//  The semaphore passed as an argument is a
//  semaphore structure unique to the invoking kernel
//  thread.


void fl_sem_timer_callback(void *lock_wait,		// -- IN
                           void *lock_cb,		// -- IN
                           unsigned int NOINTERRUPT)	// -- IN
{
        mutex_t *mutex_wait;
        mutex_t *mutex_cb;


	// NOTE: Arguments
	//
	//  This is a private ad-hoc function  and therefore,
	//  does not check its arguments.



	// NOTE: Processing


        //FL_DEBUG_PRINTF(("[fl_sem_timer_callback] INFO: lock_wait = %p\n",
			 //lock_wait))


        //FL_DEBUG_PRINTF(("[fl_sem_timer_callback] INFO: lock_cb = %p\n",
			 //lock_cb))


        //FL_DEBUG_PRINTF(("[fl_sem_timer_callback] INFO: NOINTERRUPT = %d\n",
			 //NOINTERRUPT))


        // Check for lock type
        if (NOINTERRUPT == 1)
        {
                // Set mutexes
                mutex_wait = (mutex_t *) lock_wait;
                mutex_cb = (mutex_t *) lock_cb;


next_wait:
                if (MUTEX_TRYLOCK(mutex_wait) != 0)
                {
                        // NOTE: Mutex taken by this thread


                        MUTEX_UNLOCK(mutex_wait);
                        goto next_wait;
                }


                MUTEX_LOCK(mutex_cb,
			   -1);
                MUTEX_UNLOCK(mutex_wait);
        }
        else
        {
                // Revive wait semaphore
                vsema((sema_t *) lock_wait);


                // Revive callback semaphore
                vsema((sema_t *) lock_cb);
        }


        //FL_DEBUG_PRINTF(("[fl_sem_timer_callback] INFO: vsemas done\n"))
}


#               endif
#       endif
#endif


int fl_sem_incr(fl_sem_t *sem,		// -- IN
                fl_uint64_t value,	// -- IN
                fl_uint32_t switches,	// -- IN
		struct fl_args_s *args)	// -- IN
{
        fl_int32_t ret;
        fl_int32_t overflow;
        fl_uint32_t fl_loop_forever;
	fl_uint32_t cont;
	fl_uint32_t timed_out;
	fl_uint32_t int_switch;
	fl_uint32_t ros_switch;
	fl_uint64_t maxvalue;


#if FL_OS_NAME_V_OSF1
#       if FL_OS_VERSION_V_5_1
#               if FL_OS_RELEASE_V_B


#               endif
#       endif
#elif FL_OS_NAME_V_IRIX
#       if FL_OS_VERSION_V_6_5
#               if FL_OS_RELEASE_V_30


        toid_t tid;
	fl_int32_t pl;
        fl_int32_t wait_result;
        sema_t *sem_wait;
        sema_t *sem_cb;
        mutex_t *mutex_wait;
        mutex_t *mutex_cb;


#               endif
#       endif
#endif


        // NOTE: Arguments


	// IN: Check for semaphore
        if (sem == NULL)
        {
                if (args != NULL)
                {
                        // Check for return value
                        if (args->ret.var != NULL)
                        {
                                *args->ret.var = 0;
                        }


                        // Check for reset on return
                        if (args->ret.reset == 1)
                        {
                                bzero((void *) args,
                                      sizeof(struct fl_args_s));
                        }
                }


                return 0;       // ERROR: No semaphore
        }


	// IN: Check for valid switches
	if ((switches & ~FL_SEM_SWITCH_MASK) != 0)
	{
                if (args != NULL)
                {
                        // Check for return value
                        if (args->ret.var != NULL)
                        {
                                *args->ret.var = -1;
                        }


                        // Check for reset on return
                        if (args->ret.reset == 1)
                        {
                                bzero((void *) args,
                                      sizeof(struct fl_args_s));
                        }
                }


		return -1;	// ERROR: Invalid switches
	}


	// IN: Check for args
	if (args == NULL)
	{
		return -2;	// ERROR: No args
	}


	// NOTE: Initialization


	// Return value
        ret = 1;


	// Continuation switch
        cont = 0;


	// Timed out switch
        timed_out = 0;


	// Overflow switch
        overflow = 0;


	// Maximum value of semaphore value
        maxvalue = 18446744073709551615LLU;


	// NOTE: Processing


	// Check for enable interrupt switch
	if ((switches & FL_SEM_SWITCH_ENABLE_INTERRUPT) != 0)
	{
		int_switch = 1;
	}


	// Check for return on sleep switch
	if ((switches & FL_SEM_SWITCH_RETURN_ON_SLEEP) != 0)
	{
		ros_switch = 1;
	}


#if FL_OS_NAME_V_OSF1
#       if FL_OS_VERSION_V_5_1
#               if FL_OS_RELEASE_V_B


#               endif
#       endif
#elif FL_OS_NAME_V_IRIX
#       if FL_OS_VERSION_V_6_5
#               if FL_OS_RELEASE_V_30


        // Allocate wait semaphore
        fl_mem_alloc((void **) &sem_wait,
                     sizeof(sema_t),
                     FL_MEM_ZERO |
		     FL_MEM_WAITOK,
                     "fl_sem_decr",
                     "Allocate wait semaphore");


        // Allocate callback semaphore
        fl_mem_alloc((void **) &sem_cb,
                     sizeof(sema_t),
                     FL_MEM_ZERO |
		     FL_MEM_WAITOK,
                     "fl_sem_decr",
                     "Allocate callback semaphore");


        // NOTE: IRIX semaphores vs mutexes
        //
        //  For embedded libsem handling of process signals
        //  like Ctrl-C (intr), it is necessary to implement semaphore
        //  operations like incrementation or decrementation.
        //
        //  On IRIX, when using a semaphore and interrupting once,
        //  successive calls on the same semaphore yields an immediate
        //  return value of -1.
        //
        //  This results in a de facto spinning wait loop, as opposed
        //  to an interval based loop, where the loop is only
        //  traversed once in a predefined time interval.
        //
        //  To avoid this, once the semaphore is interrupted for the first time,
        //  the code switches to a mutex instead, which does ignore any process
        //  signals. At the same time, mutexes allow the process to
        //  sleep too, what ensures the same behaviour achieved with a
        //  semaphore.


        // Allocate wait mutex
        mutex_wait = MUTEX_ALLOC(MUTEX_DEFAULT,
                                 KM_SLEEP,
                                 NULL);


        // Allocate callback mutex
        mutex_cb = MUTEX_ALLOC(MUTEX_DEFAULT,
                               KM_SLEEP,
                               NULL);


#               endif
#       endif
#endif


        // Spin
        fl_loop_forever = 1;
        while (fl_loop_forever)
        {


#if FL_OS_NAME_V_OSF1
#       if FL_OS_VERSION_V_5_1
#               if FL_OS_RELEASE_V_B


#               endif
#       endif
#elif FL_OS_NAME_V_IRIX
#       if FL_OS_VERSION_V_6_5
#               if FL_OS_RELEASE_V_30


		// Lock semaphore
                pl = LOCK(sem->slock,
			  plhi);


#               endif
#       endif
#endif


                // Check for inf
                if (sem->inf == FL_SEM_INF_INCR)
                {


#if FL_OS_NAME_V_OSF1
#       if FL_OS_VERSION_V_5_1
#               if FL_OS_RELEASE_V_B


#               endif
#       endif
#elif FL_OS_NAME_V_IRIX
#       if FL_OS_VERSION_V_6_5
#               if FL_OS_RELEASE_V_30


			// Unlock semaphore
                        UNLOCK(sem->slock,
			       pl);


#               endif
#       endif
#endif


                        // Check for return value
                        if (args->ret.var != NULL)
                        {
                                *args->ret.var = 2;
                        }


                        // Check for reset on return
                        if (args->ret.reset == 1)
                        {
                                bzero((void *) args,
                                      sizeof(struct fl_args_s));
                        }


                        return 2;       // SUCCESS
                }


                // Check for overflow
                if (value > (maxvalue - sem->value))
                {
                        overflow = 1;
                }
                else
                {
                        overflow = 0;
                        sem->value += value;
                }


#if FL_OS_NAME_V_OSF1
#       if FL_OS_VERSION_V_5_1
#               if FL_OS_RELEASE_V_B


#               endif
#       endif
#elif FL_OS_NAME_V_IRIX
#       if FL_OS_VERSION_V_6_5
#               if FL_OS_RELEASE_V_30


		// Unlock semaphore
                UNLOCK(sem->slock,
		       pl);


#               endif
#       endif
#endif


                if (overflow == 0)
                {
                        break;
                }
		else
		{
			// NOTE: Sleep condition


			if (ros_switch == 1)
			{
				// Check for return value
				if (args->ret.var != NULL)
				{
					*args->ret.var = -3;
				}


				// Check for reset on return
				if (args->ret.reset == 1)
				{
					bzero((void *) args,
					      sizeof(struct fl_args_s));
				}


				return -3;	// ERROR: Sleep condition
			}
		}


                cont = 0;
                timed_out = 0;


#if FL_OS_NAME_V_OSF1
#       if FL_OS_VERSION_V_5_1
#               if FL_OS_RELEASE_V_B


#               endif
#       endif
#elif FL_OS_NAME_V_IRIX
#       if FL_OS_VERSION_V_6_5
#               if FL_OS_RELEASE_V_30


                // Check for first interruption
		if (int_switch != 1)
                {
                        //FL_DEBUG_PRINTF(("[fl_sem_incr] INFO: itimeout()\n"))


                        // Register timeout function for mutex signalling
                        tid = itimeout(fl_sem_timer_callback,
                                       (void *) mutex_wait,
                                       drv_usectohz(1000000 / 100),
                                       plhi,
                                       (void *) mutex_cb,
                                       !int_switch);


			//FL_DEBUG_PRINTF(("[fl_sem_incr] INFO: Sleeping on mutex\n"))


                        // Deplete semaphore
                        MUTEX_LOCK(mutex_wait,
				   -1);


next_cb:
                        if (MUTEX_TRYLOCK(mutex_cb) != 0)
                        {
                                // NOTE: Mutex is taken by this thread


                                MUTEX_UNLOCK(mutex_cb);
                                goto next_cb;
                        }


                        MUTEX_UNLOCK(mutex_cb);


                        //FL_DEBUG_PRINTF(("[fl_sem_incr] INFO: Sleeping done\n"))
                }
                else
                {
                        // Initialize wait semaphore
                        initnsema(sem_wait,
                                  0,
                                  "semwait");


                        // Initialize callback semaphore
                        initnsema(sem_cb,
                                  0,
                                  "semcb");


			//FL_DEBUG_PRINTF(("[fl_sem_incr] INFO: itimeout(5 seconds)\n"))


                        // Register timeout function for semaphore signalling
                        tid = itimeout(fl_sem_timer_callback,
                                       (void *) sem_wait,
                                       drv_usectohz(1000000 / 100),
                                       plhi,
                                       (void *) sem_cb,
                                       !int_switch);


			//FL_DEBUG_PRINTF(("[fl_sem_incr] INFO: Sleeping on semaphore\n"))


                        // Deplete semaphore
                        wait_result = psema(sem_wait,
					    PUSER);


                        //FL_DEBUG_PRINTF(("[fl_sem_incr] INFO: Sleeping done (wait_result = %d)\n",
					   //wait_result))
                }


                // Check wait result
                if (wait_result == -1)
                {
                        // NOTE: Interrupted


                        //FL_DEBUG_PRINTF(("[fl_sem_incr] INFO: Interrupted\n"))


                        // Cancel pending timers
                        ret = untimeout(tid);


                        // NOTE: Timer callback
                        //
                        //  A return value of zero, 0, indicates failure.
                        //
                        //  This can occur if fl_sem_timer_callback
                        //  is either already executing or already finished.


                        // Check for cancelled timeout
                        if (ret == 0)
                        {
                                // Wait for callback signal
                                psema(sem_cb,
				      0);
                        }


                        // Check interrupt switch
                        if (int_switch != 1)
                        {
                                cont = 1;
                                ret = 1;
                        }
                        else
                        {
                                ret = -4;
                        }
                }
                else
                {
                        // NOTE: Timed-out


                        timed_out = 1;
                }


                // Check for continuation
                if (cont == 1)
                {
                        continue;
                }


                // Check if timed out
                if (timed_out == 1)
                {
                        continue;
                }


                break;


#               endif
#       endif
#endif


        }


#if FL_OS_NAME_V_IRIX
#       if FL_OS_VERSION_V_6_5
#               if FL_OS_RELEASE_V_30


        // NOTE: Free semaphores


        // Free wait semaphore
        fl_mem_free((void *) sem_wait);


        // Free callback semaphore
        fl_mem_free((void *) sem_cb);


        // Deallocate wait mutex
        MUTEX_DEALLOC(mutex_wait);


        // Deallocate callback mutex
        MUTEX_DEALLOC(mutex_cb);


#               endif
#       endif
#endif


	// Check for return value
	if (args->ret.var != NULL)
	{
		*args->ret.var = ret;
	}


	// Check for reset on return
	if (args->ret.reset == 1)
	{
		bzero((void *) args,
		      sizeof(struct fl_args_s));
	}


        //FL_DEBUG_PRINTF(("[fl_sem_incr] INFO: Sleep\n"))


        return ret;
}


int fl_sem_incr_inf(fl_sem_t *sem)	// -- IN
{


#if FL_OS_NAME_V_IRIX
#       if FL_OS_VERSION_V_6_5
#               if FL_OS_RELEASE_V_30


        int pl;


#               endif
#       endif
#endif


	// NOTE: Arguments


        // IN: Check for semaphore
        if (sem == NULL)
        {
                return 0;       // ERROR: No semaphore
        }


        // NOTE: Processing


#if FL_OS_NAME_V_OSF1
#       if FL_OS_VERSION_V_5_1
#               if FL_OS_RELEASE_V_B


#               endif
#       endif
#elif FL_OS_NAME_V_IRIX
#       if FL_OS_VERSION_V_6_5
#               if FL_OS_RELEASE_V_30


	// Lock semaphore
        pl = LOCK(sem->slock,
		  plhi);


#               endif
#       endif
#endif


        // Check for infinite decrement
        if (sem->inf == FL_SEM_INF_NULL)
        {
                sem->inf = FL_SEM_INF_INCR;
        }
        else if (sem->inf == FL_SEM_INF_DECR)
        {
                sem->inf = FL_SEM_INF_NULL;
        }


#if FL_OS_NAME_V_OSF1
#       if FL_OS_VERSION_V_5_1
#               if FL_OS_RELEASE_V_B


#               endif
#       endif
#elif FL_OS_NAME_V_IRIX
#       if FL_OS_VERSION_V_6_5
#               if FL_OS_RELEASE_V_30


	// Unlock semaphore
        UNLOCK(sem->slock,
	       pl);


#               endif
#       endif
#endif


        return 1;       // SUCCESS
}


int fl_sem_decr(fl_sem_t *sem,		// -- IN
                fl_uint64_t value,	// -- IN
                fl_uint32_t switches,	// -- IN
		struct fl_args_s *args)	// -- IN
{
        fl_int32_t ret;
        fl_int32_t underflow;
        fl_uint32_t cont;
        fl_uint32_t timed_out;
	fl_uint32_t int_switch;
	fl_uint32_t ros_switch;
        fl_uint32_t fl_loop_forever;


#if FL_OS_NAME_V_OSF1
#       if FL_OS_VERSION_V_5_1
#               if FL_OS_RELEASE_V_B


#               endif
#       endif
#elif FL_OS_NAME_V_IRIX
#       if FL_OS_VERSION_V_6_5
#               if FL_OS_RELEASE_V_30


        toid_t tid;
        fl_int32_t pl;
        fl_int32_t wait_result;
        sema_t *sem_wait;
        sema_t *sem_cb;
        mutex_t *mutex_wait;
        mutex_t *mutex_cb;


#               endif
#       endif
#endif


        // NOTE: Arguments


	// IN: Check for semaphore
        if (sem == NULL)
        {
                if (args != NULL)
                {
                        // Check for return value
                        if (args->ret.var != NULL)
                        {
                                *args->ret.var = 0;
                        }


                        // Check for reset on return
                        if (args->ret.reset == 1)
                        {
                                bzero((void *) args,
                                      sizeof(struct fl_args_s));
                        }
                }


                return 0;       // ERROR: No semaphore
        }


	// IN: Check for valid switches
	if ((switches & ~FL_SEM_SWITCH_MASK) != 0)
	{
                if (args != NULL)
                {
                        // Check for return value
                        if (args->ret.var != NULL)
                        {
                                *args->ret.var = -1;
                        }


                        // Check for reset on return
                        if (args->ret.reset == 1)
                        {
                                bzero((void *) args,
                                      sizeof(struct fl_args_s));
                        }
                }


		return -1;	// ERROR: Invalid switches
	}


	// IN: Check for args
	if (args == NULL)
	{
		return -2;	// ERROR: No args
	}


	// NOTE: Initialization


	// Return value
        ret = 1;


	// Continuation switch
        cont = 0;


	// Timed out switch
        timed_out = 0;


	// Underflow switch
        underflow = 0;


	// NOTE: Processing


	// Check for enable interrupt switch
	if ((switches & FL_SEM_SWITCH_ENABLE_INTERRUPT) != 0)
	{
		int_switch = 1;
	}


	// Check for return on sleep switch
	if ((switches & FL_SEM_SWITCH_RETURN_ON_SLEEP) != 0)
	{
		ros_switch = 1;
	}


#if FL_OS_NAME_V_OSF1
#       if FL_OS_VERSION_V_5_1
#               if FL_OS_RELEASE_V_B


#               endif
#       endif
#elif FL_OS_NAME_V_IRIX
#       if FL_OS_VERSION_V_6_5
#               if FL_OS_RELEASE_V_30


        // Allocate wait semaphore
	fl_mem_alloc((void **) &sem_wait,
                     sizeof(sema_t),
                     FL_MEM_ZERO |
		     FL_MEM_WAITOK,
                     "fl_sem_decr",
                     "Allocate wait semaphore");


        // Allocate callback semaphore
        fl_mem_alloc((void **) &sem_cb,
                     sizeof(sema_t),
                     FL_MEM_ZERO |
		     FL_MEM_WAITOK,
                     "fl_sem_decr",
                     "Allocate callback semaphore");


        // NOTE: IRIX semaphores vs mutexes
        //
        //  For embedded libsem, handling of process signals
        //  like Ctrl-C (intr) is necessary to implement semaphore
        //  operations like incrementation or decrementation.
        //
        //  In IRIX, when using a semaphore and interrupting once,
        //  successive calls on the same semaphore yields an immediate
        //  return with value -1.
        //
        //  This results in a de facto spinning wait loop, as opposed
        //  to an interval based loop, where the loop is only
        //  traversed once in a predefined time interval.
        //
        //  To avoid this, once the semaphore is interrupted, the code
        //  switches to a mutex instead, which does ignore any process
        //  signals. At the same time, mutexes allow the process to
        //  sleep too, what ensures the same behaviour achieved with a
        //  semaphore.


        // Allocate wait mutex
        mutex_wait = MUTEX_ALLOC(MUTEX_DEFAULT,
                                 KM_SLEEP,
                                 NULL);


        // Allocate callback mutex
        mutex_cb = MUTEX_ALLOC(MUTEX_DEFAULT,
                               KM_SLEEP,
                               NULL);


#               endif
#       endif
#endif


        // Loop
        fl_loop_forever = 1;
        while (fl_loop_forever)
        {


#if FL_OS_NAME_V_OSF1
#       if FL_OS_VERSION_V_5_1
#               if FL_OS_RELEASE_V_B


#               endif
#       endif
#elif FL_OS_NAME_V_IRIX
#       if FL_OS_VERSION_V_6_5
#               if FL_OS_RELEASE_V_30


		// Lock semaphore
                pl = LOCK(sem->slock,
			  plhi);


#               endif
#       endif
#endif


                // Check for infinite
                if (sem->inf == FL_SEM_INF_DECR)
                {


#if FL_OS_NAME_V_OSF1
#       if FL_OS_VERSION_V_5_1
#               if FL_OS_RELEASE_V_B


#               endif
#       endif
#elif FL_OS_NAME_V_IRIX
#       if FL_OS_VERSION_V_6_5
#               if FL_OS_RELEASE_V_30


			// Unlock semaphore
                        UNLOCK(sem->slock,
			       pl);


#               endif
#       endif
#endif


                        // Check for return value
                        if (args->ret.var != NULL)
                        {
                                *args->ret.var = 2;
                        }


                        // Check for reset on return
                        if (args->ret.reset == 1)
                        {
                                bzero((void *) args,
                                      sizeof(struct fl_args_s));
                        }


                        return 2;       // SUCCESS
                }


                if (value > sem->value)
                {
                        underflow = 1;
                }
                else
                {
                        underflow = 0;
                        sem->value -= value;
                        ret = 1;
                }


#if FL_OS_NAME_V_OSF1
#       if FL_OS_VERSION_V_5_1
#               if FL_OS_RELEASE_V_B


#               endif
#       endif
#elif FL_OS_NAME_V_IRIX
#       if FL_OS_VERSION_V_6_5
#               if FL_OS_RELEASE_V_30


		// Unlock semaphore
                UNLOCK(sem->slock,
		       pl);


#               endif
#       endif
#endif


                if (underflow == 0)
                {
                        break;
                }
		else
		{
			// NOTE: Sleep condition


			if (ros_switch == 1)
			{
				// Check for return value
				if (args->ret.var != NULL)
				{
					*args->ret.var = -3;
				}


				// Check for reset on return
				if (args->ret.reset == 1)
				{
					bzero((void *) args,
					      sizeof(struct fl_args_s));
				}


				return -3;	// ERROR: Sleep condition
			}
		}


                cont = 0;
                timed_out = 0;


#if FL_OS_NAME_V_OSF1
#       if FL_OS_VERSION_V_5_1
#               if FL_OS_RELEASE_V_B


#               endif
#       endif
#elif FL_OS_NAME_V_IRIX
#       if FL_OS_VERSION_V_6_5
#               if FL_OS_RELEASE_V_30


                // Check for first interruption
                if (int_switch != 1)
                {
                        //FL_DEBUG_PRINTF(("[fl_sem_decr] INFO: itimeout()\n"))


                        // Register timeout function for mutex signalling
                        tid = itimeout(fl_sem_timer_callback,
                                       (void *) mutex_wait,
                                       drv_usectohz(1000000 / 100),
                                       plhi,
                                       (void *) mutex_cb,
                                       !int_switch);


			//FL_DEBUG_PRINTF(("[fl_sem_decr] INFO: Sleeping on mutex\n"))


                        // Deplete semaphore
                        MUTEX_LOCK(mutex_wait,
				   -1);


next_cb:
                        if (MUTEX_TRYLOCK(mutex_cb) != 0)
                        {
                                // NOTE: Mutex is taken by this thread


                                MUTEX_UNLOCK(mutex_cb);
                                goto next_cb;
                        }


                        MUTEX_UNLOCK(mutex_cb);


                        //FL_DEBUG_PRINTF(("[fl_sem_decr] INFO: Sleeping done\n"))
                }
                else
                {
                        // Initialize wait semaphore
                        initnsema(sem_wait,
                                  0,
                                  "semwait");


                        // Initialize callback semaphore
                        initnsema(sem_cb,
                                  0,
                                  "semcb");


			//FL_DEBUG_PRINTF(("[fl_sem_decr] INFO: itimeout(5 seconds)\n"))


                        // Register timeout function for semaphore signalling
                        tid = itimeout(fl_sem_timer_callback,
                                       (void *) sem_wait,
                                       drv_usectohz(1000000 / 100),
                                       plhi,
                                       (void *) sem_cb,
                                       !int_switch);


			//FL_DEBUG_PRINTF(("[fl_sem_decr] INFO: Sleeping on semaphore\n"))


                        // Deplete semaphore
                        wait_result = psema(sem_wait,
					    PUSER);


			//FL_DEBUG_PRINTF(("[fl_sem_decr] INFO: Sleeping done (wait_result = %d)\n",
					 //wait_result))


                        // Free wait semaphore
                }


                // Check wait result
                if (wait_result == -1)
                {
                        // NOTE: Interrupted


			//FL_DEBUG_PRINTF(("[fl_sem_decr] INFO: Interrupted\n"))


                        // Cancel pending timers
                        ret = untimeout(tid);


                        // NOTE: Timer callback
                        //
                        //  A return value of zero, 0, indicates failure.
                        //
                        //  This can occur if fl_sem_timer_callback
                        //  is either already executing or already finished.


                        // Check for cancelled timeout
                        if (ret == 0)
                        {
                                // Wait for callback signal
                                psema(sem_cb,
				      0);
                        }


                        // Check interruption flag
                        if (int_switch != 1)
                        {
                                cont = 1;
                                ret = 1;
                        }
                        else
                        {
                                ret = -4;
                        }
                }
                else
                {
                        // NOTE: Timed-out


                        timed_out = 1;
                }


#               endif
#       endif
#endif


                // Check for continuation
                if (cont == 1)
                {
                        continue;
                }


                // Check if timed out
                if (timed_out == 1)
                {
                        continue;
                }


                break;
        }


#if FL_OS_NAME_V_IRIX
#       if FL_OS_VERSION_V_6_5
#               if FL_OS_RELEASE_V_30


	// NOTE: Free semaphores


        // Free wait semaphore
        fl_mem_free((void *) sem_wait);


        // Free callback semaphore
        fl_mem_free((void *) sem_cb);


        // Deallocate wait mutex
        MUTEX_DEALLOC(mutex_wait);


        // Deallocate callback mutex
        MUTEX_DEALLOC(mutex_cb);


#               endif
#       endif
#endif


	// Check for return value
	if (args->ret.var != NULL)
	{
		*args->ret.var = ret;
	}


	// Check for reset on return
	if (args->ret.reset == 1)
	{
		bzero((void *) args,
		      sizeof(struct fl_args_s));
	}


	//FL_DEBUG_PRINTF(("[fl_sem_decr] INFO: STOP\n"))


        return ret;
}


int fl_sem_decr_inf(fl_sem_t *sem)	// -- IN
{


#if FL_OS_NAME_V_IRIX
#       if FL_OS_VERSION_V_6_5
#               if FL_OS_RELEASE_V_30


        int pl;


#               endif
#       endif
#endif


	// NOTE: Arguments


        // IN: Check for semaphore
        if (sem == NULL)
        {
                return 0;       // ERROR: No semaphore
        }


#if FL_OS_NAME_V_OSF1
#       if FL_OS_VERSION_V_5_1
#               if FL_OS_RELEASE_V_B


#               endif
#       endif
#elif FL_OS_NAME_V_IRIX
#       if FL_OS_VERSION_V_6_5
#               if FL_OS_RELEASE_V_30


	// Lock semaphore
        pl = LOCK(sem->slock,
		  plhi);


#               endif
#       endif
#endif


        // Check for infinite increment
        if (sem->inf == FL_SEM_INF_NULL)
        {
                sem->inf = FL_SEM_INF_DECR;
        }
        else if (sem->inf == FL_SEM_INF_INCR)
        {
                sem->inf = FL_SEM_INF_NULL;
        }


#if FL_OS_NAME_V_OSF1
#       if FL_OS_VERSION_V_5_1
#               if FL_OS_RELEASE_V_B


#               endif
#       endif
#elif FL_OS_NAME_V_IRIX
#       if FL_OS_VERSION_V_6_5
#               if FL_OS_RELEASE_V_30


	// Unlock semaphore
	UNLOCK(sem->slock, pl);


#               endif
#       endif
#endif


        return 1;       // SUCCESS
}


