

// NOTE: System header files


#if FL_OS_NAME_V_OSF1
#       if FL_OS_VERSION_V_5_1
#               if FL_OS_RELEASE_V_B


#               endif
#       endif
#elif FL_OS_NAME_V_IRIX
#       if FL_OS_VERSION_V_6_5
#               if FL_OS_RELEASE_V_30


#               endif
#       endif
#endif


// NOTE: Local header files


#include "fl_spin.h"


int fl_spin_new(struct fl_spin_s **spin)	// -- OUT
{
	// NOTE: Arguments


	// OUT: Check for spinlock
	if (spin == NULL)
	{
		return 0;	// ERROR: No spinlock
	}


	// NOTE: Processing


	// Allocate spinlock
        fl_mem_alloc((void **) spin,
                     sizeof(struct fl_spin_s),
                     FL_MEM_WAITOK |
		     FL_MEM_ZERO,
                     "fl_spin_new",
                     "Allocate spinlock");


#if FL_OS_NAME_V_OSF1
#       if FL_OS_VERSION_V_5_1
#               if FL_OS_RELEASE_V_B


#               endif
#       endif
#elif FL_OS_NAME_V_IRIX
#       if FL_OS_VERSION_V_6_5
#               if FL_OS_RELEASE_V_30


	// Allocate lock
        (*spin)->lock = LOCK_ALLOC(-1,
                                   plhi,
                                   (lkinfo_t *) -1,
                                   KM_SLEEP);


#               endif
#       endif
#endif


	return 1;	// SUCCESS
}


int fl_spin_free(struct fl_spin_s *spin)	// -- IN
{
	// NOTE: Arguments


	// IN: Check for spinlock
	if (spin == NULL)
	{
		return 0;	// ERROR: No spinlock
	}


	// NOTE: Processing


#if FL_OS_NAME_V_OSF1
#       if FL_OS_VERSION_V_5_1
#               if FL_OS_RELEASE_V_B


#               endif
#       endif
#elif FL_OS_NAME_V_IRIX
#       if FL_OS_VERSION_V_6_5
#               if FL_OS_RELEASE_V_30


	// Deallocate lock
	LOCK_DEALLOC(spin->lock);


#               endif
#       endif
#endif


	// Free spinlock
	fl_mem_free(spin);


        return 1;       // SUCCESS
}


int fl_spin_lock(struct fl_spin_s *spin)	// -- IN
{
	// NOTE: Arguments


	// IN: Check for spinlock
	if (spin == NULL)
	{
		return 0;	// ERROR: No spinlock
	}


	// NOTE: Processing


#if FL_OS_NAME_V_OSF1
#       if FL_OS_VERSION_V_5_1
#               if FL_OS_RELEASE_V_B


#               endif
#       endif
#elif FL_OS_NAME_V_IRIX
#       if FL_OS_VERSION_V_6_5
#               if FL_OS_RELEASE_V_30


	// Lock spinlock
	spin->pl = LOCK(spin->lock,
			plhi);


#               endif
#       endif
#endif


	return 1;	// SUCCESS
}


int fl_spin_unlock(struct fl_spin_s *spin)	// -- IN
{
	// NOTE: Arguments


	// IN: Check for spinlock
	if (spin == NULL)
	{
		return 0;	// ERROR: No spinlock
	}


	// NOTE: Processing


#if FL_OS_NAME_V_OSF1
#       if FL_OS_VERSION_V_5_1
#               if FL_OS_RELEASE_V_B


#               endif
#       endif
#elif FL_OS_NAME_V_IRIX
#       if FL_OS_VERSION_V_6_5
#               if FL_OS_RELEASE_V_30


	// Unlock spinlock
	UNLOCK(spin->lock,
	       spin->pl);


#               endif
#       endif
#endif


	return 1;	// SUCCESS
}


int fl_spin_trylock(struct fl_spin_s *spin,	// -- IN
		    fl_uint32_t *is_locked)	// -- OUT
{
	// NOTE: Arguments


	// IN: Check for spinlock
	if (spin == NULL)
	{
		return 0;	// ERROR: No spinlock
	}


	// OUT: Check for is locked switch
	if (is_locked == NULL)
	{
		return -1;	// ERROR: No is locked switch
	}


	// NOTE: Processing


#if FL_OS_NAME_V_OSF1
#       if FL_OS_VERSION_V_5_1
#               if FL_OS_RELEASE_V_B


#               endif
#       endif
#elif FL_OS_NAME_V_IRIX
#       if FL_OS_VERSION_V_6_5
#               if FL_OS_RELEASE_V_30


	// Try to lock semaphore
	*is_locked = 0;
	spin->pl = TRYLOCK(spin->lock,
			   plhi);
	if (spin->pl != INVPL)
	{
		*is_locked = 1;
	}


#               endif
#       endif
#endif


	return 1;	// SUCCESS
}


