

#include <unistd.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <string.h>
#include <stdlib.h>
#include <limits.h>
#include <errno.h>

#include "fl_test.h"


unsigned long long fl_test_env;
char *fl_test_env_function_name;
char *fl_test_env_test_name;


void fl_test_init()
{
	//int ret;
	char *val;
	char *colon;
	unsigned long long len;
	char *function_name;
	char *test_name;


	// Check environment variables
	if ((val= getenv("FL_TEST")) != NULL)
	{
		// Get index of first ':'
		if ((colon = strchr(val, ':')) == NULL)
		{
			printf("[fl_test.c][fl_test_init] ERROR: Wrong test name format, must be: function_name:test_name\n");
			exit(-1);
		}


		// Get test function name
		len = colon - val;
		//printf("function name len = %llu\n", len);
		if (len == 0)
		{
			printf("[fl_test.c][fl_test_init] ERROR: Empty test function name\n");
			exit(-1);
		}


		// Allocate space for function name
		if ((function_name = (char *) malloc(len + 1)) == NULL)
		{
			printf("[fl_test.c][fl_test_init] ERROR: Couldn't allocate space for test function name\n");
			exit(-1);
		}


		// Get test name
		//colon += 1;
		len = strlen(colon + 1);
		//printf("test name len = %llu\n", len);
		if (len == 0)
		{
			free(function_name);
			printf("[fl_test.c][fl_test_init] ERROR: Empty test name\n");
                        exit(-1);
		}


		// Allocate space for test name
                if ((test_name = (char *) malloc(len + 1)) == NULL)
                {
			free(function_name);
                        printf("[fl_test.c][fl_test_init] ERROR: Couldn't allocate space for test name\n");
                        exit(-1);
                }


		//printf("val = %s\n", val);

		*colon = ' ';
		//if ((ret = sscanf(val, "%s %s", function_name, test_name)) != 2)
		if (sscanf(val, "%s %s", function_name, test_name) != 2)
		{
			//printf("sprintf.ret = %i\n", ret);
			free(function_name);
			free(test_name);
			printf("[fl_test.c][fl_test_init] ERROR: Couldn't parse test environment variable\n");
			exit(-1);
		}
		*colon = ':';

		fl_test_env = 1;
		fl_test_env_function_name = function_name;
		fl_test_env_test_name = test_name;
	}
	else
	{
		fl_test_env = 0;
		fl_test_env_function_name = NULL;
		fl_test_env_test_name = NULL;
	}


	//printf("[fl_test.c][fl_test_init] fl_test_env = %u\n", fl_test_env);
        //printf("[fl_test.c][fl_test_init] fl_test_num = %llu\n", fl_test_num);
	//printf("[fl_test.c][fl_test_init] fl_test_env_functon_name = %s\n", fl_test_env_function_name);
	//printf("[fl_test.c][fl_test_init] fl_test_env_test_name = %s\n", fl_test_env_test_name);
}


void fl_test_deinit()
{
	free(fl_test_env_function_name);
	free(fl_test_env_test_name);
}


int fl_test_check_this_test(char *function_name,
                            char *test_name)
{
	if ((strcmp(function_name,
                    fl_test_env_function_name) != 0) || 
	    (strcmp(test_name,
		    fl_test_env_test_name) != 0))
	{
		return 0;	// ERROR: Test doesn't match
	}


	return 1;	// SUCCESS
}


// TEST FUNCTIONS


%GLOB:fl_test:START%


%FUNC:test_functions:START%


int fl_test_f_main_n_NOT_SUSER()
{
        if (fl_test_env == 1)
        {
                if (fl_test_check_this_test("main", "NOT_SUSER") != 1)
                {
                        return INT_MAX;       // SUCCESS: Unrequested test
                }
        }
        else
        {
                return INT_MAX;       // SUCCESS: Unrequested test
        }


        //return 0;       // SUCCESS
        return %ufl_c:main:NOT_SUSER[on,,]%;       // SUCCESS
}


int fl_test_f_main_n_ATTACH_EACCES_DEV()
{
        if (fl_test_env == 1)
        {
                if (fl_test_check_this_test("main", "ATTACH_EACCES_DEV") != 1)
                {
                        return INT_MAX;       // SUCCESS: Unrequested test
                }
        }
        else
        {
                return INT_MAX;       // SUCCESS: Unrequested test
        }

	printf("[fl_test_f_main_n_ATTACH_EACCES_DEV] START\n");

	errno = EACCES;


        //return -10;       // SUCCESS
        return %ufl_c:main:ATTACH_EACCES_DEV[on,,]%;       // SUCCESS
}


int fl_test_f_main_n_ATTACH_EIO_DEV()
{
        if (fl_test_env == 1)
        {
                if (fl_test_check_this_test("main", "ATTACH_EIO_DEV") != 1)
                {
                        return INT_MAX;       // SUCCESS: Unrequested test
                }
        }
        else
        {
                return INT_MAX;       // SUCCESS: Unrequested test
        }


	errno = EIO;


        //return -11;       // SUCCESS
        return %ufl_c:main:ATTACH_EIO_DEV[on,,]%;       // SUCCESS
}


int fl_test_f_main_n_ATTACH_ELOOP_DEV()
{
        if (fl_test_env == 1)
        {
                if (fl_test_check_this_test("main", "ATTACH_ELOOP_DEV") != 1)
                {
                        return INT_MAX;       // SUCCESS: Unrequested test
                }
        }
        else
        {
                return INT_MAX;       // SUCCESS: Unrequested test
        }


	errno = ELOOP;


        //return -12;       // SUCCESS
        return %ufl_c:main:ATTACH_ELOOP_DEV[on,,]%;       // SUCCESS
}


int fl_test_f_main_n_ATTACH_ENOMEM_DEV()
{
        if (fl_test_env == 1)
        {
                if (fl_test_check_this_test("main", "ATTACH_ENOMEM_DEV") != 1)
                {
                        return INT_MAX;       // SUCCESS: Unrequested test
                }
        }
        else
        {
                return INT_MAX;       // SUCCESS: Unrequested test
        }


	errno = ENOMEM;


        //return -15;       // SUCCESS
        return %ufl_c:main:ATTACH_ENOMEM_DEV[on,,]%;       // SUCCESS
}


int fl_test_f_main_n_ATTACH_ENOTDIR_DEV()
{
        if (fl_test_env == 1)
        {
                if (fl_test_check_this_test("main", "ATTACH_ENOTDIR_DEV") != 1)
                {
                        return INT_MAX;       // SUCCESS: Unrequested test
                }
        }
        else
        {
                return INT_MAX;       // SUCCESS: Unrequested test
        }


	errno = ENOTDIR;


        //return -16;       // SUCCESS
        return %ufl_c:main:ATTACH_ENOTDIR_DEV[on,,]%;       // SUCCESS
}


int fl_test_f_main_n_ATTACH_EACCES_FILE()
{
        if (fl_test_env == 1)
        {
                if (fl_test_check_this_test("main", "ATTACH_EACCES_FILE") != 1)
                {
                        return INT_MAX;       // SUCCESS: Unrequested test
                }
        }
        else
        {
                return INT_MAX;       // SUCCESS: Unrequested test
        }

        printf("[fl_test_f_main_n_ATTACH_EACCES_FILE] START\n");

        errno = EACCES;


        //return -18;       // SUCCESS
        return %ufl_c:main:ATTACH_EACCES_FILE[on,,]%;       // SUCCESS
}


int fl_test_f_main_n_ATTACH_EIO_FILE()
{
        if (fl_test_env == 1)
        {
                if (fl_test_check_this_test("main", "ATTACH_EIO_FILE") != 1)
                {
                        return INT_MAX;       // SUCCESS: Unrequested test
                }
        }
        else
        {
                return INT_MAX;       // SUCCESS: Unrequested test
        }


        errno = EIO;


        //return -19;       // SUCCESS
        return %ufl_c:main:ATTACH_EIO_FILE[on,,]%;       // SUCCESS
}


int fl_test_f_main_n_ATTACH_ELOOP_FILE()
{
        if (fl_test_env == 1)
        {
                if (fl_test_check_this_test("main", "ATTACH_ELOOP_FILE") != 1)
                {
                        return INT_MAX;       // SUCCESS: Unrequested test
                }
        }
        else
        {
                return INT_MAX;       // SUCCESS: Unrequested test
        }


        errno = ELOOP;


        //return -20;       // SUCCESS
        return %ufl_c:main:ATTACH_ELOOP_FILE[on,,]%;       // SUCCESS
}


int fl_test_f_main_n_ATTACH_ENOMEM_FILE()
{
        if (fl_test_env == 1)
        {
                if (fl_test_check_this_test("main", "ATTACH_ENOMEM_FILE") != 1)
                {
                        return INT_MAX;       // SUCCESS: Unrequested test
                }
        }
        else
        {
                return INT_MAX;       // SUCCESS: Unrequested test
        }


        errno = ENOMEM;


        //return -23;       // SUCCESS
        return %ufl_c:main:ATTACH_ENOMEM_FILE[on,,]%;       // SUCCESS
}


int fl_test_f_main_n_ATTACH_ENOTDIR_FILE()
{
        if (fl_test_env == 1)
        {
                if (fl_test_check_this_test("main", "ATTACH_ENOTDIR_FILE") != 1)
                {
                        return INT_MAX;       // SUCCESS: Unrequested test
                }
        }
        else
        {
                return INT_MAX;       // SUCCESS: Unrequested test
        }


        errno = ENOTDIR;


        //return -24;       // SUCCESS
        return %ufl_c:main:ATTACH_ENOTDIR_FILE[on,,]%;       // SUCCESS
}


int fl_test_f_main_n_DETACH_EACCES()
{
        if (fl_test_env == 1)
        {
                if (fl_test_check_this_test("main", "DETACH_EACCES") != 1)
                {
                        return INT_MAX;       // SUCCESS: Unrequested test
                }
        }
        else
        {
                return INT_MAX;       // SUCCESS: Unrequested test
        }

        printf("[fl_test_f_main_n_DETACH_EACCES] START\n");

        errno = EACCES;


        //return -25;       // SUCCESS
        return %ufl_c:main:DETACH_EACCES[on,,]%;       // SUCCESS
}


int fl_test_f_main_n_DETACH_EIO()
{
        if (fl_test_env == 1)
        {
                if (fl_test_check_this_test("main", "DETACH_EIO") != 1)
                {
                        return INT_MAX;       // SUCCESS: Unrequested test
                }
        }
        else
        {
                return INT_MAX;       // SUCCESS: Unrequested test
        }


        errno = EIO;


        //return -26;       // SUCCESS
        return %ufl_c:main:DETACH_EIO[on,,]%;       // SUCCESS
}


int fl_test_f_main_n_DETACH_ELOOP()
{
        if (fl_test_env == 1)
        {
                if (fl_test_check_this_test("main", "DETACH_ELOOP") != 1)
                {
                        return INT_MAX;       // SUCCESS: Unrequested test
                }
        }
        else
        {
                return INT_MAX;       // SUCCESS: Unrequested test
        }


        errno = ELOOP;


        //return -27;       // SUCCESS
        return %ufl_c:main:DETACH_ELOOP[on,,]%;       // SUCCESS
}


int fl_test_f_main_n_DETACH_ENOMEM()
{
        if (fl_test_env == 1)
        {
                if (fl_test_check_this_test("main", "DETACH_ENOMEM") != 1)
                {
                        return INT_MAX;       // SUCCESS: Unrequested test
                }
        }
        else
        {
                return INT_MAX;       // SUCCESS: Unrequested test
        }


        errno = ENOMEM;


        //return -30;       // SUCCESS
        return %ufl_c:main:DETACH_ENOMEM[on,,]%;       // SUCCESS
}


int fl_test_f_main_n_DETACH_ENOTDIR()
{
        if (fl_test_env == 1)
        {
                if (fl_test_check_this_test("main", "DETACH_ENOTDIR") != 1)
                {
                        return INT_MAX;       // SUCCESS: Unrequested test
                }
        }
        else
        {
                return INT_MAX;       // SUCCESS: Unrequested test
        }


        errno = ENOTDIR;


        //return -31;       // SUCCESS
        return %ufl_c:main:DETACH_ENOTDIR[on,,]%;       // SUCCESS
}


int fl_test_f_main_n_NOT_ALLOC_RESPONSE_PACKET()
{
        if (fl_test_env == 1)
        {
                if (fl_test_check_this_test("main", "NOT_ALLOC_RESPONSE_PACKET") != 1)
                {
                        return 0;       // SUCCESS: Unrequested test
                }
        }
        else
        {
                return 0;       // SUCCESS: Unrequested test
        }


        //return -39;       // SUCCESS
        return %ufl_c:main:NOT_ALLOC_RESPONSE_PACKET[on,,]%;       // SUCCESS
}


int fl_test_f_main_n_NO_PERMISSION_PACKET_DATA()
{
        if (fl_test_env == 1)
        {
                if (fl_test_check_this_test("main", "NO_PERMISSION_PACKET_DATA") != 1)
                {
                        return 0;       // SUCCESS: Unrequested test
                }
        }
        else
        {
                return 0;       // SUCCESS: Unrequested test
        }


	errno = EACCES;


        //return -40;       // SUCCESS
        return %ufl_c:main:NO_PERMISSION_PACKET_DATA[on,,]%;       // SUCCESS
}


int fl_test_f_main_n_NOT_PARSE_RESPONSE_PACKET()
{
        if (fl_test_env == 1)
        {
                if (fl_test_check_this_test("main", "NOT_PARSE_RESPONSE_PACKET") != 1)
                {
                        return 0;       // SUCCESS: Unrequested test
                }
        }
        else
        {
                return 0;       // SUCCESS: Unrequested test
        }


        //return -42;       // SUCCESS
        return %ufl_c:main:NOT_PARSE_RESPONSE_PACKET[on,,]%;       // SUCCESS
}


%FUNC:test_functions:STOP%


%GLOB:fl_test:STOP%


