

#ifndef __FL_TEST_H__
#define __FL_TEST_H__


#include "fl_conf_id.h"


// FL_TEST is a marker for test cases
#define FL_TEST(FUNC, NAME)


// FL_TEST_MCOMMENT() allows commenting inside a macro definition
#define FL_TEST_MCOMMENT(C)


#ifdef FL_TESTS


// NOTE: Global declarations


#define FL_TEST_DECLS \
extern char fl_test_func[300];\
extern char fl_test_name[300];\
extern char fl_test_err[300];\
extern char fl_test_suc[300];\
extern fl_int32_t fl_test_ret;\
extern fl_int32_t fl_test_notify;\
extern fl_uint64_t fl_test_counter;\
extern fl_uint64_t fl_test_ctrl_sem;


#define FL_TEST_STATE_BEFORE 0
#define FL_TEST_STATE_ERROR 1
#define FL_TEST_STATE_AFTER 2


#define FL_TEST_NO_ARGS NULL


// NOTE: Local declarations


//#define FL_TEST_TYPE_DECL int fl_test_type;
#define FL_TEST_STATE_DECL int fl_test_state;
#define FL_TEST_STATE_INIT fl_test_state = 0;


#define FL_TEST_COND(COND, OP) COND ## OP


// NOTE: Test functions
//
//  There are two type of tests, depending on the test source:
//
//    User command (also called foreign tests)
//    Kernel module


// Driver foreign test functions (fl command)
#define FL_TEST_F_main_N_NOT_SUSER_OPEN_DEVFL(TYPE, RET) \
if ((strcmp(fl_test_func, "main") == 0) &&\
    (strcmp(fl_test_name, "NOT_SUSER_OPEN_DEVFL") == 0))\
{\
        return EUSERS;\
}


#define FL_TEST_F_main_N_NO_PERMISSION_CPU_CHECK(TYPE, RET) \
if ((strcmp(fl_test_func, "main") == 0) &&\
    (strcmp(fl_test_name, "NO_PERMISSION_CPU_CHECK") == 0))\
{\
        return EACCES;\
}


#define FL_TEST_F_main_N_NOT_OPEN_DEVFL(TYPE, RET) \
if ((strcmp(fl_test_func, "main") == 0) &&\
    (strcmp(fl_test_name, "NOT_OPEN_DEVFL") == 0))\
{\
        return EIO;\
}


#define FL_TEST_F_main_N_NO_PERMISSION_WRITE_REQUEST_PACKET(TYPE, RET) \
if ((strcmp(fl_test_func, "main") == 0) &&\
    (strcmp(fl_test_name, "NO_PERMISSION_WRITE_REQUEST_PACKET") == 0))\
{\
        return EACCES;\
}


#define FL_TEST_F_main_N_NOT_WRITE_REQUEST_PACKET_DEVFL(TYPE, RET) \
if ((strcmp(fl_test_func, "main") == 0) &&\
    (strcmp(fl_test_name, "NOT_WRITE_REQUEST_PACKET_DEVFL") == 0))\
{\
        return EIO;\
}


#define FL_TEST_F_main_N_NO_PERMISSION_PACKET_SIZE(TYPE, RET) \
if ((strcmp(fl_test_func, "main") == 0) &&\
    (strcmp(fl_test_name, "NO_PERMISSION_PACKET_SIZE") == 0))\
{\
        return EACCES;\
}


#define FL_TEST_F_main_N_NOT_READ_RESPONSE_PACKET_SIZE_DEVFL(TYPE, RET) \
if ((strcmp(fl_test_func, "main") == 0) &&\
    (strcmp(fl_test_name, "NOT_READ_RESPONSE_PACKET_SIZE_DEVFL") == 0))\
{\
        return EIO;\
}


#define FL_TEST_F_main_N_NO_PERMISSION_PACKET_DATA(TYPE, RET) \
if ((strcmp(fl_test_func, "main") == 0) &&\
    (strcmp(fl_test_name, "NO_PERMISSION_PACKET_DATA") == 0))\
{\
	if (fl_test_passes++ == 1)\
	{\
		fl_test_passes = 0;\
        	return EACCES;\
	}\
}


#define FL_TEST_F_main_N_NOT_READ_RESPONSE_PACKET_DEVFL(TYPE, RET) \
if ((strcmp(fl_test_func, "main") == 0) &&\
    (strcmp(fl_test_name, "NOT_READ_RESPONSE_PACKET_DEVFL") == 0))\
{\
	if (fl_test_passes++ == 1)\
        {\
                fl_test_passes = 0;\
                return EIO;\
        }\
}


#define FL_TEST_F_main_N_NOT_ALLOC_LOOPENTRY(TYPE, RET) \
if ((strcmp(fl_test_func, "main") == 0) &&\
    (strcmp(fl_test_name, "NOT_ALLOC_LOOPENTRY") == 0))\
{\
	fl_test_type = TYPE;\
        if (fl_test_type == 0)\
        {\
                fl_test_ret = RET;\
        }\
        else if (fl_test_type == 1)\
        {\
                fl_test_ret = RET;\
        }\
}


#define FL_TEST_F_main_N_NOT_CREATE_CHR_PIDLIST(TYPE, RET) \
if ((strcmp(fl_test_func, "main") == 0) &&\
    (strcmp(fl_test_name, "NOT_CREATE_CHR_PIDLIST") == 0))\
{\
	fl_test_type = TYPE;\
	if (fl_test_type == 0)\
        {\
                fl_test_ret = RET;\
        }\
        else if (fl_test_type == 1)\
        {\
                fl_test_ret = RET;\
        }\
}


#define FL_TEST_F_main_N_NOT_CREATE_BLK_PIDLIST(TYPE, RET) \
if ((strcmp(fl_test_func, "main") == 0) &&\
    (strcmp(fl_test_name, "NOT_CREATE_BLK_PIDLIST") == 0))\
{\
	fl_test_type = TYPE;\
	if (fl_test_type == 0)\
        {\
                fl_test_ret = RET;\
        }\
        else if (fl_test_type == 1)\
        {\
                fl_test_ret = RET;\
        }\
}


#define FL_TEST_F_main_N_NOT_ALLOC_VATTR(TYPE, RET) \
if ((strcmp(fl_test_func, "main") == 0) &&\
    (strcmp(fl_test_name, "NOT_ALLOC_VATTR") == 0))\
{\
	fl_test_type = TYPE;\
	if (fl_test_type == 0)\
        {\
                fl_test_ret = RET;\
        }\
        else if (fl_test_type == 1)\
        {\
                fl_test_ret = RET;\
        }\
}


#define FL_TEST_F_main_N_NOT_ALLOC_CHR_DEV_NAME(TYPE, RET) \
if ((strcmp(fl_test_func, "main") == 0) &&\
    (strcmp(fl_test_name, "NOT_ALLOC_CHR_DEV_NAME") == 0))\
{\
	fl_test_type = TYPE;\
	if (fl_test_type == 0)\
        {\
                fl_test_ret = RET;\
        }\
        else if (fl_test_type == 1)\
        {\
                fl_test_ret = RET;\
                strcpy(fl_test_err, "ERROR: Couldn't allocate CHR device file name");\
        }\
}


#define FL_TEST_F_main_N_NOT_ALLOC_UTASK(TYPE, RET) \
if ((strcmp(fl_test_func, "main") == 0) &&\
    (strcmp(fl_test_name, "NOT_ALLOC_UTASK") == 0))\
{\
	fl_test_type = TYPE;\
	if (fl_test_type == 0)\
        {\
                fl_test_ret = RET;\
        }\
        else if (fl_test_type == 1)\
        {\
                fl_test_ret = RET;\
        }\
}


#define FL_TEST_F_main_N_NOT_ALLOC_NAMEIDATA(TYPE, RET) \
if ((strcmp(fl_test_func, "main") == 0) &&\
    (strcmp(fl_test_name, "NOT_ALLOC_NAMEIDATA") == 0))\
{\
	fl_test_type = TYPE;\
	if (fl_test_type == 0)\
        {\
                fl_test_ret = RET;\
        }\
        else if (fl_test_type == 1)\
        {\
                fl_test_ret = RET;\
        }\
}


#define FL_TEST_F_main_N_NOT_DUP_NAMEIDATA(TYPE, RET) \
if ((strcmp(fl_test_func, "main") == 0) &&\
    (strcmp(fl_test_name, "NOT_DUP_NAMEIDATA") == 0))\
{\
	fl_test_type = TYPE;\
	if (fl_test_type == 0)\
        {\
                fl_test_ret = RET;\
        }\
        else if (fl_test_type == 1)\
        {\
                fl_test_ret = RET;\
        }\
}


#define FL_TEST_F_main_N_BLK_DEV_ALREADY_ATTACHED(TYPE, RET) \
if ((strcmp(fl_test_func, "main") == 0) &&\
    (strcmp(fl_test_name, "BLK_DEV_ALREADY_ATTACHED") == 0))\
{\
	fl_test_type = TYPE;\
        if (fl_test_type == 0)\
        {\
                is_attached = RET;\
        }\
        else if (fl_test_type == 1)\
        {\
                is_attached = 0;\
                fl_test_ret = RET;\
        }\
}


#define FL_TEST_F_main_N_NOT_LOOKUP_FILENAME(TYPE, RET) \
if ((strcmp(fl_test_func, "main") == 0) &&\
    (strcmp(fl_test_name, "NOT_LOOKUP_FILENAME") == 0))\
{\
	fl_test_type = TYPE;\
        if (fl_test_type == 0)\
        {\
                fl_test_ret = RET;\
        }\
        else if (fl_test_type == 1)\
        {\
                fl_test_ret = RET;\
        }\
}


#define FL_TEST_F_main_N_NOT_FILE_STAT(TYPE, RET) \
if ((strcmp(fl_test_func, "main") == 0) &&\
    (strcmp(fl_test_name, "NOT_FILE_STAT") == 0))\
{\
	fl_test_type = TYPE;\
        if (fl_test_type == 0)\
        {\
                fl_test_ret = RET;\
        }\
        else if (fl_test_type == 1)\
        {\
                fl_test_ret = RET;\
        }\
}


#define FL_TEST_F_main_N_NOT_OPEN_FILE(TYPE, RET) \
if ((strcmp(fl_test_func, "main") == 0) &&\
    (strcmp(fl_test_name, "NOT_OPEN_FILE") == 0))\
{\
	fl_test_type = TYPE;\
        if (fl_test_type == 0)\
        {\
                fl_test_ret = RET;\
        }\
        else if (fl_test_type == 1)\
        {\
                fl_test_ret = RET;\
        }\
}


#define FL_TEST_F_main_N_NOT_BLK_DEV_LOOKUP(TYPE, RET) \
if ((strcmp(fl_test_func, "main") == 0) &&\
    (strcmp(fl_test_name, "NOT_BLK_DEV_LOOKUP") == 0))\
{\
	fl_test_type = TYPE;\
        if (fl_test_type == 0)\
        {\
                fl_test_ret = RET;\
        }\
        else if (fl_test_type == 1)\
        {\
                fl_test_ret = RET;\
        }\
}


#define FL_TEST_F_main_N_NOT_BLK_DEV_FILE_DELETE(TYPE, RET) \
if ((strcmp(fl_test_func, "main") == 0) &&\
    (strcmp(fl_test_name, "NOT_BLK_DEV_FILE_DELETE") == 0))\
{\
	fl_test_type = TYPE;\
        if (fl_test_type == 0)\
        {\
                fl_test_ret = RET;\
        }\
        else if (fl_test_type == 1)\
        {\
                fl_test_ret = RET;\
        }\
}


#define FL_TEST_F_main_N_NOT_CHR_DEV_LOOKUP(TYPE, RET) \
if ((strcmp(fl_test_func, "main") == 0) &&\
    (strcmp(fl_test_name, "NOT_CHR_DEV_LOOKUP") == 0))\
{\
	fl_test_type = TYPE;\
        if (fl_test_type == 0)\
        {\
                fl_test_ret = RET;\
        }\
        else if (fl_test_type == 1)\
        {\
                fl_test_ret = RET;\
        }\
}


#define FL_TEST_F_main_N_NOT_CHR_DEV_FILE_DELETE(TYPE, RET) \
if ((strcmp(fl_test_func, "main") == 0) &&\
    (strcmp(fl_test_name, "NOT_CHR_DEV_FILE_DELETE") == 0))\
{\
	fl_test_type = TYPE;\
        if (fl_test_type == 0)\
        {\
                fl_test_ret = RET;\
        }\
        else if (fl_test_type == 1)\
        {\
                fl_test_ret = RET;\
        }\
}


#define FL_TEST_F_main_N_BITFIELD_IS_FULL(TYPE, RET) \
if ((strcmp(fl_test_func, "main") == 0) &&\
    (strcmp(fl_test_name, "BITFIELD_IS_FULL") == 0))\
{\
	fl_test_type = TYPE;\
        if (fl_test_type == 0)\
        {\
                fl_test_ret = RET;\
        }\
        else if (fl_test_type == 1)\
        {\
                fl_test_ret = RET;\
        }\
}


#define FL_TEST_F_main_N_NOT_PREPARE_BLK_DEV_CREATION(TYPE, RET) \
if ((strcmp(fl_test_func, "main") == 0) &&\
    (strcmp(fl_test_name, "NOT_PREPARE_BLK_DEV_CREATION") == 0))\
{\
	fl_test_type = TYPE;\
        if (fl_test_type == 0)\
        {\
                fl_test_ret = RET;\
        }\
        else if (fl_test_type == 1)\
        {\
                fl_test_ret = RET;\
        }\
}


#define FL_TEST_F_main_N_NOT_CREATE_BLK_DEV_FILE(TYPE, RET) \
if ((strcmp(fl_test_func, "main") == 0) &&\
    (strcmp(fl_test_name, "NOT_CREATE_BLK_DEV_FILE") == 0))\
{\
	fl_test_type = TYPE;\
        if (fl_test_type == 0)\
        {\
                fl_test_ret = RET;\
        }\
        else if (fl_test_type == 1)\
        {\
                fl_test_ret = RET;\
        }\
}


#define FL_TEST_F_main_N_NOT_NEW_NODE(TYPE, RET) \
if ((strcmp(fl_test_func, "main") == 0) &&\
    (strcmp(fl_test_name, "NOT_NEW_NODE") == 0))\
{\
	fl_test_type = TYPE;\
        if (fl_test_type == 0)\
        {\
                fl_test_ret = RET;\
        }\
        else if (fl_test_type == 1)\
        {\
                fl_test_ret = RET;\
        }\
}


#define FL_TEST_F_main_N_LIST_TRUNCATED(TYPE, RET) \
if ((strcmp(fl_test_func, "main") == 0) &&\
    (strcmp(fl_test_name, "LIST_TRUNCATED") == 0))\
{\
	fl_test_type = TYPE;\
        if (fl_test_type == 0)\
        {\
                fl_test_ret = RET;\
                dsf_control.ret = RET;\
        }\
}


#define FL_TEST_F_main_N_NOT_ALLOC_PACKET_DATA(TYPE, RET) \
if ((strcmp(fl_test_func, "main") == 0) &&\
    (strcmp(fl_test_name, "NOT_ALLOC_PACKET_DATA") == 0))\
{\
	fl_test_type = TYPE;\
        if (fl_test_type == 0)\
        {\
                fl_test_ret = RET;\
        }\
        else if (fl_test_type == 1)\
        {\
                fl_test_ret = RET;\
        }\
}


// Driver test functions
#define FL_TEST_VARS_F_fl_loop_new() \
char *file_name;


#define FL_TEST_F_fl_open_N_NOT_MAP_DEVICE(STATE, RET, ARGS) \
if ((strcmp(fl_test_func, "fl_open") == 0) &&\
    (strcmp(fl_test_name, "NOT_MAP_DEVICE") == 0))\
{\
	fl_test_state = STATE;\
        if (fl_test_state == FL_TEST_STATE_BEFORE)\
        {\
                fl_test_ret = RET;\
		args.ret.value = 0;\
        	args.ret.use_value = 1;\
        	args.ret.reset = 1;\
        }\
        else if (fl_test_state == FL_TEST_STATE_ERROR)\
        {\
                fl_test_ret = RET;\
                strcpy(fl_test_err, "ERROR: Couldn't map device numbers");\
        }\
}


#define FL_TEST_F_fl_open_N_HAS_RESET_SWITCH_ON(STATE, RET, ARGS) \
if ((strcmp(fl_test_func, "fl_open") == 0) &&\
    (strcmp(fl_test_name, "HAS_RESET_SWITCH_ON") == 0))\
{\
	fl_test_state = STATE;\
        if (fl_test_state == FL_TEST_STATE_BEFORE)\
        {\
                fl_test_ret = RET;\
		reset_switch = 1;\
        }\
        else if (fl_test_state == FL_TEST_STATE_ERROR)\
        {\
                fl_test_ret = RET;\
                strcpy(fl_test_err, "ERROR: Semaphore value overflow");\
        }\
}


#define FL_TEST_F_fl_open_N_NOT_SUPER_USER(STATE, RET, ARGS) \
if ((strcmp(fl_test_func, "fl_open") == 0) &&\
    (strcmp(fl_test_name, "NOT_SUPER_USER") == 0))\
{\
	fl_test_state = STATE;\
        if (fl_test_state == FL_TEST_STATE_BEFORE)\
        {\
                fl_test_ret = RET;\
		args.ret.value = 0;\
        	args.ret.use_value = 1;\
        	args.ret.reset = 1;\
        }\
        else if (fl_test_state == FL_TEST_STATE_ERROR)\
        {\
                fl_test_ret = RET;\
                strcpy(fl_test_err, "ERROR: Not super user");\
        }\
}


#define FL_TEST_F_fl_open_N_INTERRUPT_CTRL_SEM(STATE, RET, ARGS) \
if ((strcmp(fl_test_func, "fl_open") == 0) &&\
    (strcmp(fl_test_name, "INTERRUPT_CTRL_SEM") == 0))\
{\
	fl_test_state = STATE;\
        if (fl_test_state == FL_TEST_STATE_BEFORE)\
        {\
                fl_test_ret = RET;\
		args.ret.value = 0;\
        	args.ret.use_value = 1;\
        	args.ret.reset = 1;\
        }\
        else if (fl_test_state == FL_TEST_STATE_ERROR)\
        {\
                fl_test_ret = RET;\
                strcpy(fl_test_err, "ERROR: Execution interrupted by user");\
        }\
}


#if FL_OS_NAME_V_OSF1
#       if FL_OS_VERSION_V_5_1
#               if FL_OS_RELEASE_V_B


#               endif
#       endif
#elif FL_OS_NAME_V_IRIX
#       if FL_OS_VERSION_V_6_5
#               if FL_OS_RELEASE_V_30


#define FL_TEST_F_fl_open_N_OPEN_CONTENTION(STATE, RET, ARGS)


#               endif
#       endif
#endif


#define FL_TEST_F_fl_open_N_RESET_SWITCH_ON_INCR(STATE, RET, ARGS) \
if ((strcmp(fl_test_func, "fl_open") == 0) &&\
    (strcmp(fl_test_name, "RESET_SWITCH_ON_INCR") == 0))\
{\
	fl_test_state = STATE;\
        if (fl_test_state == FL_TEST_STATE_BEFORE)\
        {\
                fl_test_ret = RET;\
		args.ret.value = 0;\
        	args.ret.use_value = 1;\
        	args.ret.reset = 1;\
        }\
        else if (fl_test_state == FL_TEST_STATE_ERROR)\
        {\
                fl_test_ret = RET;\
                strcpy(fl_test_err, "ERROR: Execution interrupted by user");\
        }\
}


#define FL_TEST_F_fl_open_N_INVALID_OPEN_FLAGS_CHR(STATE, RET, ARGS) \
if ((strcmp(fl_test_func, "fl_open") == 0) &&\
    (strcmp(fl_test_name, "INVALID_OPEN_FLAGS_CHR") == 0))\
{\
	fl_test_state = STATE;\
        if (fl_test_state == FL_TEST_STATE_ERROR)\
        {\
                fl_test_ret = RET;\
                strcpy(fl_test_err, "ERROR: Invalid open flags for CHR device");\
        }\
}


#define FL_TEST_F_fl_open_N_NOT_FOUND_LOOP_ENTRY(STATE, RET, ARGS) \
if ((strcmp(fl_test_func, "fl_open") == 0) &&\
    (strcmp(fl_test_name, "NOT_FOUND_LOOP_ENTRY") == 0))\
{\
	fl_test_state = STATE;\
        if (fl_test_type == FL_TEST_STATE_BEFORE)\
        {\
                fl_test_ret = RET;\
		args.ret.value = 0;\
        	args.ret.use_value = 1;\
        	args.ret.reset = 1;\
        }\
        else if (fl_test_type == FL_TEST_STATE_ERROR)\
        {\
                fl_test_ret = RET;\
                strcpy(fl_test_err, "ERROR: Couldn't get loop entry");\
        }\
}


#define FL_TEST_F_fl_open_N_IS_IN_EXCLUSIVE_USE(STATE, RET, ARGS) \
if ((strcmp(fl_test_func, "fl_open") == 0) &&\
    (strcmp(fl_test_name, "IS_IN_EXCLUSIVE_USE") == 0))\
{\
	fl_test_state = STATE;\
        if (fl_test_type == FL_TEST_STATE_BEFORE)\
        {\
                fl_test_ret = RET;\
		loop_entry->is_exclusive = 1;\
        }\
        else if (fl_test_type == FL_TEST_STATE_ERROR)\
        {\
                fl_test_ret = RET;\
                strcpy(fl_test_err, "ERROR: Device is being used in exclusive access mode");\
        }\
}


#define FL_TEST_F_fl_open_N_INVALID_OPEN_FLAGS_BLK(STATE, RET, ARGS) \
if ((strcmp(fl_test_func, "fl_open") == 0) &&\
    (strcmp(fl_test_name, "INVALID_OPEN_FLAGS_BLK") == 0))\
{\
	fl_test_state = STATE;\
        if (fl_test_state == FL_TEST_STATE_ERROR)\
        {\
                fl_test_ret = RET;\
                strcpy(fl_test_err, "ERROR: Invalid open flags for BLK device");\
        }\
}


#define FL_TEST_F_fl_read_N_WRONG_SESSION_STATE(TYPE, RET) \
if ((strcmp(fl_test_func, "fl_read") == 0) &&\
    (strcmp(fl_test_name, "WRONG_SESSION_STATE") == 0))\
{\
fl_test_ret = RET;\
strcpy(fl_test_err, "ERROR: Wrong session state in read entry point");\
}


#define FL_TEST_F_fl_write_N_WRONG_SESSION_STATE(TYPE, RET) \
if ((strcmp(fl_test_func, "fl_write") == 0) &&\
    (strcmp(fl_test_name, "WRONG_SESSION_STATE") == 0))\
{\
fl_test_ret = RET;\
strcpy(fl_test_err, "ERROR: Wrong session state in write entry point");\
}


#define FL_TEST_F_fl_write_N_NOT_ENOUGH_DATA(TYPE, RET) \
if ((strcmp(fl_test_func, "fl_write") == 0) &&\
    (strcmp(fl_test_name, "NOT_ENOUGH_DATA") == 0))\
{\
fl_test_ret = RET;\
strcpy(fl_test_err, "ERROR: Not enough data");\
}


#define FL_TEST_F_fl_write_N_NOT_COPY_DATA_USERSPACE(TYPE, RET) \
if ((strcmp(fl_test_func, "fl_write") == 0) &&\
    (strcmp(fl_test_name, "NOT_COPY_DATA_USERSPACE") == 0))\
{\
	fl_test_type = TYPE;\
        if (fl_test_type == 0)\
        {\
                fl_test_ret = RET;\
        }\
        else if (fl_test_type == 1)\
        {\
                fl_test_ret = RET;\
		strcpy(fl_test_err, "ERROR: Couldn't copy data from user space");\
        }\
}


#define FL_TEST_F_fl_write_N_NOT_ALLOC_REQ_PACKET_AT(TYPE, RET) \
if ((strcmp(fl_test_func, "fl_write") == 0) &&\
    (strcmp(fl_test_name, "NOT_ALLOC_REQ_PACKET_AT") == 0))\
{\
	fl_test_type = TYPE;\
        if (fl_test_type == 0)\
        {\
                fl_test_ret = RET;\
        }\
        else if (fl_test_type == 1)\
        {\
                fl_test_ret = RET;\
		strcpy(fl_test_err, "ERROR: Couldn't allocate request packet (FL_CMD_ATTACH)");\
        }\
}


#define FL_TEST_F_fl_packet_read_N_NOT_ALLOC_DEV_NAME_AT(TYPE, RET) \
if ((strcmp(fl_test_func, "fl_packet_read") == 0) &&\
    (strcmp(fl_test_name, "NOT_ALLOC_DEV_NAME_AT") == 0))\
{\
	fl_test_type = TYPE;\
	if (fl_test_type == 0)\
	{\
		fl_test_ret = RET;\
	}\
        else if (fl_test_type == 1)\
        {\
                fl_test_ret = RET;\
                strcpy(fl_test_err, "ERROR: Couldn't allocate device file name (FL_CMD_ATTACH)");\
        }\
}


#define FL_TEST_F_fl_packet_read_N_NOT_ALLOC_FILE_NAME_AT(TYPE, RET) \
if ((strcmp(fl_test_func, "fl_packet_read") == 0) &&\
    (strcmp(fl_test_name, "NOT_ALLOC_FILE_NAME_AT") == 0))\
{\
	fl_test_type = TYPE;\
        if (fl_test_type == 0)\
        {\
                fl_test_ret = RET;\
        }\
        else if (fl_test_type == 1)\
        {\
                fl_test_ret = RET;\
                strcpy(fl_test_err, "ERROR: Couldn't allocate file name (FL_CMD_ATTACH)");\
        }\
}


#define FL_TEST_F_fl_write_N_NOT_READ_REQ_PACKET_DE(TYPE, RET) \
if ((strcmp(fl_test_func, "fl_write") == 0) &&\
    (strcmp(fl_test_name, "NOT_READ_REQ_PACKET_DE") == 0))\
{\
	fl_test_type = TYPE;\
        if (fl_test_type == 0)\
        {\
                fl_test_ret = RET;\
        }\
        else if (fl_test_type == 1)\
        {\
                fl_test_ret = RET;\
                strcpy(fl_test_err, "ERROR: Couldn't read request packet (FL_CMD_DETACH)");\
        }\
}


#define FL_TEST_F_fl_packet_read_N_NOT_ALLOC_BLK_DEV_NAME_DE(TYPE, RET) \
if ((strcmp(fl_test_func, "fl_packet_read") == 0) &&\
    (strcmp(fl_test_name, "NOT_ALLOC_BLK_DEV_NAME_DE") == 0))\
{\
	fl_test_type = TYPE;\
        if (fl_test_type == 0)\
        {\
                fl_test_ret = RET;\
        }\
        else if (fl_test_type == 1)\
        {\
                fl_test_ret = RET;\
                strcpy(fl_test_err, "ERROR: Couldn't allocate BLK device file name (FL_CMD_DETACH)");\
        }\
}


#define FL_TEST_F_fl_proc_get_N_NOT_FOUND_PID(TYPE, RET) \
if ((strcmp(fl_test_func, "fl_proc_get") == 0) &&\
    (strcmp(fl_test_name, "NOT_FOUND_PID") == 0))\
{\
	fl_test_type = TYPE;\
        if (fl_test_type == 0)\
        {\
                fl_test_ret = RET;\
        }\
        else if (fl_test_type == 1)\
        {\
                fl_test_ret = RET;\
                strcpy(fl_test_err, "ERROR: Couldn't find pid");\
        }\
}


#define FL_TEST_F_fl_nameidata_dup_N_not_alloc_utask(TYPE, RET) \
if ((strcmp(fl_test_func, "fl_nameidata_dup") == 0) &&\
    (strcmp(fl_test_name, "not_alloc_utask") == 0))\
{\
	fl_test_type = TYPE;\
        if (fl_test_type == 0)\
        {\
                fl_test_ret = RET;\
        }\
        else if (fl_test_type == 1)\
        {\
                fl_test_ret = RET;\
		strcpy(fl_test_err, "ERROR: Couldn't allocate utask");\
        }\
}


#define FL_TEST_F_fl_nameidata_dup_N_not_alloc_nameidata(TYPE, RET) \
if ((strcmp(fl_test_func, "fl_nameidata_dup") == 0) &&\
    (strcmp(fl_test_name, "not_alloc_nameidata") == 0))\
{\
	fl_test_type = TYPE;\
        if (fl_test_type == 0)\
        {\
                fl_test_ret = RET;\
        }\
        else if (fl_test_type == 1)\
        {\
                fl_test_ret = RET;\
		strcpy(fl_test_err, "ERROR: Couldn't allocate nameidata structure");\
        }\
}


#define FL_TEST_F_fl_nameidata_dup_N_not_dup_nameidata(TYPE, RET) \
if ((strcmp(fl_test_func, "fl_nameidata_dup") == 0) &&\
    (strcmp(fl_test_name, "not_dup_nameidata") == 0))\
{\
	fl_test_type = TYPE;\
        if (fl_test_type == 0)\
        {\
                fl_test_ret = RET;\
        }\
        else if (fl_test_type == 1)\
        {\
                fl_test_ret = RET;\
		strcpy(fl_test_err, "ERROR: Couldn't duplicate nameidata");\
        }\
}


#define FL_TEST_F_fl_open_N_SIMPLE_OPEN_DEVFL(TYPE, RET) \
if ((strcmp(fl_test_func, "fl_open") == 0) &&\
    (strcmp(fl_test_name, "SIMPLE_OPEN_DEVFL") == 0))\
{\
	fl_test_type = TYPE;\
        if (fl_test_type == 0)\
        {\
                fl_test_ret = RET;\
                strcpy(fl_test_suc, "SUCCESS: /dev/fl device has been opened");\
        }\
}


#define FL_TEST_F_fl_open_N_SIMPLE_OPEN_ATTACHED_BLK_DEV(TYPE, RET) \
if ((strcmp(fl_test_func, "fl_open") == 0) &&\
    (strcmp(fl_test_name, "SIMPLE_OPEN_ATTACHED_BLK_DEV") == 0))\
{\
	fl_test_type = TYPE;\
        if (fl_test_type == 0)\
        {\
                fl_test_ret = RET;\
                strcpy(fl_test_suc, "SUCCESS: Attached BLK device has been opened");\
        }\
}


#define FL_TEST_F_fl_open_N_NOT_FOUND_LOOP_ENTRY_CHR_DEV(TYPE, RET) \
if ((strcmp(fl_test_func, "fl_open") == 0) &&\
    (strcmp(fl_test_name, "NOT_FOUND_LOOP_ENTRY_CHR_DEV") == 0))\
{\
	fl_test_type = TYPE;\
        if (fl_test_type == 0)\
        {\
                fl_test_ret = RET;\
                strcpy(fl_test_err, "ERROR: Loop entry not found for CHR device");\
        }\
}


#define FL_TEST_F_fl_open_N_NOT_FOUND_LOOP_ENTRY_BLK_DEV(TYPE, RET) \
if ((strcmp(fl_test_func, "fl_open") == 0) &&\
    (strcmp(fl_test_name, "NOT_FOUND_LOOP_ENTRY_BLK_DEV") == 0))\
{\
	fl_test_type = TYPE;\
        if (fl_test_type == 0)\
        {\
                fl_test_ret = RET;\
                strcpy(fl_test_err, "ERROR: Loop entry not found for BLK device");\
        }\
}


#define FL_TEST_F_fl_list_push_data_N_NOT_CREATE_NODE(TYPE, RET) \
if ((strcmp(fl_test_func, "fl_list_push_data") == 0) &&\
    (strcmp(fl_test_name, "NOT_CREATE_NODE") == 0))\
{\
	fl_test_type = TYPE;\
        if (fl_test_type == 0)\
        {\
                fl_test_ret = RET;\
        }\
        else if (fl_test_type == 1)\
        {\
                fl_test_ret = RET;\
                strcpy(fl_test_err, "ERROR: Couldn't create node");\
        }\
}


#define FL_TEST_F_fl_strategy_N_READ_MODE_ONLY(TYPE, RET) \
if ((strcmp(fl_test_func, "fl_strategy") == 0) &&\
    (strcmp(fl_test_name, "READ_MODE_ONLY") == 0))\
{\
	fl_test_type = TYPE;\
	if (fl_test_type == 0)\
        {\
		fl_test_notify++;\
        }\
        else if (fl_test_type == 1)\
        {\
                fl_test_ret = RET;\
                strcpy(fl_test_err, "ERROR: Only read mode is supported");\
		fl_test_notify++;\
        }\
}


#define FL_TEST_F_fl_strategy_N_OFFSET_OUT_OF_RANGE(TYPE, RET) \
if ((strcmp(fl_test_func, "fl_strategy") == 0) &&\
    (strcmp(fl_test_name, "OFFSET_OUT_OF_RANGE") == 0))\
{\
	fl_test_type = TYPE;\
        if (fl_test_type == 0)\
        {\
                fl_test_ret = RET;\
        }\
        else if (fl_test_type == 1)\
        {\
                fl_test_ret = RET;\
                strcpy(fl_test_err, "ERROR: Request offset is out of range");\
        }\
}


#define FL_TEST_F_custom_N_PINGPONG(STATE, RET, ARGS) \
if ((strcmp(fl_test_func, "custom") == 0) &&\
    (strcmp(fl_test_name, "PINGPONG") == 0))\
{\
	fl_test_type = TYPE;\
        if (fl_test_type == 0)\
        {\
		fl_test_counter = 0;\
        }\
        else if (fl_test_type == 1)\
        {\
		fl_test_counter++;\
                fl_test_ret = EINVAL;\
                strcpy(fl_test_err, "ERROR: Process is reset");\
        }\
}


#define FL_TEST_F_fl_write_N_UNKNOWN_COMMAND(TYPE, RET) \
if ((strcmp(fl_test_func, "fl_write") == 0) &&\
    (strcmp(fl_test_name, "UNKNOWN_COMMAND") == 0))\
{\
	fl_test_type = TYPE;\
        if (fl_test_type == 0)\
        {\
                fl_test_ret = RET;\
                strcpy(fl_test_err, "ERROR: Unknown command");\
        }\
}


#else


#define FL_TEST_DECLS
#define FL_TEST_TYPE_DECL
#define FL_TEST_TYPE_INIT
#define FL_TEST_COND(COND, OP)
#define FL_TEST_VARS_F_fl_loop_new()


#define FL_TEST_F_main_N_NOT_SUSER_OPEN_DEVFL(TYPE, RET)
#define FL_TEST_F_main_N_NO_PERMISSION_CPU_CHECK(TYPE, RET)
#define FL_TEST_F_main_N_NOT_OPEN_DEVFL(TYPE, RET)
#define FL_TEST_F_main_N_NO_PERMISSION_WRITE_REQUEST_PACKET(TYPE, RET)
#define FL_TEST_F_main_N_NOT_WRITE_REQUEST_PACKET_DEVFL(TYPE, RET)
#define FL_TEST_F_main_N_NO_PERMISSION_PACKET_SIZE(TYPE, RET)
#define FL_TEST_F_main_N_NOT_READ_RESPONSE_PACKET_SIZE_DEVFL(TYPE, RET)
#define FL_TEST_F_main_N_NO_PERMISSION_PACKET_DATA(TYPE, RET)
#define FL_TEST_F_main_N_NOT_READ_RESPONSE_PACKET_DEVFL(TYPE, RET)
#define FL_TEST_F_main_N_NOT_ALLOC_LOOPENTRY(TYPE, RET)
#define FL_TEST_F_main_N_NOT_CREATE_CHR_PIDLIST(TYPE, RET)
#define FL_TEST_F_main_N_NOT_CREATE_BLK_PIDLIST(TYPE, RET)
#define FL_TEST_F_main_N_NOT_ALLOC_VATTR(TYPE, RET)
#define FL_TEST_F_main_N_NOT_ALLOC_CHR_DEV_NAME(TYPE, RET)
#define FL_TEST_F_main_N_NOT_ALLOC_UTASK(TYPE, RET)
#define FL_TEST_F_main_N_NOT_ALLOC_NAMEIDATA(TYPE, RET)
#define FL_TEST_F_main_N_NOT_DUP_NAMEIDATA(TYPE, RET)
#define FL_TEST_F_main_N_BLK_DEV_ALREADY_ATTACHED(TYPE, RET)
#define FL_TEST_F_main_N_NOT_LOOKUP_FILENAME(TYPE, RET)
#define FL_TEST_F_main_N_NOT_FILE_STAT(TYPE, RET)
#define FL_TEST_F_main_N_NOT_OPEN_FILE(TYPE, RET)
#define FL_TEST_F_main_N_NOT_BLK_DEV_LOOKUP(TYPE, RET)
#define FL_TEST_F_main_N_NOT_BLK_DEV_FILE_DELETE(TYPE, RET)
#define FL_TEST_F_main_N_NOT_CHR_DEV_LOOKUP(TYPE, RET)
#define FL_TEST_F_main_N_NOT_CHR_DEV_FILE_DELETE(TYPE, RET)
#define FL_TEST_F_main_N_BITFIELD_IS_FULL(TYPE, RET)
#define FL_TEST_F_main_N_NOT_PREPARE_BLK_DEV_CREATION(TYPE, RET)
#define FL_TEST_F_main_N_NOT_CREATE_BLK_DEV_FILE(TYPE, RET)
#define FL_TEST_F_main_N_NOT_NEW_NODE(TYPE, RET)
#define FL_TEST_F_main_N_LIST_TRUNCATED(TYPE, RET)
#define FL_TEST_F_main_N_NOT_ALLOC_PACKET_DATA(TYPE, RET)
#define FL_TEST_F_fl_open_N_NOT_MAP_DEVICE(STATE, RET, ARGS)
#define FL_TEST_F_fl_open_N_HAS_RESET_SWITCH_ON(STATE, RET, ARGS)
#define FL_TEST_F_fl_open_N_NOT_SUPER_USER(STATE, RET, ARGS)
#define FL_TEST_F_fl_open_N_INTERRUPT_CTRL_SEM(STATE, RET, ARGS)
#define FL_TEST_F_fl_open_N_OPEN_CONTENTION(STATE, RET, ARGS)
#define FL_TEST_F_fl_open_N_NOT_FOUND_LOOP_ENTRY(STATE, RET, ARGS)
#define FL_TEST_F_fl_open_N_IS_IN_EXCLUSIVE_USE(STATE, RET, ARGS)
#define FL_TEST_F_fl_open_N_INVALID_OPEN_FLAGS_BLK(STATE, RET, ARGS)

#define FL_TEST_F_fl_read_N_WRONG_SESSION_STATE(TYPE, RET)
#define FL_TEST_F_fl_write_N_WRONG_SESSION_STATE(TYPE, RET)
#define FL_TEST_F_fl_write_N_NOT_ENOUGH_DATA(TYPE, RET)
#define FL_TEST_F_fl_write_N_NOT_COPY_DATA_USERSPACE(TYPE, RET)
#define FL_TEST_F_fl_write_N_NOT_ALLOC_REQ_PACKET_AT(TYPE, RET)
#define FL_TEST_F_fl_packet_read_N_NOT_ALLOC_DEV_NAME_AT(TYPE, RET)
#define FL_TEST_F_fl_packet_read_N_NOT_ALLOC_FILE_NAME_AT(TYPE, RET)
#define FL_TEST_F_fl_write_N_NOT_READ_REQ_PACKET_DE(TYPE, RET)
#define FL_TEST_F_fl_packet_read_N_NOT_ALLOC_BLK_DEV_NAME_DE(TYPE, RET)
#define FL_TEST_F_fl_proc_get_N_NOT_FOUND_PID(TYPE, RET)
#define FL_TEST_F_fl_nameidata_dup_N_not_alloc_utask(TYPE, RET)
#define FL_TEST_F_fl_nameidata_dup_N_not_alloc_nameidata(TYPE, RET)
#define FL_TEST_F_fl_nameidata_dup_N_not_dup_nameidata(TYPE, RET)
#define FL_TEST_F_fl_open_N_SIMPLE_OPEN_DEVFL(STATE, RET, ARGS)
#define FL_TEST_F_fl_open_N_SIMPLE_OPEN_ATTACHED_BLK_DEV(STATE, RET, ARGS)
#define FL_TEST_F_fl_open_N_INVALID_OPEN_FLAGS_CHR(STATE, RET, ARGS)
#define FL_TEST_F_fl_open_N_NOT_FOUND_LOOP_ENTRY_CHR_DEV(STATE, RET, ARGS)
#define FL_TEST_F_fl_open_N_INVALID_OPEN_FLAGS_BLK(STATE, RET, ARGS)
#define FL_TEST_F_fl_open_N_NOT_FOUND_LOOP_ENTRY_BLK_DEV(STATE, RET, ARGS)
#define FL_TEST_F_fl_list_push_data_N_NOT_CREATE_NODE(TYPE, RET)
#define FL_TEST_F_fl_strategy_N_READ_MODE_ONLY(TYPE, RET)
#define FL_TEST_F_fl_strategy_N_OFFSET_OUT_OF_RANGE(TYPE, RET)
//#define FL_TEST_F_custom_N_PINGPONG(TYPE, RET)
#define FL_TEST_F_custom_N_PINGPONG(STATE, RET, ARGS)
#define FL_TEST_F_fl_write_N_UNKNOWN_COMMAND(TYPE, RET)


#endif


#endif	// __FL_TEST_H__


