// NOTE: File looper mounts
//
//  In order to use the system mount(1M) user command with file looper,
//  the system mount(1M) executable must be started with some of its functions
//  masked by ad hoc versions using IRIX's prelinking/preloading facility.
//
//  These ad hoc functions are stored in a dynamic shared object located
//  in the same path as file looper's mount_iso9660:
//
//	/etc/mount_iso9660 - file looper's mount
//	/etc/mount_iso9660.so - file looper's ad hoc functions
//
//  The flow diagram for a system mount(1M) invokation changes to:
//
//			mount(1M)
//
//			    |
//			    V
//
//		   /usr/bin/mount_iso9660 (file looper's mount)
//
//			    |
//			    V
//
//		   /usr/etc/mount_iso9660 (system mount(1M) helper)
//
//
//  File looper's mount modifies the environment for the system mount(1M) helper,
//  by adding an _RLD_LIST environment variable pointing to the ad hoc functions
//  in /usr/lib32/mount_iso9660.so.
//
//  When the system mount(1M) helper gets executed, it preloads this DSO, masking its
//  own functions with the ad hoc functions.


#include <stdio.h>
#include <unistd.h>
#include <errno.h>
#include <limits.h>


int main(int argc,
	 char *argv[])
{
	char *RLD_LIST = "_RLD_LIST=/usr/lib32/mount_iso9660.so:DEFAULT";
	char *rld_list;
	unsigned int list_length = strlen(RLD_LIST);


	// NOTE: Add _RLD_LIST to environment


	// Create _RLD_LIST value
	rld_list = (char *) malloc(list_length + 1);
	if (rld_list == NULL)
	{
		fprintf(stderr,
			"ERROR: Couldn't allocate _RLD_LIST environment variable\n");


		// Set error number
		errno = ENOMEM;
		exit(1);
	}


	// Copy _RLD_LIST
	rld_list[0] = '\0';
	strcat(rld_list, RLD_LIST);
	rld_list[list_length] = '\0';


	// Set _RLD_LIST
	if (putenv(rld_list) != 0)
	{
		fprintf(stderr,
			"ERROR: Couldn't set _RLD_LIST environment variable\n");


		// Set error number
		errno = ENOMEM;
		exit(1);
	}


	// NOTE: Execute mount_iso9660
	//
	//  Execute system's mount_iso9660 helper (/usr/etc/mount_iso9660).


	if (execv("/usr/etc/mount_iso9660", argv) == -1)
	{
		fprintf(stderr,
			"ERROR: execv failed for /usr/etc/mount_iso9660\n");


		exit(1);
	}


	// NOTE: Never reached


	return 0;
}


