

#include <stdio.h>
#include <stdarg.h>
#include <dslib.h>
#include <string.h>
#include <limits.h>
#include <sys/types.h>
#include <sys/conf.h>
#include <sys/iograph.h>
#include <sys/stat.h>
#include <unistd.h>
#include <errno.h>
#include <dirent.h>


//#define MOUNT_DEBUG


#ifdef MOUNT_DEBUG


#define MOUNT_DEBUG_PRINTF(ARG) printf ## ARG;


#else


#define MOUNT_DEBUG_PRINTF(ARG)


#endif


// NOTE: External references


//extern int __Argc;
//extern char **__Argv;
//int argc;
//char **argv;


// NOTE: Global variables


int fd_to_scsi_name(int dkfd,
		    char *sdev)
{
	DIR *dirp;
	unsigned int dev_found;
	struct dirent *direntp;
	struct stat dk_stat;
	struct stat dent_stat;
	char chr_name[PATH_MAX + 1];
	char devname[MAXDEVNAME];
	char *tmpname;
	int devnmlen = MAXDEVNAME;


	MOUNT_DEBUG_PRINTF(("[fd_to_scsi_name] INFO: dkfd = %i sdev = \"%s\"\n",
			    dkfd,
			    sdev))


	// Get driver name for file descriptor
	dev_found = 0;
	if (fdes_to_drivername(dkfd,
			       devname,
			       &devnmlen))
	{
		// NOTE: Found driver name


		MOUNT_DEBUG_PRINTF(("[fd_to_scsi_name] INFO: devname = \"%s\"\n",
				    devname))


		// Check for dksc driver
		if(strcmp(devname,
			  "dksc") == 0)
		{
			// NOTE: dksc driver


			// Get device name for file descriptor
			devnmlen = MAXDEVNAME;
			if (fdes_to_devname(dkfd,
					    devname,
					    &devnmlen))
			{
				// NOTE: SCSI name


				// Substitute "scsi" for "disk"
				tmpname = strstr(devname,
						 EDGE_LBL_DISK);
				strcpy(tmpname,
				       EDGE_LBL_SCSI);


				// Set scsi name
				strcpy(sdev,
				       devname);


				return 0;
			}
		}
		else
		{
			// NOTE: Other driver
			//
			//  Check for /dev/loop file descriptor.


			MOUNT_DEBUG_PRINTF(("[fd_to_scsi_name] INFO: fstat()\n"))


			// Get stat for dkfd
			errno = 0;
			if (fstat(dkfd,
				  &dk_stat) == 0)
			{
				// Traverse /dev/loop
				dirp = opendir("/dev/loop");
				direntp = readdir(dirp);
				while (direntp != NULL)
				{
					MOUNT_DEBUG_PRINTF(("[fd_to_scsi_name] INFO: direntp->d_name = %s\n",
							    direntp->d_name))


					// Skip . and ..
					if (direntp->d_name[0] == '.')
					{
						if (direntp->d_name[1] == '\0')
						{
							MOUNT_DEBUG_PRINTF(("[fd_to_scsi_name] INFO: Skipped\n"))


							// Read next entry
							direntp = readdir(dirp);


							continue;
						}


						if (direntp->d_name[1] == '.')
						{
							if (direntp->d_name[2] == '\0')
							{
								MOUNT_DEBUG_PRINTF(("[fd_to_scsi_name] INFO: Skipped\n"))


								// Read next entry
								direntp = readdir(dirp);


								continue;
							}
						}
					}


					// Check for buffer overflow
					if (strlen(direntp->d_name) + 10 > PATH_MAX - 1)
					{
						MOUNT_DEBUG_PRINTF(("[fd_to_scsi_name] INFO: Skipped\n"))


						// Read next entry
						direntp = readdir(dirp);


						continue;
					}


					// Create absolute path to loop device file
					if (sprintf(chr_name,
						    "/dev/loop/%s",
						    direntp->d_name) < 0)
					{
						MOUNT_DEBUG_PRINTF(("[fd_to_scsi_name] ERROR: sprintf(%p, /dev/loop/%%s, \"%s\"\n",
								    chr_name,
								    direntp->d_name))


						break;
					}


					// Get stat for entry
					if (stat(chr_name,
						 &dent_stat) == 0)
					{
						MOUNT_DEBUG_PRINTF(("[fd_to_scsi_name] INFO: dent_stat.st_ino = %llu\n",
								    dent_stat.st_ino))


						// Check inode
						if (dent_stat.st_ino == dk_stat.st_ino)
						{
							MOUNT_DEBUG_PRINTF(("INFO: Found %s\n",
									    direntp->d_name))


							// Set SCSI name
							sdev[0] = '\0';
							strcat(sdev,
							       chr_name);


							// Set device found switch
							dev_found = 1;


							break;
						}
					}


					// Read next entry
					direntp = readdir(dirp);
				}


				closedir(dirp);
			}
		}
	}


	MOUNT_DEBUG_PRINTF(("[fd_to_scsi_name] INFO: dkfd = %i sdev = \"%s\"\n",
			    dkfd,
			    sdev))


	if (dev_found != 1)
	{
		return -1;
	}


	return 0;	// SUCCESS: /dev/loop/r* device found
}


int svc_register(int arg0,
		  int arg1,
		  int arg2,
		  int arg3,
		  int arg4)
{
	int ret;


	MOUNT_DEBUG_PRINTF(("[svc_register] INFO: START\n"))


	MOUNT_DEBUG_PRINTF(("[svc_register] INFO: __svc_register_local()\n"))


	ret = __svc_register_local(arg1,
				   arg2,
				   arg3,
				   arg4);
	if (ret == 0)
	{
		MOUNT_DEBUG_PRINTF(("[svc_register] ERROR: ret = 0\n"))


		return 0;	// ERROR
	}


	if (arg4 == 0)
	{
		MOUNT_DEBUG_PRINTF(("[svc_register] SUCCESS: arg4 = 0\n"))


		return 1;	// SUCCESS
	}


	MOUNT_DEBUG_PRINTF(("[svc_register] INFO: _pmap_set()\n"))


	//ret = _pmap_set(arg1,
	_pmap_set(arg1,
		  arg2,
		  arg4,
		  *((short *) (arg0 + 4)));


	MOUNT_DEBUG_PRINTF(("[svc_register] SUCCESS: ret = %i\n",
			    ret))


	MOUNT_DEBUG_PRINTF(("[svc_register] INFO: STOP\n"))


	return 1;	// SUCCESS
}


